/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.monitor;

import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;
import ostore.util.*;
import java.util.*;

/**
 * A resource we are monitoring.
 *
 * @author  Jeff Pang
 * @version $Id: Resource.java,v 1.2 2003/02/18 05:33:20 jeffpang Exp $
 **/
public abstract class Resource
{
    // the resource log
    private ResourceLog log;
    // name of resource
    private String name;

    // If the resource log is only monitoring a single resource value
    // then this is a temp double[1] array used to pass values to the log.
    // Otherwise this is null.
    private double[] tempUtil;
    
    /**
     * New resource.
     *
     * @param id node id
     * @param name name of resource
     * @param initial initial utilization or status values. The size of
     * this array determines the number of values in each log entry.
     */
    protected Resource(NodeId id, String name, double[] initial)
    {
	this.name = name;
	this.log = new ResourceLog(id, initial);
	this.tempUtil = new double[1];
    }

    /**
     * New resource. Assuming a single utilization value.
     *
     * @param id node id
     * @param name name of resource
     * @param initial initial utilization value.
     */
    protected Resource(NodeId id, String name, double initial)
    {
	this.name = name;
	this.tempUtil = new double[1];
	tempUtil[0] = initial;
	this.log = new ResourceLog(id, tempUtil);
    }

    /**
     * Get the resource log.
     */
    public ResourceLog getLog()
    {
	return log;
    }

    /**
     * Log a new set of utilization or status values.
     */
    public void log(double[] val)
    {
	log.log(val);
    }

    /**
     * Log a new utilization value. Assumes this only has a single utilization
     * value per log entry.
     */
    public void log(double val)
    {
	tempUtil[0] = val;
	log(tempUtil);
    }

    /**
     * Get the name of the resource.
     */
    public String getName()
    {
	return name;
    }
}
