/**
 * NetworkMessage.
 *
 * @author     Sean C. Rhea
 * @version    $Id: NetworkMessage.java,v 1.22 2003/06/27 03:45:59 srhea Exp $
 */

package ostore.network;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;

public abstract class NetworkMessage 
implements QueueElementIF, QuickSerializable, Cloneable {

    /**
     * If <code>inbound=true</code> then <code>peer</code> is the NodeId
     * of the node where this message came from; otherwise, it's the NodeId
     * of the node where this message should be sent.
     */
    public NodeId peer;
    public boolean inbound;

    /**
     * DO NOT TOUCH THIS FIELD.  That said, one of you will ignore me; if
     * you are that person, you had better understand
     * ostore.dispatch.Classifier and ostore.network.MultiplexedNetwork.
     */
    public NodeId sender;

    /** 
     * Times to attempt sending the message before giving up.  Depending on
     * the underlying network layer, this field may be ignored.
     */
    public int tries;

    /**
     * A completion queue to enqueue a success or failure on after this
     * message is successfully acknowledged or times out tries times,
     * respectively; the enqueued object will be of type NetworkMessageResult;
     * if this field is null, no event will be enqueued.  Depending on the
     * underlying network layer, this field may be ignored.
     */
    public SinkIF comp_q;

    /**
     * A value to place in the NetworkMessageResult so that the calling stage
     * can associate the result with this request.  Depending on the
     * underlying network layer, this field may be ignored.
     */
    public Object user_data;

    public NetworkMessage (NodeId peer, boolean inbound) {
	this.peer = peer;
	this.inbound = inbound;
    }

    public NetworkMessage (InputBuffer buffer) throws QSException {
	//peer = new NodeId (buffer); // don't use typecode
	//sender = new NodeId (buffer); // don't use typecode
    }

    public void serialize (OutputBuffer buffer) {
	// TODO: only need to serialize the port numbers...
	//peer.serialize (buffer); // don't use typecode
	//sender.serialize (buffer); // don't use typecode
    }

    public Object clone () throws CloneNotSupportedException {
	NetworkMessage result = (NetworkMessage) super.clone ();
	result.peer = peer;
	result.sender = sender;
	result.inbound = inbound;
	return result;
    }

    public String toString () {
	return "(NetworkMessage peer=" + peer + ", sender=" + sender +
	    " " + (inbound ? "inbound" : "outbound") + ")";
    }
}

