/**
 * NodeStatusMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: NodeStatusMsg.java,v 1.2 2003/02/25 23:21:24 strib Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;
import ostore.util.NodeId;
import seda.sandStorm.api.QueueElementIF;

/**
 * Sent by Patchwork to other stages who care if nodes fail or come back.
 * Contains a status field, which denotes how the status of the node being
 * monitored has changed.
 *
 * @author Jeremy Stribling
 * @version    $Id: NodeStatusMsg.java,v 1.2 2003/02/25 23:21:24 strib Exp $
 */
public class NodeStatusMsg implements QueueElementIF {

    public NodeId peer;
    public int status;
    public Object userData;

    public static final int STATUS_DEAD = 0;
    public static final int STATUS_ALIVE = 1;

    /**
     */
    public NodeStatusMsg (NodeId peer, int status, Object userData) {
	this.peer = peer;
	this.status = status;
	this.userData = userData;
    }

    public Object clone () throws CloneNotSupportedException {
	NodeStatusMsg newone = (NodeStatusMsg) super.clone ();
	newone.peer = peer;
	newone.status = status;
	newone.userData = userData;
	return newone;
    }

    public String toString () {
	return "(NodeStatusMsg, peer " + peer + ", status " + status + ")";
    }
}

