/**
 * TapestryLocationMsg.
 *
 * @author     Sean C. Rhea
 * @version    $Id: TapestryLocateMsg.java,v 1.17 2002/10/30 09:26:13 strib Exp $
 */
package ostore.tapestry.api;
import ostore.tapestry.channel.MsgQos;
import ostore.tapestry.impl.Router;
import ostore.util.SecureHash;
import ostore.util.QuickSerializable;
import seda.sandStorm.api.QueueElementIF;

/**
 * Messages sent from one OceanStore machine to one or more machines
 * avertising the given GUID and satisfying the given tag query.  The
 * intended functionality of this event (outgoing) is as follows.  The
 * message is routed toward the given {@link #guid} in the Tapestry mesh,
 * and at each hop the {@link #query} is applied to any backpointers
 * associated with the given GUID, along with the current {@link
 * #query_state}.  The query should return a result, and possibly modify
 * the query state as well.  The result can be one of {@link
 * TapestryQueryResultMatch} or {@link TapestryQueryResultNoMatch}.  A
 * matching result indicates which backpointers to follow, as well as
 * whether or not to continue up the tree.  A non-matching result causes
 * the query to continue towards the root without following any
 * backpointers.
 */
public abstract class TapestryLocateMsg 
implements QueueElementIF, QuickSerializable {

    /**
     * If <code>inbound</code>, the sender; undefined otherwise.
     */
    public SecureHash peer;

     /**
     * Whether this message is being received (<code>true</code>) or sent
     * (<code>false</code>).
     */
    public boolean inbound;

    /**
     * The queried GUID.
     */
    public SecureHash guid;

    /**
     * The query to execute on each backpointer.  See the class comments,
     * above.
     */
    public TapestryQuery query;

    /**
     * The state to pass the query executed on each backpointer.  See the
     * class comments, above.
     */
    public TapestryQueryState query_state;

    /**
     * Number of hops taken by the message on its way to the application
     */
    public int hopCount;

    /**
     * Number of hops this message should take before delivering payload.
     */
    public int TTL;

    /**
     * A unique ID for this message, in order that a
     * <code>TapestryLocateFailure</code> can be matched to the location
     * request; set to <code>null</code> if you do not want failure
     * notification.
     */
    public SecureHash id;

    /**
     * The delivery qos of this msg, added by Ling Huang
     */
    public int qos;

    /**
     * A bitmask indicating which local optimizations to use.
     * Bits indicate the following optimizations, from least significant bit
     *   to most significant bit:
     *   Bit 3) Before leaving a LAN, misroute and try to locate.
     * Note: "a few hops" is defined by Router.LOCAL_OPT_HOPS
     */
    public int local_opts;

    /**
     * Construct with the given guid, query, and state; will return a
     * {@link TapestryLocateFailure} with the same <code>id</code> if
     * the location fails.  Set <CODE>id</CODE> to <CODE>null</CODE>
     * if a failure respons is not needed. Use for new outbound messages.
     */
    public TapestryLocateMsg (SecureHash guid, TapestryQuery query, 
	    TapestryQueryState query_state, SecureHash id) {

	this.peer = null;
	this.inbound = false;
	this.guid = guid;
	this.query = query;
	this.query_state = query_state;
	this.id = id;
	this.TTL = -1;
	this.qos = MsgQos.UNRELIABLE;
	local_opts = 0;
    }

    /**
     * Empty constructor.  For use by derived types when reconstructing
     * themselves from a byte array.
     */
    public TapestryLocateMsg () 
    {
	this.qos = MsgQos.UNRELIABLE;
    }

    public void setLocalOptLocalMisroute() {
	local_opts = local_opts | Router.LOCAL_OPT_PUBLISH_LOCAL_MISROUTE;
    }

}













