/*
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.api;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;

/**
 * A (NodeId, Guid, Latency) tuple.  rtt_ms is only an estimate
 * (obviously), and may not be filled in at all; e.g. it will be equal to
 * Double.MAX_VALUE if we don't have an estimate for it yet.  If so, and
 * you want a better estimate, use Patchwork.
 * 
 * @author Sean C. Rhea
 * @version $Id: TapestryNeighborInfo.java,v 1.1 2003/03/03 05:55:09 srhea Exp $
 */
public class TapestryNeighborInfo {

    public NodeId node_id;
    public SecureHash guid;
    public double rtt_ms = Double.MAX_VALUE;

    public TapestryNeighborInfo (NodeId n, SecureHash g, double r) {
	node_id = n; guid = g; rtt_ms = r;
    }

    public TapestryNeighborInfo (NodeId n, SecureHash g) {
	node_id = n; guid = g; rtt_ms = Double.MAX_VALUE;
    }

    public boolean equals (Object rhs) {
	if (! (rhs instanceof TapestryNeighborInfo))
	    return false;
	TapestryNeighborInfo other = (TapestryNeighborInfo) rhs;
	return node_id.equals (other.node_id) && guid.equals (other.guid);
    }

    public int hashCode () {
	return node_id.hashCode () ^ guid.hashCode ();
    }

    public String toString () {
	return "(" + node_id.address ().getHostAddress () + ":" + 
	    node_id.port () + ", " + guid + 
	    ((rtt_ms == Double.MAX_VALUE) ? "" : (", " + rtt_ms + " ms")) + 
	    ")";
    }
}

