/**
 * ChannelRequestMsg class.
 *
 * @author     Ling Huang
 * @version    $Id: ChannelRequestMsg.java,v 1.2 2002/07/20 19:18:15 srhea Exp $
 */

package ostore.tapestry.channel;
import java.util.Random;
import ostore.tapestry.impl.*;
import ostore.network.NetworkMessage;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.CountBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSArray;
import ostore.util.QSByteArray;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.TypeTable;


//Source should send ChannelRequestMsg to destination for constructing a
public class ChannelRequestMsg extends RouteMsg
{
    protected int channel_id;  //At this time, we make it local unique. 

    public ChannelRequestMsg()
    {
	super();
	this.channel_id = 0;

	this.prefix_only = false; //Channel never do prefix only routing.

	// Default values for hop and TTL: hop start from 0, and TTL is big enough!
	this.hop = 0;
	this.TTL = 100;
    }

    public ChannelRequestMsg (InputBuffer buffer) throws QSException 
    {
	super (buffer);
	channel_id = buffer.nextInt ();
    }

    public ChannelRequestMsg(NodeId peer, SecureHash src, SecureHash dest)
    {
	super(peer,src,dest);
	this.channel_id = 0;

	this.prefix_only = false; //ChannelRequestMsg never do prefix only routing.

	// Default values for hop and TTL: hop start from 0, and TTL is big enough!
	this.hop = 0;
	this.TTL = 100;
    }

    public ChannelRequestMsg(NodeId peer, SecureHash src, SecureHash dest,int channel_id)
    {
	super(peer,src,dest);
	this.channel_id = channel_id;

	this.prefix_only = false; //ChannelRequestMsg never do prefix only routing.

	// Default values for hop and TTL: hop start from 0, and TTL is big enough!
	this.hop = 0;
	this.TTL = 100;
    }

    public ChannelRequestMsg(NodeId peer, SecureHash src, SecureHash dest,int channel_id,
		      int hop, int ttl)
    {
	super(peer,src,dest);
	this.channel_id = channel_id;

	this.prefix_only = false; //ChannelRequestMsg never do prefix only routing.

	this.hop = hop;
	this.TTL = ttl;
    }

//  Functions to get member variables:
    public int get_channel_id()                 { return channel_id; }
    public NodeId get_peer()                    { return peer; }
    public SecureHash get_src()                 { return src; }
    public SecureHash get_dest()                { return dest; }

//  Functions to set member variables:
    public void set_channel_id(int id)           { channel_id = id; }

//  Serialize functions:
    public void serialize (OutputBuffer buffer) 
    {
	System.out.println("peer = " + peer + "   src = " + src + "    dest = "+ dest);
	super.serialize (buffer);

	buffer.add (channel_id);
    }

//  Clone functions:
    public Object clone () throws CloneNotSupportedException 
    {
	ChannelRequestMsg result = (ChannelRequestMsg) super.clone ();
	
	result.channel_id = channel_id;

	return result;
    }

//  Clone functions:
    public String toString () 
    {
	return "(ChannelRequestMsg " + super.toString () + " " + src + "->" + dest + 
	    " hop " + hop + " TTL " + TTL;
    }
}













