package ostore.tapestry.channel;
import ostore.tapestry.api.TapestryRouteMsg;
import ostore.network.NetworkMessage;
import ostore.tapestry.impl.RouteMsg;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.CountBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSArray;
import ostore.util.QSByteArray;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.TypeTable;

public class ICMPPongMsg
  extends NetworkMessage implements QuickSerializable {
    public NodeId     nodeid;
    public SecureHash guid;   //guid of the node who sends this msg.
    public SecureHash msg_id;
    public int hop;           //Overlay hops of this node in th path.
    
    public ICMPPongMsg(NodeId peer, NodeId nodeid, SecureHash guid, 
		       SecureHash msg_id, int hop){
	super(peer, false /*outbound*/);
	this.nodeid = nodeid;
	this.guid = guid;
	this.msg_id = msg_id;
	this.hop  = hop;
    }
    
    public ICMPPongMsg( InputBuffer buffer ) throws QSException {
	super(buffer);
	nodeid = (NodeId) buffer.nextObject();
	guid = (SecureHash)buffer.nextObject();
	msg_id  = (SecureHash)buffer.nextObject();
	hop  = buffer.nextInt();
    }
    
    public void serialize (OutputBuffer buffer){
	super.serialize(buffer);
	buffer.add (nodeid);
	buffer.add (guid);
	buffer.add (msg_id);
	buffer.add(hop);
    }

    public String toString(){
	String _string = "ICMP-PONG-MSSG: at hop = " + hop +
	    " at which nodeid = " + peer + "; from nodeid = " + nodeid +
	    "; guid = " + guid + "; for msg_id = " + msg_id;
	return _string;
    }
}










