package ostore.tapestry.channel;

import seda.sandStorm.api.*;
import seda.sandStorm.core.ssTimer;
import seda.sandStorm.internal.SinkProxy;

import ostore.util.Carp;
import ostore.util.SHA1Hash;

import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.Pair;
import ostore.util.SecureHash;
import ostore.util.QuickSerializable;
import ostore.util.TypeTable;
import ostore.util.Types;
import ostore.dispatch.Classifier;
import ostore.dispatch.Signal;
import ostore.dispatch.Filter;
import ostore.tapestry.api.TapestryPublishMsg;
import ostore.tapestry.impl.TapestryReadyMsg;
import ostore.util.QSString;
import ostore.util.QSIO;
import ostore.security.QSPublicKey;
import ostore.tapestry.api.TapestryLocateFailure;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;
import seda.sandStorm.internal.SinkProxy;
import java.io.FileInputStream;

public class PongStage implements EventHandlerIF{
    
    private NodeId my_node_id;
    private NodeId dest_node_id;
    private Classifier _classifier;
    private Signal signal;
    private SinkIF _this_sink;
    private ssTimer _timer;
    private String stageName;
    private QSPublicKey _pkey;
    private SecureHash _my_id;
    
    public void handleEvent(QueueElementIF element)
	throws EventHandlerException{
	if (element instanceof StagesInitializedSignal){
	    
	}else if(element instanceof TapestryReadyMsg){
	    signal.alarm("pongStage", 1000);
	}else if(element instanceof Signal.AlarmSignal){
	    
	}else if(element instanceof ICMPPingMsg){
	    ICMPPingMsg msg = (ICMPPingMsg)element;
	    System.out.println("PongStage >> ICMPPingMsg.src: " + msg.src);//xxx
	    //System.out.println("ICMPPingMsg.dest: " + msg.dest);//xxx
	    System.out.println("PongStage >> ICMPPingMsg.peer: " + msg.peer);//xxx
	}
    }
    
    public void handleEvents(QueueElementIF elements[])
	throws EventHandlerException{
	for(int i = 0 ; i < elements.length; i++){
	    handleEvent(elements[i]);
	}
    }
    
    public void init(ConfigDataIF config) throws Exception{
	
	TypeTable.register_type("ostore.tapestry.channel.ICMPPingMsg");
	TypeTable.register_type("ostore.tapestry.channel.ICMPPongMsg");
	TypeTable.register_type("ostore.tapestry.channel.ICMPPingRouteMsg");
	Types.registerAllTypes();
	
	my_node_id = new NodeId(config.getString("node_id"));
	//dest_node_id = new NodeId(config.getString("dest_node_id"));
	_classifier = Classifier.getClassifier(my_node_id);
	signal = Signal.getSignal(my_node_id);
	
	String keyfilename = config.getString("pkey");
	try{
	    FileInputStream keyfile = new FileInputStream(keyfilename);
	    _pkey = (QSPublicKey) QSIO.read(keyfile);
	    keyfile.close();
	}catch(Exception e){
	    Carp.die ( "TClient.init: Caught exception " + e + 
		       "while trying to read pkey from disk.");
	}
	
	_my_id = new SHA1Hash(_pkey);
	String stageName = config.getStage().getName();
	try{
	    if( config.getBoolean("Profile") ){
		SinkIF sink = 
		    config.getManager().getStage(stageName).getSink();
		_this_sink = new SinkProxy(sink,
					   config.getManager(),
					   config.getStage().getWrapper());
		config.getManager().getProfiler().add("ostore.proxy.SDTcpClient", (SinkProxy)_this_sink );
	    }else{
		_this_sink =
		    config.getManager().getStage(stageName).getSink();
	    }
	}catch(NoSuchStageException e){
	    Carp.die(stageName + ".init:ascouldn ot get own sink.");
	}
	
	signal.register( Signal.ALRM, "pongStage", _this_sink  );
	registerInitialStageFilters();
    }
    
    private void registerInitialStageFilters() 
	throws Exception
    {
	String[] msg_types = {
	    "ostore.tapestry.impl.TapestryReadyMsg",
            "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.tapestry.channel.ICMPPingMsg"
	};
	
        for (int i = 0; i < msg_types.length; ++i)
        {
            Filter filter = new Filter();
            if (! filter.requireType(Class.forName(msg_types[i])))
            {
                Carp.die(stageName + ": Could not filter on type " +
                         msg_types[i]);
            }
            if (! _classifier.subscribe(filter, _this_sink))
            {
                Carp.die(stageName + ": Could not subscribe for type " +
                         msg_types[i]);
            }
        }
	
    }

    public void destroy() throws Exception{
	
    }
}
