/**
 * DynamicBeginDelete.java
 *
 * Message sent by a node to its backpointers to indicate its intention to 
 * detach from the Tapestry network.  It includes a list of replacement
 * candidates at each routing level, where possible.
 *
 * @author Ben Y. Zhao
 * @version $Id: DynamicBeginDelete.java,v 1.3 2003/03/28 00:57:00 strib Exp $
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import java.util.Iterator;
import java.util.HashMap;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.Carp;

public class DynamicBeginDelete
extends NetworkMessage implements QuickSerializable {

  public SecureHash src;  //the GUID of whoever sent this message
  public HashMap replacements; //single level (hop) of the routing table

  public DynamicBeginDelete (NodeId peer, SecureHash src,
			     HashMap replacements) {
    super (peer, false /*outbound*/);
    this.src = src;
    this.replacements = replacements;
  }


  public DynamicBeginDelete (InputBuffer buffer) throws QSException {
    super(buffer) ;
    src = (SecureHash) buffer.nextObject();
    SecureHash name;
    NodeId addr;
    int size = buffer.nextInt();
    replacements = new HashMap(size);
    for (int i =0 ; i < size; i++) {
      name = (SecureHash) buffer.nextObject();
      addr = (NodeId) buffer.nextObject();
      if (!name.isNull()) 
	//if null guid, means pair of guid/addr is useless
	replacements.put(new Integer(i), new GuidAddr(name, addr));
      else 
	replacements.put(new Integer(i), new GuidAddr(null, addr));
    }
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(src);
    buffer.add(replacements.size());
    
    SecureHash name;
    NodeId addr;
    GuidAddr ga;
    for (Iterator i=replacements.keySet().iterator(); i.hasNext(); ) {
      ga = (GuidAddr) replacements.get(i.next());
      name = ga.getGuid();
      if (name == null) 
	buffer.add(src.nullHash());
      else 
	buffer.add(name);
      addr = ga.getAddr();
      if (addr == null)
	Carp.die("Should never happen!!!!");
      buffer.add(addr);
    }
  }
  
  public Object clone () throws CloneNotSupportedException {
    DynamicBeginDelete result = (DynamicBeginDelete) super.clone();
    result.src = src;
    result.replacements = replacements;
    return result;
  }

  public String toString() {
    String s = "(DynamicBeginDelete:";
    s += " src=" + src;
    s += " peer=" + peer;
    s += "\n[";
    for (int i=0; i<replacements.size(); i++) {
      s += (GuidAddr) replacements.get(new Integer(i)) +", ";
    }
    return s;
  }

}
