/**
 * DynamicDMCastMsg.java
 *
 * The message that carries an acknowledged multicast to the "need to know"
 * nodes during a dynamic node insertion.
 *
 * @author Ben Y. Zhao
 * @version $Id: DynamicDMCastMsg.java,v 1.20 2003/03/05 19:41:36 strib Exp $
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.NonceMsg;
import ostore.util.SecureHash;
import ostore.util.TypeTable;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;

/**
 * Message used by a new node to do directed multicast to all nodes 
 * sharing a common suffix.
 */
public class DynamicDMCastMsg 
extends NonceMsg implements QuickSerializable {

  public SecureHash src; //root of mcast tree
  public NodeId srcAddr;
  public SecureHash newGuid;  //guid of new node
  public NodeId newAddr;      //nodeid of new node
  public WatchList watchlist;
  public int alpha;        //the length of the longest matching prefix
  public boolean fromLock;
  //to show if this msg came from a lock, therefore TTL=1
    
  public DynamicDMCastMsg (NodeId peer, SecureHash src, NodeId srcAddr,
			   SecureHash newGuid, NodeId newAddr, 
			   WatchList watchlist, int alpha, boolean lock) {
    super (peer);
    this.src = src;
    this.srcAddr = srcAddr;
    this.newGuid = newGuid;
    this.newAddr = newAddr;
    this.watchlist = watchlist;
    this.alpha = alpha;
    this.fromLock = lock;
  }


  public DynamicDMCastMsg (InputBuffer buffer) throws QSException {
    super(buffer);
    src = (SecureHash) buffer.nextObject();
    srcAddr = (NodeId) buffer.nextObject();
    newGuid = (SecureHash) buffer.nextObject();
    newAddr = (NodeId) buffer.nextObject();
    watchlist = (WatchList) buffer.nextObject();
    alpha = buffer.nextInt();
    fromLock = buffer.nextBoolean();
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(src);
    buffer.add(srcAddr);
    buffer.add(newGuid);
    buffer.add(newAddr);
    buffer.add(watchlist);
    buffer.add(alpha);
    buffer.add(fromLock);
  }

  public Object clone () throws CloneNotSupportedException {
    DynamicDMCastMsg result = (DynamicDMCastMsg) super.clone();
    result.src = src;
    result.srcAddr = srcAddr;
    result.newGuid = newGuid;
    result.newAddr = newAddr;
    result.watchlist = watchlist;
    result.alpha = alpha;
    result.fromLock = fromLock;
    return result;
  }

  public String toString() {
    return "(DynamicDMCastMsg: newGuid="+newGuid+", root="+src+
      (inbound?", coming from=":", going to=") +peer+", alpha="+ alpha+
      (fromLock?" from lock)":" )");
  }

}
