/**
 * DynamicObjReqMsg.java
 *
 * @author Ben Y. Zhao
 * @version $Id: DynamicObjReqMsg.java,v 1.16 2002/11/05 00:45:13 ravenben Exp $
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import ostore.util.TypeTable;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.OutputBuffer;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;

public class DynamicObjReqMsg 
extends NetworkMessage implements QuickSerializable {

  public SecureHash src;  //source of message
  public int msgtype;     //is this request(0) or confirmation(1)
  public SecureHash obj;  //object either requested or acked

  public static final int REQUEST = 0;
  public static final int CONFIRM = 1;

  public DynamicObjReqMsg (NodeId peer, SecureHash src,
			   SecureHash obj, int msgtype) {
    super (peer, false /*outbound*/);
    this.src = src;
    this.msgtype = msgtype;
    this.obj = obj;
  }

  public DynamicObjReqMsg (InputBuffer buffer) throws QSException {
    super(buffer);
    src = (SecureHash) buffer.nextObject();
    msgtype = buffer.nextInt();
    obj = (SecureHash) buffer.nextObject();
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(src);
    buffer.add(msgtype);
    buffer.add(obj);
  }

  public Object clone () throws CloneNotSupportedException {
    DynamicObjReqMsg result = (DynamicObjReqMsg) super.clone();
    result.src = src;
    result.msgtype = msgtype;
    result.obj = obj;
    return result;
  }

  public String toString() {
    if (msgtype == CONFIRM)
      return "(DynamicObjReqMsg: peer="+peer+", Confirm-->"+obj+")";
    else return "(DynamicObjReqMsg: peer="+peer+", Request-->"+obj+")";

  }
}
