/**
 * DynamicSeedMsg.java<BR>
 * A message sent by a node who's being inserted into the Tapestry
 * dynamically.  It sends this message to a set of nodes to request their
 * backpointers in order to construct the next set of nearest neighbors.
 *
 * @author Ben Y. Zhao
 * @version $Id: DynamicSeedMsg.java,v 1.3 2003/03/05 19:41:36 strib Exp $
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.NonceMsg;
import ostore.util.SecureHash;
import ostore.util.QSException;
import ostore.util.OutputBuffer;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;

public class DynamicSeedMsg 
  extends NonceMsg implements QuickSerializable {

  public SecureHash src;  //source of message
  public int level;

  public DynamicSeedMsg (NodeId peer, SecureHash src, int level) {
    super (peer);
    this.src = src;
    this.level = level;
  }

  public DynamicSeedMsg (InputBuffer buffer) throws QSException {
    super(buffer);
    src = (SecureHash) buffer.nextObject();
    level = buffer.nextInt();
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(src);
    buffer.add(level);
  }

  public Object clone () throws CloneNotSupportedException {
    DynamicSeedMsg result = (DynamicSeedMsg) super.clone();
    result.src = src;
    result.level = level;
    return result;
  }

  public String toString() {
    return "(DynamicSeedMsg: from-->"+src+", " + 
	(inbound?"from":"to") + "-->"+peer+", level="+level+")";
  }

}
