package ostore.tapestry.impl;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.SortedMap;
import java.util.StringTokenizer;
import java.util.Set;
import soss.network.TransitStub;

/**
 * @author     Ben Zhao
 */
public class Parser {

  public class Link {    //class to encapsulate topological link information
    public String from;  //unique IP address (after alias-resolution)
    public String to;    //unique IP address (after alias-resolution)

    public Link (String from, String to) {
      this.from = from;
      this.to = to;
    }
    public boolean equals(Object newlink) {
      Link nlink = (Link) newlink;
      /*      System.out.println("CALLED OFTEN, "+this.from+" "+nlink.from+
	      ", "+this.to+" "+nlink.to+", "+
	      (this.from.equals(nlink.from) &&
	      this.to.equals(nlink.to))); */
      return (this.from.equals(nlink.from) &&
	      this.to.equals(nlink.to));
    }
    public int hashCode() {
      return (this.from.hashCode() ^ this.to.hashCode());
    }
    public String toString() {
      return from+"->"+to;
    }
  }

  public class Path {  //class to encapsulate topological path information
                       //between edge (tapestry) nodes
    public String from;  //unique IP address (after alias-resolution)
    public String to;    //unique IP address (after alias-resolution)
    public HashSet links;

    public Path (String from, String to) {
      this.from = from;
      this.to = to;
    }
    public boolean equals(Object newpath) {
      Path npath = (Path) newpath;
      return (this.from.equals(npath.from) &&
	      this.to.equals(npath.to));
    }
    public int hashCode() {
      return (this.from.hashCode() ^ this.to.hashCode());
    }
    public String toString() {
      return from+"-"+to;
    }
  }

  //sequence of events
  //1. read in ally-planetlab
  //2. parse in data into router_ip_to_id and router_id_to_ips
  //3. now read in traceroute.dat
  //4. output new traceroute with unique router IDs
  //5. collect set of all links and map links to one way paths

  void parse_in_aliases(String fname) {
    
    BufferedReader br = null;
    try {
      br = new BufferedReader(new FileReader(fname));
    } catch (FileNotFoundException fnfe) {
      System.out.println("ERROR, file not found: "+fname);
      System.exit(-1);
    }
  
    aliased_ips = new HashMap(3700);
    unique_ips = new HashMap(200);
    StringTokenizer st = null;
    String currline;
    String nameIP, aliasIP, a, b;
    HashSet aliases;
    try {
      while ((currline = br.readLine()) != null) {
	a = b = nameIP = aliasIP = null;

	st = new StringTokenizer(currline, " ");
	a = st.nextToken();
	st.nextToken();
	b = st.nextToken();

	if (unique_ips.containsKey(a)) {
	  //if name is unique, we just add the alias to its list of aliases
	  //and then add alias to the aliasedIPs
	  nameIP = a;
	  aliasIP = b;
	} else if (unique_ips.containsKey(b)) {
	  nameIP = b;
	  aliasIP = a;
	} else if (aliased_ips.containsKey(a)) {
	  nameIP = (String) aliased_ips.get(a); //get real unique IP
	  aliasIP = b;
	} else if (aliased_ips.containsKey(b)) {
	  nameIP = (String) aliased_ips.get(b);  //get real unique IP
	  aliasIP = a;
	}
	//	System.out.println("nameIP:"+nameIP+", aliasIP:"+aliasIP);
	if (nameIP == null) {  //both are new
	  aliases = new HashSet();
	  unique_ips.put(a, aliases);
	  aliases.add(b);
	  aliased_ips.put(b, a);
	} else {   //do the normal thing
	  aliases = (HashSet) unique_ips.get(nameIP);
	  aliases.add(aliasIP);
	  aliased_ips.put(aliasIP, nameIP);

	  //now check if this alias link connected two unique IPs
	  if (unique_ips.containsKey(aliasIP)) {
	    HashSet newaliases = (HashSet) unique_ips.remove(aliasIP);
	    //iterate through those aliases and point them to nameIP
	    for (Iterator i = newaliases.iterator(); i.hasNext(); ){
	      aliasIP = (String) i.next();
	      aliased_ips.remove(aliasIP);
	      aliased_ips.put(aliasIP, nameIP);
	      aliases.add(aliasIP);
	    }
	  }
	}
      }
    } catch (Exception e) {
      System.out.println("Caught exception :"+e);
      e.printStackTrace();
      System.exit(-1);
    }
  }

  public void translate_traceroute(String fname) {

    BufferedReader br = null;
    try {
      br = new BufferedReader(new FileReader(fname));
    } catch (FileNotFoundException fnfe) {
      System.out.println("ERROR, file not found: "+fname);
      System.exit(-1);
    }

    try {
      br.readLine();  //ignore first line
    } catch (java.io.IOException jii) {
      System.out.println("ERROR: "+jii);
      System.exit(-1);
    }

    StringTokenizer st = null;
    String currline;
    ArrayList nodes = new ArrayList();
    String currnode=null, lastnode=null, destnode, a, b;
    Link currlink;
    Path currpath;
    HashSet paths;
    int index;
    
    if (links_to_paths == null) 
      links_to_paths = new HashMap(400);
    try {

      currline = br.readLine();  //read in list of node IP addresses
      st = new StringTokenizer (currline, " ");
      while (st.hasMoreTokens()) {
	currnode = st.nextToken();
	nodes.add(currnode);
      }

      while ((currline = br.readLine()) != null) {
	if (currline.charAt(0) == '#') {
	  //start of links for a new node
	  currnode = currline.substring(4, currline.length()-1);
	  //determine the source node for this set of links

	  lastnode = currnode;
	} else {
	  //first figure out destination node
	  index = currline.lastIndexOf('/');
	  //System.out.println("currline: "+currline+", "+
	  //		     (currline.lastIndexOf(' ',index)+1)+","+index);
	  destnode = currline.substring(currline.lastIndexOf(' ',index)+1,
					index);
	  //System.out.println("currline:"+currline+", "+
	  //  (currline.lastIndexOf(' ',index)+1)+", "+index);

	  currpath = new Path(currnode, destnode);
	  //now remember the current end to end path

	  //	  System.out.print(currpath+":");

	  st = new StringTokenizer(currline);
	  while (st.hasMoreTokens()) {
	    a = st.nextToken();
	    while (a.length() == 7)
	      a = st.nextToken();
	    //if the next token is just ***/***, get another token
	    b = a.substring(0, a.indexOf('/'));
	    a = b;
	    if (aliased_ips.containsKey(b)) {
	      b = (String) aliased_ips.get(b);
	      //System.out.println("Alias found: "+a+"=>"+b);
	    }
	    
	    currlink = new Link(lastnode, b);
	    if (links_to_paths.containsKey(currlink)) {
	      paths = (HashSet) links_to_paths.get(currlink);
	      //	      System.out.println("never being called");
	    } else {
	      paths = new HashSet();
	      links_to_paths.put(currlink, paths);
	    }
	    paths.add(currpath);
	    //	    System.out.print(currlink+", ");
	    lastnode = b;
	  } //end while (static.hasMoreTokens())

	  //	  System.out.println();
	}
      }
    } catch (Exception e) {
      System.out.println("Caught exception: "+e);
      e.printStackTrace();
    }
  }

  public void output_links() {

    System.out.println("Data file: Mapping IP paths to unaliased links");
    System.out.println("Format:  linkFrom:linkTo; pathFrom-pathTo pathFrom-pathTo ...");
    Link l;
    Path p;
    HashSet hs;
    for(Iterator i = links_to_paths.keySet().iterator(); i.hasNext(); ) {
      l = (Link) i.next();
      System.out.print(l+";");
      hs = (HashSet) links_to_paths.get(l);
      for (Iterator it = hs.iterator(); it.hasNext();) {
	p = (Path) it.next();
	System.out.print(" "+p);
      }
      System.out.println();
    }
  }

  public void input_links() {
    BufferedReader br = null;
    try {
      br = new BufferedReader(new FileReader(input_filename));
    } catch (FileNotFoundException fnfe) {
      System.out.println("ERROR, file not found: "+input_filename);
      System.exit(-1);
    }

    try {
      br.readLine();  //ignore first line
      br.readLine();  //ignore second line
    } catch (java.io.IOException jii) {
      System.out.println("ERROR: "+jii);
      System.exit(-1);
    }

    StringTokenizer st = null;
    String currline;
    String currnode=null, lastnode=null, destnode, a, b;
    Link currlink;
    Path currpath;
    HashSet paths;

    if (links_to_paths == null) 
      links_to_paths = new HashMap(400);
    try {

      while ((currline = br.readLine()) !=null) {
	a = currline.substring(0, currline.indexOf('-'));
	b = currline.substring(currline.indexOf('>')+1,
			       currline.indexOf(';'));
	currlink = new Link(a, b);
	paths = new HashSet();
	links_to_paths.put(currlink, paths);
	//	System.out.print("Link:"+currlink+":");
	
	st = new StringTokenizer (currline, " ");
	st.nextToken();    //ignore first token (link representation)
	while (st.hasMoreTokens()) {
	  a = st.nextToken();  //get next path in list
	  b = a.substring(0, a.indexOf('-'));
	  currpath = new Path(b, a.substring(a.indexOf('-')+1, a.length()));
	  paths.add(currpath);
	  //	  System.out.print(" "+currpath);
	}
	//	System.out.println();
      }
    } catch (Exception e) {
      System.out.println("Caught exception: "+e);
      e.printStackTrace();
      System.exit(-1);
    }
    
  }

  public void translate_topology(int ports[], int num_ports) 
  {
      TransitStub _ts;
      Integer head, tail;
      HashSet paths;
      Path    curr_path;
      Link    curr_link;
      int i, j;

      links_to_paths = new HashMap(400);

      try {
	  _ts = new TransitStub(topology_file);
      }
      catch (IOException e) {
	  System.out.println("Caught exception when " +
			     "creating graph: "+e);
	  return;
      }

      for (i = 0; i < num_ports; i++)
      {
	  for (j = 0; j < num_ports; j++)
	  {
	      if (i == j)
		  continue;

	      curr_path = new Path( Integer.toString(ports[i]),
				   Integer.toString(ports[j]) );

	      TransitStub.PathInfo path_info = 
		  _ts.shortest_path(ports[i], ports[j], 0, false, 0);
	      if (path_info == null)
		  System.out.println("Horrible! No PathInfo between " + 
				     ports[i] + " and " + ports[j]);
	      else
	      {
		  /*
		  System.out.println("Shortest path information of (" +
				     ports[i] + "," + ports[j] + 
				     "): total_hops = " + 
				     path_info.total_hops +
				     "; path = " + 
				     path_info.path);
		  */
		  if (path_info.path == null)
		  {
		      System.out.println("Horrible! No path between "
					 + ports[i] + " and " + ports[j]);
		  }
		  else
		  {
		      Iterator it = path_info.path.iterator ();
		      head = (Integer) it.next();
		      while ( it.hasNext () ) 
		      {
			  tail = (Integer)it.next();
			  //System.out.println("curr_link = [" + head +
			  //	     "," + tail + "]");
			  if ( head.intValue() != tail.intValue() )
			  {
			      curr_link = new Link(head.toString(), 
						   tail.toString());
			      paths = (HashSet)links_to_paths.get(curr_link);
			      if (paths != null)
				  paths.add(curr_path);
			      else
			      {
				  paths = new HashSet();
				  links_to_paths.put(curr_link, paths);
				  paths.add(curr_path);
			      }
			  }

  			  head = tail;
		      }
		  }
	      }
	  }
      }
      links = (Link [])links_to_paths.keySet().toArray(new Link[links_to_paths.size()]);
  }

  public int getTotalNumberofLinks() {
    return links_to_paths.size();
  }

  public void addFailedLinks(int number, Integer unique_id)
    throws Exception {  //throws exception if unique_id already used once
    int counter = number;
    if (id_to_links.containsKey(unique_id))
      throw new Exception ("Unique ID: "+unique_id+" already used!!");

    Link l;

    System.out.println("Adding "+number+" failed links: ");
    HashSet picked = new HashSet(number);
    int index;
    while (true) {
      index = (int) (java.lang.Math.random() * links.length);

      if (links_to_status.get(links[index]) == null &&
	  !picked.contains(links[index])) {
	picked.add(links[index]);
	links_to_status.put(links[index], unique_id);
	counter --;
	if (counter == 0) break;
      }
    }
    id_to_links.put(unique_id, picked);
  }

  public HashSet getFailedPath(Integer unique_id) 
    throws Exception {  //throws exception if unique_id has not been used
    HashSet links;
    links = (HashSet) id_to_links.get(unique_id);
    System.out.println("***Getting failedlinks for id:"+unique_id.hashCode()+ 
		       ", "+ links);
    System.out.println(" keyset="+id_to_links.keySet());
    
    if (links == null) 
      throw new Exception("Unique ID: "+unique_id.hashCode()+" not found!!");

    HashSet totalpaths = new HashSet();
    Link l;
    HashSet paths;
    for (Iterator i = links.iterator(); i.hasNext(); ) {
    //    for (Iterator i = links_to_paths.keySet().iterator(); i.hasNext(); ) {
      l = (Link) i.next();
      paths = (HashSet) links_to_paths.get(l);
      //      System.out.println("Paths = "+paths);
      if (paths != null)
	  totalpaths.addAll(paths);
    }
    return totalpaths;
  }

  public void print_aliases() {
    System.out.println("Unique Nodes");
    String a,b;
    HashSet hs;
    for (Iterator i = unique_ips.keySet().iterator(); i.hasNext(); ) {
      a = (String) i.next();
      hs = (HashSet) unique_ips.get(a);
      System.out.print(a+":"+hs+"\n");
    }
  }

  public static void main (String[] args) {

    Parser parser = new Parser();

    //   parser.parse_in_aliases (alias_filename);
    //parse in the aliases, then fill int router_ip_to_id and router_id_to_ips

    //parser.print_aliases ();

    //  parser.translate_traceroute(tracer_filename);
    //read in traceroute data, write out new traceroutes with unique IDs
    //parse in all links, map hashmap from unique one-way links to 
    //unidirectional paths

    //  parser.output_links();
    //output resulting info to file

        parser.links_to_paths = null;
        parser.input_links();
	//        parser.translate_topology();
    //read back in the output from parser.output_links()
    
  }

  private static final boolean DEBUG = false;
  private static final String alias_filename = "ally-planetlab";
  private static final String tracer_filename = "traceroute.dat";
  private static final String input_filename = "links-paths.dat";
  private static final String topology_file = 
      "/home/hling/oceanstore/pond/run/graphs/ts5000-1.hier";

  public Parser () {
    links_to_paths = new HashMap(500);
    links_to_status = new HashMap(500);
    id_to_links = new HashMap();
  }

  Link [] links;
  HashMap links_to_status;   
  //mapping from links to the ID that they failed under. null means they 
  //did not fail
  HashMap id_to_links;
  //mapping from failure ID to all the links that failed with it

  public HashMap links_to_paths;  
                           //mapping from Links to all Paths that traverse it

  HashMap router_ip_to_id;  //hashes router IP to unique router ID
  HashMap router_id_to_ips; //hashes unique router ID to list of router IPs
  HashMap aliased_ips; //mapping from aliased IP to the real unique IP
  HashMap unique_ips;  //mapping from unique Ip to list of aliases

}

