package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import ostore.util.QSException;
import ostore.util.Carp;
import java.util.Iterator;
import java.util.HashMap;
import java.util.HashSet;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.Pair;
import ostore.util.QuickSerializable;

public class PeriodicRoutePatch
  extends NetworkMessage implements QuickSerializable {

  public static final int REQUEST = 1;
  public static final int RESPONSE = 2;

  public int type;  //request or response
  public SecureHash srcguid;
  public HashSet set;

  public PeriodicRoutePatch (int type, NodeId peer, 
			     SecureHash srcguid, HashSet set) {
    super (peer, false /*outbound*/);
    this.srcguid = srcguid;
    this.type = type;
    this.set = set;
  }

  public PeriodicRoutePatch (InputBuffer buffer) throws QSException {
    super(buffer) ;
    int size;
    type = buffer.nextInt();
    srcguid = (SecureHash) buffer.nextObject();
    size = buffer.nextInt();
    set = new HashSet(size);
    if (type == REQUEST) {
      for (int i=0; i<size; i++) {
	set.add(new Pair(new Integer(buffer.nextInt()),
			 new Integer(buffer.nextInt())));
      }
    } else if (type == RESPONSE) {
      for (int i=0; i<size; i++) {
	set.add((GuidAddr) buffer.nextObject());
      }
    } else 
      Carp.die("PeriodicRoutePatch: Unrecognized type:"+this);
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize(buffer);
    buffer.add(type);
    buffer.add(srcguid);
    buffer.add(set.size());
    if (type == REQUEST) {
      Pair p;
      for (Iterator i = set.iterator(); i.hasNext(); ){
	p = (Pair) i.next();
	buffer.add(((Integer) p.first()).intValue());
	buffer.add(((Integer) p.second()).intValue());
      }
    } else if (type == RESPONSE) {
      for (Iterator i = set.iterator(); i.hasNext(); ) {
	buffer.add((GuidAddr) i.next());
      }
    } else {
      System.err.println("Serialization error in PeriodicRoutePatch");
    }
  }
  
  public Object clone () throws CloneNotSupportedException {
    PeriodicRoutePatch result = (PeriodicRoutePatch) super.clone();
    result.peer = peer;
    result.type = type;
    result.srcguid = srcguid;
    result.set = set;
    return result;
  }

  public String toString() {
    String s = "(PeriodicRoutePatch: from " + peer;
    return s;
  }
}

