/**
 * RepublishRequestMsg.java
 *
 * @author     Ben Y. Zhao
 * @version    $Id: RepublishRequestMsg.java,v 1.1 2002/09/05 23:26:56 ravenben Exp $
 */

package ostore.tapestry.impl;
import ostore.util.QSException;
import ostore.util.SecureHash;
import ostore.util.OutputBuffer;
import ostore.tapestry.api.TapestryRouteMsg;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;

/**
 * The message sent by DynamicTClient (or another Tapestry stage) to the
 * source of a particular object replica requesting that it republish the
 * object. This is generally due to changes in the Tapestry mesh which
 * require the publication information to be propagated to a new publication
 * path.
 */
public class RepublishRequestMsg 
  extends TapestryRouteMsg implements QuickSerializable {

  public SecureHash objGuid;
  public SecureHash requestor;

  public RepublishRequestMsg (SecureHash objGuid, SecureHash dest_guid,
			      SecureHash requestor) {
    super (dest_guid);
    this.objGuid = objGuid;
    this.requestor = requestor;
  }

  public RepublishRequestMsg (InputBuffer buffer) throws QSException {
    super(buffer);
    objGuid = (SecureHash) buffer.nextObject();
    requestor = (SecureHash) buffer.nextObject();
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(objGuid);
    buffer.add(requestor);
  }

  public String toString() {
    return "(RepublishRequestMsg for object "+objGuid+", from="+requestor+")";
  }
}


