/**
 * RouteMsg.
 *
 * @author     Sean C. Rhea
 * @version    $Id: RouteMsg.java,v 1.33 2002/08/18 22:39:41 srhea Exp $
 */

package ostore.tapestry.impl;
import ostore.tapestry.channel.MsgQos;
import ostore.network.NetworkMessage;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.CountBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSArray;
import ostore.util.QSByteArray;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import ostore.util.TypeTable;

public class RouteMsg 
extends TapestryMessage implements QuickSerializable {

    public SecureHash src;
    public SecureHash dest;
    public int hop;
    public int TTL;
    public QSByteArray mac;
    public boolean prefix_only;

    //Modified by Ling Huang. For inheriting ChannelMsg from RouteMsg.
    //    private QuickSerializable _user_data;
    //    private int _user_data_length;
    //    private InputBuffer _user_data_buffer;

    protected QuickSerializable _user_data;
    protected int _user_data_length;
    protected InputBuffer _user_data_buffer;

  /* They are moved to TapestryMessage
    //Added by Ling Huang for reliable msg delivery.
    protected int qos;                //The QOS of this message, should it be reliable or un-.
    protected SecureHash msg_id;      //Unique identity for a message for retransmission.
  */

    public QuickSerializable user_data () throws QSException {
	if (_user_data == null) {
	    _user_data = _user_data_buffer.nextObject ();
	    _user_data_buffer = null; // save memory
	}
	return _user_data;
    }

    public RouteMsg()
    {
	super (null, false /* outbound */);

	this.src = null;
	this.dest = null;

	this.prefix_only = false;
	this.hop = 0;
	this.TTL = -1;
	this._user_data = null;
	this.mac = null;

	this.qos = MsgQos.UNRELIABLE;
	this.msg_id = null;
    }

    public RouteMsg(NodeId peer,SecureHash src, SecureHash dest)
    {
	super (peer, false /* outbound */);

	this.src = src;
	this.dest = dest;

	this.prefix_only = false;
	this.hop = 0;
	this.TTL = -1;
	this._user_data = null;
	this.mac = null;

	this.qos = MsgQos.UNRELIABLE;
	this.msg_id = null;
    }

    public RouteMsg (NodeId peer, SecureHash src, SecureHash dest,
		     boolean prefix_only, int hop, 
		     QuickSerializable user_data) {
	super (peer, false /* outbound */);
	this.src = src;
	this.dest = dest;
	this.prefix_only = prefix_only;
	this.hop = hop;
	this.TTL = -1;
	this._user_data = user_data; 
	this.mac = null;

	this.qos = MsgQos.UNRELIABLE;
	this.msg_id = null;
    }

    public RouteMsg (NodeId peer, SecureHash src, SecureHash dest,
		     boolean prefix_only, int hop, 
		     QuickSerializable user_data, 
		     int TTL) {
	super (peer, false /* outbound */);
	this.src = src;
	this.dest = dest;
	this.prefix_only = prefix_only;
	this.hop = hop;
	this.TTL = TTL;
	this._user_data = user_data;
	this.mac = null;

	this.qos = MsgQos.UNRELIABLE;
	this.msg_id = null;
    }

    public RouteMsg (InputBuffer buffer) throws QSException {
	super (buffer);
	src = (SecureHash) buffer.nextObject ();
	dest = (SecureHash) buffer.nextObject ();
	prefix_only = buffer.nextBoolean ();
	hop = buffer.nextInt ();
	TTL = buffer.nextInt ();
	_user_data_length = buffer.nextInt ();
	_user_data_buffer = buffer.subBuffer (_user_data_length);
	if (buffer.nextBoolean ())
	    mac = (QSByteArray) buffer.nextObject ();

/*  This is moved to TapestryMessage by Ling Huang 07/31/02
	qos = buffer.nextInt ();
	if (qos == MsgQos.RELIABLE)
	    msg_id = (SecureHash) buffer.nextObject ();
*/
    }

    public QuickSerializable data_to_mac () throws QSException {
	user_data ();
	QuickSerializable [] a = {src, dest, _user_data};
	return new QSArray (a);
    }

    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);
	buffer.add (src);
	buffer.add (dest);
	buffer.add (prefix_only);
	buffer.add (hop);
	buffer.add (TTL);
	if (_user_data != null) {
	    // TODO: find a better way to get size of message.
	    CountBuffer cb = new CountBuffer ();
	    cb.add (_user_data);
	    buffer.add (cb.size ());
	    buffer.add (_user_data);
	}
	else {
	    buffer.add (_user_data_length);
	    try {
		// In order to make the add_to_digest idempotent, I have to
		// clone the _user_data_buffer before reading the data out
		// of it.  Otherwise, it will be empty the next time this
		// function is called.
		buffer.add ((InputBuffer) _user_data_buffer.clone ());
	    }
	    catch (CloneNotSupportedException e) {
		throw new AssertionViolatedException ("must be cloneable");
	    }
	}
	if (mac == null)
	    buffer.add (false);
	else {
	    buffer.add (true);
	    buffer.add (mac);
	}

/*  This is moved to TapestryMessage by Ling Huang 07/31/02
	buffer.add (qos);
	if (qos == MsgQos.RELIABLE)
	    buffer.add (msg_id);
*/
    }

    public Object clone () throws CloneNotSupportedException {
	RouteMsg result = (RouteMsg) super.clone ();
	result.src = src;
	result.dest = dest;
	result.prefix_only = prefix_only;
	result.hop = hop;
	result.TTL = TTL;
	result._user_data = _user_data;
	result._user_data_length = _user_data_length;
	if (_user_data_buffer != null)
	    result._user_data_buffer = (InputBuffer) _user_data_buffer.clone ();
	result.mac = mac;

	//This is moved to TapestryMessage by Ling Huang 07/31/02.
	//result.qos = qos;
	//result.msg_id = msg_id;

	return result;
    }

/*  This is moved to TapestryMessage by Ling Huang 07/31/02
    // Added by Ling Huang for msg retransmission.
    public int get_qos()                    {   return qos; }
    public boolean is_reliable()
    {
	if (qos == MsgQos.RELIABLE)
	    return true;
	return false;
    }
    public SecureHash get_msg_id()          {   return msg_id; }
    public void set_msg_id(SecureHash id)   {   msg_id = id; }
*/

    public void set_qos(int qos)
    {   
	this.qos = qos;
	//	if (qos == MsgQos.RELIABLE && msg_id == null)
	if (msg_id == null)
	    generate_msg_id();
    }
    public SecureHash generate_msg_id()
    {
	double  id;
	String  id_string;

	id = java.lang.Math.random();
	id_string = src.toString() + dest.toString() + Double.toString(id);
	msg_id = new SHA1Hash(id_string);

	return msg_id;
    }
    public SecureHash generate_msg_id(Object id)
    {
	String  id_string;

	id_string = src.toString() + dest.toString() + id.toString();
	msg_id = new SHA1Hash(id_string);

	return msg_id;
    }

    public String toString () {
	return "(RouteMsg " + super.toString () + " " + src + "->" + dest + 
	    " hop " + hop + " TTL " + TTL +
	    (mac == null ? " without" : " with") + " mac " + _user_data + ")";
    }
}















