/**  Router.java
 *
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.ListIterator;
import java.util.Date;
import java.util.Set;
import java.io.FileInputStream;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.StagesInitializedSignal;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.ProfilableIF;
import seda.sandStorm.internal.SinkProxy;
import ostore.bloom.BloomQuery;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.mac.MacResponse;
import ostore.mac.MacRequest;
import ostore.mac.HMACDHHandler;
import ostore.mac.MacHandler;
import ostore.network.NetworkMessage;
import ostore.security.QSPublicKey;
import ostore.tapestry.channel.ChannelACKMsg;
import ostore.tapestry.channel.RetransmissionCheckMsg;
import ostore.tapestry.api.TapestryDetachConfirm;
import ostore.tapestry.api.TapestryIntegrateMsg;
import ostore.tapestry.api.TapestryLocateFailure;
import ostore.tapestry.api.TapestryLocateMsg;
import ostore.tapestry.api.TapestryMacRouteMsg;
import ostore.tapestry.api.TapestryPrefixRouteMsg;
import ostore.tapestry.api.TapestryPublishMsg;
import ostore.tapestry.api.TapestryObjPointerSetReq;
import ostore.tapestry.api.TapestryObjPointerSetResp;
import ostore.tapestry.api.TapestryQuery;
import ostore.tapestry.api.TapestryQueryResult;
import ostore.tapestry.api.TapestryQueryResultMatch;
import ostore.tapestry.api.TapestryQueryState;
import ostore.tapestry.api.TapestryRouteMsg;
import ostore.tapestry.api.TapestryTag;
import ostore.tapestry.api.TapestryUnpublishMsg;
import ostore.tapestry.channel.ICMPPingMsg;
import ostore.tapestry.channel.ICMPPongMsg;
import ostore.tapestry.channel.ICMPPingRouteMsg;
import ostore.network.patchwork.PatchworkTable;
import ostore.network.patchwork.DeltaMsg;
import ostore.network.patchwork.QuickStatReq;
import ostore.tapestry.patchwork.RouterFaultMsg;
import ostore.tapestry.patchwork.FaultController;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.Clock;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import ostore.util.QSByteArray;
import ostore.util.QSException;
import ostore.util.QSIO;
import ostore.util.QuickSerializable;
import ostore.util.SHA1Hash;
import ostore.util.SecureHash;
import ostore.util.TypeTable;

/**
 * Handles the routing functions of Tapestry.
 *
 * TODO:  The back pointers are currently maintained in sorted linked lists
 * which are sorted by distance.  All the elements in a list have the same
 * guid, and the lists are stored in a map indexed by those guids.  I don't
 * have to tell you that this is inefficient, but I can't get Java's
 * TreeSet class to replace the lists.  I think it has something to do with
 * what the Javadocs call compare being "compatible with" equals.  I don't
 * know.  In any case, the backpointers will need to be stored on disk at
 * some point, not all in memory, so this will all need an overhauling
 * then.  --srhea, 12/18/2001
 *
 * @author     Sean C. Rhea and Ben Y. Zhao
 * @version    $Id: Router.java,v 1.144 2003/06/20 23:04:49 strib Exp $
 */
public class Router implements SingleThreadedEventHandlerIF, EventHandlerIF, ProfilableIF {

  private String tag = "Router";
    private static final boolean DEBUG = false;
  //    private static final boolean DEBUG = true;
  private static final boolean MAC_DEBUG = false;
  private boolean DO_MACS = true;
  public static final int LOCALROOTOBJS = -1;
  
  /**
   ** The constant that defines whether multicast is turned on or not.
   ** This value has no effect, and the run time value is determined by
   ** config scripts. If it is not found in the config files, it's set to
   ** false by default.  Config name = ft_multicast
   **/
  public static boolean MULTICAST = false;
  
  /**
   ** The constant that defines when multicast is triggered for a message.
   ** If the loss rate is greater than THRESHOLD_MULTICAST*100, then
   ** the message is multicast across the links.  
   ** Config name = multicast_threshold
   **/
  public static int THRESHOLD_MULTICAST = 20;

  /**
   ** The size of the multicast convergence queue, which we use to
   ** drop duplicate messages which are produced by the multicast routing
   ** mechanism.  Config name = multicast_queue_size
   **/  
  public static int MULTICAST_QUEUE_SIZE = 5;
  
  /**
   ** Boolean flag that turns FRLS mode on or off.  This is set here by
   ** default, and can only be turned off by specifying the config
   ** argument "disable_frls" to true. Config name = disable_frls
   **/
  public static boolean FRLS = true;
    //  public static boolean FRLS = false;

  /**
   ** The maximum failure rate tolerable that defines "reachable" in our
   ** FRLS scheme.  FRLS is determined by scanning according to link
   ** latency and picking the first node with loss rate less than
   ** THRESHOLD_FRLS. Config name = frls_threshold
   **/
  public static int THRESHOLD_FRLS = 20;
  
  /**
   ** Threshold for determining when a node is "dead."  If a node's
   ** updated link loss rate is rated above THRESHOLD_DEAD for this
   ** current period, then it is "dead." 
   ** Config name = inactive_node_threshold
   **/
  public static int THRESHOLD_DEAD = 60;
  
  /**
   ** If a node has been "dead/inactive" for a consecutive number of periods
   ** equal or greater than THRESHOLD_REPLACE, it is removed from
   ** the routing table and we try to replace it using repair algorithms.
   ** Config name = replace_node_threshold
   **/
  public static int THRESHOLD_REPLACE = 4;
  
  
  /*
    public static final byte[] _match_bytes = {(byte)0x07, (byte)0xaa, 
    (byte)0xa4, (byte)0xde, 
    (byte)0x6a, (byte)0x6d};
  */
  public static final byte[] _match_bytes = {(byte)0x24, (byte)0x81, 
					     (byte)0xc3, (byte)0xbf, 
					     (byte)0x5d, (byte)0xfe};
  static {
    Debug.printtagln("Router.<init>",
		     " match_bytes=(SHA1:" + 
		     ByteUtils.print_bytes(_match_bytes, 0,
					   _match_bytes.length) +")");
  }

    /**
     * The following flags represent flags for local optimizations
     * during publishing.  Which optimizations to use are indicated by
     * the local_opts field in TapestryPublishMsg.  The supported 
     * operations are described in TapestryPublishMsg.
     */
    public static final int LOCAL_OPT_PUBLISH_BACKUP = 1;
    public static final int LOCAL_OPT_PUBLISH_ALL_LINKS = 2;
    public static final int LOCAL_OPT_PUBLISH_LOCAL_MISROUTE = 4;

    /**
     * The default number of backups to publish to while optimizing
     */
    public static int LOCAL_OPT_BACKUPS = RouteEntry.REDUNDANT_LVL-1;

    /**
     * The default number of links on the same level to publish to while
     * optimizing
     */
    public static int LOCAL_OPT_NEAREST = 10;

    /**
     * The factor by which to judge whether or not a link is too long
     * to take (i.e. if it is this much bigger than the last link taken)
     */
    public static int LOCAL_OPT_THRESHOLD_FACTOR = 10;

    /**
     * The default initial time-to-live for the local optimizations that only
     * get forwarded for a few hops.
     */
    public static int LOCAL_OPT_HOPS = 2;

    private void dispatch (QueueElementIF item) {
	if (DEBUG) Debug.printtagln (tag, "dispatching " +
		item.getClass () + " event: " + item + ".");

	//Modified by Ling Huang: add channel in Tapestry router.
	//Channel will handle all the dispatch(may be in or not in channel).) 
	/*
	try 
	{
	    Classifier.getClassifier(_self_node_id).dispatch(item);
	} 
	catch (SinkException e) 
	{
	    BUG (tag + ".dispatch: could not dispatch " + item);
	}
	*/

	//Every time when a multicast msg comes to a router, the router do a
	//look up in the queue. If the msg_id is in the queue, this means the
	// msg came to the router before. It is duplicate msg and should 
	//not be dispatched. This is checked and done in handleEvent().

	//Otherwise, it is the first time appearing in the router.
	//Its msg_id is added into the multicast queue.
	if (MULTICAST == true)
	{
	    if (item instanceof TapestryMessage)
	    {
		TapestryMessage msg = (TapestryMessage) item;
		if (msg.get_delivery_type() == msg.MULTICAST)
		{
		    MsgTime rmvd, addmsg;
		    SecureHash msg_id = msg.get_msg_id();
		    long time = System.currentTimeMillis();
		    addmsg = new MsgTime(msg_id, time);
		    _multicast_msgs.add(msg_id);
		    _sorted_multicast_msgs.add(addmsg);
		    if (_sorted_multicast_msgs.size() > MULTICAST_QUEUE_SIZE)
		    {
			rmvd = (MsgTime)_sorted_multicast_msgs.first();
			_sorted_multicast_msgs.remove(rmvd);
			_multicast_msgs.remove(rmvd.msg_id);
			if (DEBUG) 
			    Debug.printtagln (tag, "Two much msgs in multicast queue." +
			      "Throw out " + rmvd.msg_id + "enqueued at time = " + 
			      rmvd.time + " .");
		    }
		    if (DEBUG) Debug.printtagln (tag, "Unicast Dispatching: " +
			       "get a multicast msg and put it in the queue!");

		    /*
		    //Multicast msg will be removed after 2 seconds. 
		    //The assumption is that msgs will get to destination
		    // in 2 seconds.
		    long later_time = System.currentTimeMillis() + 20000;
		    Classifier.getClassifier(_self_node_id).dispatch_later(
			       new MulticastMsgDone(msg_id), later_time);
		    */

		}
	    }
	}

	_channel_handler.dispatch(item);
    }

    public boolean isDuplicate(QueueElementIF item)
    {
	//Added by Ling Huang for fault-tolerance routing.
	//Every time when a multicast msg comes to a router, the router do a look 
	//up in the multicast msg queue. If the msg_id is in the queue, this means 
	//the msg came to the router before. 
	//It is duplicate msg and should not be dispatched. 
	if (item instanceof TapestryMessage)
	{
	    TapestryMessage msg = (TapestryMessage) item;
	    SecureHash msg_id = msg.get_msg_id();
	    if (msg.get_delivery_type() == msg.MULTICAST)
	    {
		if ( _multicast_msgs.contains(msg_id) )
		{
		    if (DEBUG) Debug.printtagln (tag, "Duplicate: " +
			       "TapMsg " + msg_id + " came to the router before!");

		    return true;
		}
		else
		{
		    MsgTime rmvd, addmsg;
		    long time = System.currentTimeMillis();
		    addmsg = new MsgTime(msg_id, time);
		    _multicast_msgs.add(msg_id);
		    _sorted_multicast_msgs.add(addmsg);
		    if (_sorted_multicast_msgs.size() > MULTICAST_QUEUE_SIZE)
		    {
			rmvd = (MsgTime)_sorted_multicast_msgs.first();
			_sorted_multicast_msgs.remove(rmvd);
			_multicast_msgs.remove(rmvd.msg_id);
			if (DEBUG) 
			    Debug.printtagln (tag, "Two much msgs in multicast queue." +
			      "Throw out " + rmvd.msg_id + "enqueued at time = " + 
			      rmvd.time + " .");
		    }
		    /*
		    //Multicast msg will be removed after 2 seconds. 
		    //The assumption is that
		    //the msg will get to destination in 2 seconds.
		    long later_time = System.currentTimeMillis() + 20000;
		    Classifier.getClassifier(_self_node_id).dispatch_later(
			       new MulticastMsgDone(msg_id), later_time);
		    */
		}
	    }
	    if (DEBUG) Debug.printtagln (tag, "Duplicate: " + "TapMsg " + 
		       msg_id + " first time comes to the router, handle it!");

	}

	return false;
    }

    private void dispatch_multicast(TapestryMessage outb, RouteEntry entry)
    {
	TapestryMessage[] backup_outb;
	SecureHash msg_id;

	msg_id = outb.get_msg_id();
	//Every time when a multicast msg comes to a router, the router do a
	//look up in the queue. If the msg_id is in the queue, this means the
	// msg came to the router before. It is duplicate msg and should 
	//not be dispatched. 
	//This is checked and done in handleEvent().
	//Otherwise, it is the first time appearing in the router.
	//Its msg_id is added into the multicast queue.
	MsgTime rmvd, addmsg;
	long time = System.currentTimeMillis();
	addmsg = new MsgTime(msg_id, time);
	_multicast_msgs.add(msg_id);
	//There is a fixed size queue for multicast msg.
	//It uses FIFO to manange all msgs.
	_sorted_multicast_msgs.add(addmsg);
	if (_sorted_multicast_msgs.size() > MULTICAST_QUEUE_SIZE)
	{
	    rmvd = (MsgTime)_sorted_multicast_msgs.first();
	    _sorted_multicast_msgs.remove(rmvd);
	    _multicast_msgs.remove(rmvd.msg_id);
	    if (DEBUG) 
		Debug.printtagln (tag, "Two much msgs in multicast queue." +
		      "Throw out " + rmvd.msg_id + "enqueued at time = " + 
		      rmvd.time + " .");
	}

	if (DEBUG) Debug.printtagln (tag, "Multicast Dispatching: " +
		   "get a msg with id " + msg_id + " and put it in the queue!");

	try 
	{
	    outb.set_delivery_type(TapestryMessage.MULTICAST);
	    NodeId  first  = entry.getFirst();
	    if (first == null)
	    {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
		    //MARK
		    System.out.println("Doing Multicast for bad message: "+outb);
		    System.out.println(_routing_tbl);
		    Carp.out (tag + ": " + 
			      "got TapestryRouteMsg for non-existent " +
			      "destination.  Dropping it.");
		}
		return;
	    }

	    if (DEBUG) Debug.printtagln (tag, "Multicast dispatching for " +
		       outb.get_msg_id() + " to dest: " + ((RouteMsg)outb).dest + 
		       " with entries: " + first + " : " + entry.getAtIndex(1) + 
					 " : " + entry.getAtIndex(2));
	    //Dispatch the msg from primary path.
	    outb.peer  = first;
	    try 
	    {
		Classifier.getClassifier(_self_node_id).dispatch(outb);
	    } 
	    catch (SinkException e) 
	    {
		BUG (tag + ".dispatch: could not dispatch " + outb);
	    }
	    //Dispatch the msg from backup paths.
	    int k;
	    NodeId backup_routes;
	    backup_outb = new TapestryMessage[RouteEntry.REDUNDANT_LVL];
	    for (k = 2; k < RouteEntry.REDUNDANT_LVL; k++)
	    {
		backup_routes = entry.getAtIndex(k);
		if (backup_routes != null)
		{
		    backup_outb[k] = (TapestryMessage) outb.clone ();
		    backup_outb[k].set_delivery_type(TapestryMessage.MULTICAST);
		    backup_outb[k].peer = backup_routes;
		    try 
		    {
			Classifier.getClassifier(_self_node_id).dispatch(backup_outb[k]);
		    } 
		    catch (SinkException e) 
		    {
			    BUG (tag + ".dispatch: could not dispatch " + backup_outb[k]);
		    }
		}
	    }
	}
	catch (CloneNotSupportedException e) {
	    BUG (tag + ": wtf?");
	}
    }

    public static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in " +
		"ostore.tapestry.impl.Router.  " +
		"Please let Sean know about it ASAP.");
    }

    public void init (ConfigDataIF config) throws Exception {
	_self_node_id = new NodeId (config.getString ("node_id"));

	if( config.contains( "do_macs" ) ) {
	    DO_MACS = config.getBoolean( "do_macs" );
	}
	
	if (DEBUG) Debug.turnOn (tag);
	tag = tag + ".[" + _self_node_id.port () + "]";

	ostore.util.Types.registerAllTypes ();

	LOCATE_CLASS = 
	    Class.forName ("ostore.tapestry.api.TapestryLocateMsg");

	_using_bloom = config.getBoolean ("using_bloom");

	String stageName = config.getStage().getName();
	try
	{
	    if(config.getBoolean("Profile"))
	    {
		config.getManager().getProfiler().add("ostore.tapestry.impl.Router", this);
		SinkIF sink =
		    config.getManager().getStage(stageName).getSink();
		_sink = new SinkProxy(sink,
			      config.getManager(),
			      config.getStage().getWrapper());
	    }
	    else
		_sink =
		    config.getStage().getSink();
	} 
	catch (NoSuchStageException e) 
	{
	    Carp.die (tag + ": could not get own sink.");
	}

	_waiting_reqs = new LinkedList ();
	_routing_tbl = null;
	
	_network_latencies = new HashMap ();
	_pings_in_flight = new HashMap ();
	_pending_back_pointer_times = new HashMap ();

	_simulator_running = config.getBoolean("simulator_running");

	dynamic_route = config.getString ("dynamic_route");
	if (dynamic_route == null || dynamic_route.equals("static")) {
	  //starting using static tapestry, so 
	  //router needs to send _backpointers to DynamicTClient
	  _backpointers = new HashMap ();
	} else {
	  //starting using dynamic tapestry, will
	  //wait to get objectMap from DynamicTClient
	  _backpointers = null;
	    Filter filter = new Filter ();
	    if (! filter.requireType(
		  Class.forName("ostore.tapestry.impl.Router$ObjectMap")))
	      BUG (tag + ": could not require type " + 
		   "ostore.tapestry.impl.Router$ObjectMap");
	    Classifier.getClassifier(_self_node_id).subscribe (filter, _sink);
	}

	// get fault-resilient routing options
	MULTICAST = config.getBoolean("ft_multicast");
	int mcthresh = config.getInt("multicast_threshold");
	if (mcthresh >= 0 && mcthresh <= 100)
	  THRESHOLD_MULTICAST = mcthresh;
	else if (mcthresh != -1)
	  Carp.die("Invalid value for multicast threshold "+mcthresh+"%");
	if (MULTICAST) {
	  int mcqueuesize = config.getInt("multicast_queue_size");
	  if (mcqueuesize > 0)
	    MULTICAST_QUEUE_SIZE = mcqueuesize;
	  else if (mcqueuesize != -1)
	    Carp.die("Invalid value for multicast queue size "+mcqueuesize);
	}

	boolean disablefrls = config.getBoolean("disable_frls");
	if (disablefrls) FRLS = false;
	if (FRLS) {
	  System.err.println("FRLS IS TURNED ON");
	  int frlsthresh = config.getInt("frls_threshold");
	  if (frlsthresh >= 0 && frlsthresh <=100)
	    THRESHOLD_FRLS = frlsthresh;
	  else if (frlsthresh != -1)
	    Carp.die("Invalid value for FRLS maximum failure threshold "+
		     frlsthresh);
	  System.err.println("FRLS IS TURNED ON, THRESHOLD="+THRESHOLD_FRLS);
	} else
	  System.err.println("FRLS IS TURNED OFF");
	int deadthresh = config.getInt("inactive_node_threshold");
	if (deadthresh > 0)
	  THRESHOLD_DEAD = deadthresh;
	else if (deadthresh != -1)
	  Carp.die("Invalid value for inactive_node_threshold "+deadthresh);
	int replacethresh = config.getInt("replace_node_threshold");
	if (replacethresh > 0)
	  THRESHOLD_REPLACE = replacethresh;
	else if (replacethresh != -1)
	  Carp.die ("Invalid value for replace_node_threshold "+
		    replacethresh);

	// get local optimization parameters
	int backups = config.getInt( "local_opt_backups" );
	if( backups < -1 ) {
	    Carp.die( "Invalid value for local_opt_backups: " + backups );
	} else if( backups != -1 && backups < RouteEntry.REDUNDANT_LVL ) {
	    LOCAL_OPT_BACKUPS = backups;
	}
	int nearest = config.getInt( "local_opt_nearest" );
	if( nearest < -1 ) {
	    Carp.die( "Invalid value for local_opt_nearest: " + nearest );
	} else if( nearest != -1 ) {
	    LOCAL_OPT_NEAREST = nearest;
	}
	int thresh_fact = config.getInt( "local_opt_threshold_factor" );
	if( thresh_fact < -1 ) {
	    Carp.die( "Invalid value for local_opt_threshold_factor: " + 
		      thresh_fact );
	} else if( thresh_fact != -1 ) {
	    LOCAL_OPT_THRESHOLD_FACTOR = thresh_fact;
	}
	int hops = config.getInt( "local_opt_hops" );
	if( hops < -1 ) {
	    Carp.die( "Invalid value for local_opt_hops: " + hops );
	} else if( hops != -1 ) {
	    LOCAL_OPT_HOPS = hops;
	}
	_publish_to_root_surrogate = 
	    config.getBoolean( "publish_to_root_surrogate" );

	// for deciding when to drop messages
	_fault_controller = new FaultController();
	// for dealing with local objects and republishes
	localobjects = new HashMap();

	if (config.getBoolean ("fake_keys")) {
	  String fakeguid = config.getString ("fake_guid");
	  if (fakeguid != null) {
	    _self_guid = new SHA1Hash(fakeguid.getBytes(), true);
	  } else {
	    int offset = config.getInt("fake_key_offset");
	    if (offset != -1) {
	      _self_guid = new SHA1Hash (_self_node_id.toString()+offset);
	    } else 
	      _self_guid = new SHA1Hash(_self_node_id);
	  }
	} else {
	    String keyfilename = config.getString ("pkey");
	    try {
		FileInputStream keyfile = new FileInputStream (keyfilename);
		_pkey = (QSPublicKey) QSIO.read (keyfile);
		keyfile.close ();
	    }
	    catch (Exception e) {
		Carp.die ("TClient.init:  Caught exception " + e +
			" while trying to read pkey from disk.");
	    }

	    _self_guid = new SHA1Hash (_pkey);
	}

	System.out.println (tag + ": Tapestry Node ID is " + 
			    RoutingTable.printGuid(_self_guid) + ".");

	Filter filter;
	String [] event_types = {
	    "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.tapestry.impl.RoutingTable$Available",
	    "ostore.tapestry.impl.NetworkLatencyReq",
	    "ostore.tapestry.impl.NetworkLatencyResp",
	    "ostore.tapestry.api.TapestryPublishMsg",
	    "ostore.tapestry.api.TapestryUnpublishMsg",
	    "ostore.tapestry.api.TapestryIntegrateMsg",
	    "ostore.tapestry.api.TapestryObjPointerSetReq",
	    "ostore.mac.MacResponse",
	    "ostore.mac.MacHandler$KeyEstablished",
	    //	    "ostore.tapestry.patchwork.PatchworkDeltaResponseMsg",
	    "ostore.network.patchwork.DeltaMsg",
	    "ostore.tapestry.channel.RetransmissionCheckMsg",
	    "ostore.tapestry.api.TapestryDetachConfirm"
	};

	for (int i = 0; i < event_types.length; ++i) {
	    filter = new Filter ();
	    if (! filter.requireType (Class.forName (event_types [i])))
		BUG (tag + ": could not require type " + event_types [i]);
	    Classifier.getClassifier(_self_node_id).subscribe (filter, _sink);
	}

	String [] inb_msg_types = {
	    "ostore.tapestry.channel.ICMPPingRouteMsg",
	    "ostore.tapestry.impl.PongMsg",
	    "ostore.tapestry.impl.PingMsg",
	    "ostore.tapestry.impl.RouteMsg",
	    "ostore.tapestry.impl.RouteObjRootMsg",
	    "ostore.tapestry.impl.RouteObjDestMsg",
	    "ostore.tapestry.impl.PublishMsg",
	    "ostore.tapestry.impl.BloomWrapper",
	    "ostore.tapestry.patchwork.RouterFaultMsg",
	    "ostore.tapestry.channel.ChannelACKMsg",
	    "ostore.tapestry.impl.RepublishRequestMsg"
	};

	for (int i = 0; i < inb_msg_types.length; ++i) {
	    filter = new Filter ();
	    if (! filter.requireType (Class.forName (inb_msg_types [i])))
		BUG (tag + ": could not require type " + inb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (true)))
		BUG (tag + ": could not require inbound = true for " 
			+ inb_msg_types [i]);
	    Classifier.getClassifier(_self_node_id).subscribe (filter, _sink);
	}

	String [] outb_msg_types = {
	    "ostore.tapestry.channel.ICMPPingMsg",
	    "ostore.tapestry.api.TapestryRouteMsg",
	    "ostore.tapestry.api.TapestryPrefixRouteMsg",
	    "ostore.tapestry.api.TapestryLocateMsg"
	};

	for (int i = 0; i < outb_msg_types.length; ++i) {
	    filter = new Filter ();
	    if (! filter.requireType (Class.forName (outb_msg_types [i])))
		BUG (tag + ": could not require type " + outb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (false)))
		BUG (tag + ": could not require inbound = false for " 
			+ outb_msg_types [i]);
	    Classifier.getClassifier(_self_node_id).subscribe (filter, _sink);
	}
    }

    public void destroy() throws Exception {
	// TODO
    }

    //  methods for ProfilableIF
    public int profileSize()
    {
	return _sink.size();
    }
    //  end methods for ProfilableIF

    /**
     * For use in derived classes.
     */
    protected void handle_stages_init_signal () {

	if (DO_MACS) {
	    _mac_stage = HMACDHHandler.getInstance (_self_guid);
	}

	// Added by Jeremy and Ling Huang for Channel operation: 06/28/2002.
	// After all stages are created, create a channel for the router.
	// This works fine for patchwork.
	_channel_handler = new ChannelHandler(this);

	// Let other nodes know the local node guid.
	// Most stages extract local node guid from
	// TapestryReadyMsg, but some stages need
	// to know the local node guid before
	// the tapestry is ready, e.g. DynamicTClient.
	dispatch(new NodeGuid(_self_guid));

	EnqueueTimeComparator comparator = new EnqueueTimeComparator();
	_sorted_multicast_msgs  = new TreeSet(comparator);
	_multicast_msgs  = new HashSet();


	_self_digits = RoutingTable.digits (_self_guid);
	//if we're using staticTC, then send reference of backpointers
	//to the DynamicTClient stage
	if (dynamic_route == null || dynamic_route.equals("static")) {
	  dispatch(new Router.ObjectMap(_backpointers));
	}
	System.err.println("TapestryInfo: NodeID="+_self_node_id+
			   ", NodeGUID="+_self_guid);
    }

  public static void dumpRevpointers (HashMap revs, int level) {
    HashMap currlevel;
    SecureHash s;
    if (revs == null) {
      if (DEBUG)
	Debug.printtagln("Router", "Empty Revpointer being dumped.");
      return;
    }
    for (int i=0; i< level; i++) {
      currlevel = (HashMap) revs.get(new Integer(i));
      System.err.println("Revpointers for level "+i);
      if (currlevel == null) System.err.println("NONE");
      else {
	for (Iterator it = currlevel.keySet().iterator(); it.hasNext();) {
	  s = (SecureHash) it.next();
	  System.err.print("Guid: "+s+" NodeID: "+(NodeId) currlevel.get(s));
	}
	System.err.print("\n");
      }
    }
  }

  public static Set getRevpointers (Map revs, int level) {
    HashMap currlevel;
    SecureHash guid;
    Set set = new TreeSet();
    if(revs == null)
      return set;

    currlevel = (HashMap) revs.get(new Integer(level));
    
    if(currlevel == null)
      return set;

    for(Iterator it = currlevel.keySet().iterator(); it.hasNext();) {
      guid = (SecureHash) it.next();
      set.add(new GuidAddr(guid, (NodeId)currlevel.get(guid)));
    }
    return set;
  }

    private void handle_icmp_ping_route_msg(ICMPPingMsg msg){//added by Lih Chen
	if (DEBUG)
	    Debug.printtagln(tag, "handle_icmp_ping_route_msg >> ICMPPingMsg = "+ 
			     msg + "; dispatch an ICMPPongMsg.");
	ICMPPongMsg pongMsg = new ICMPPongMsg(msg.src_id, _self_node_id, 
					      _self_guid, msg.msg_id,
					      msg.hop);
	msg.hop=msg.hop+1;
	dispatch(pongMsg);
    }


    protected void handle_icmp_ping_msg( ICMPPingMsg msg ){//added by Lih Chen
	if (_using_bloom) {
	    dispatch (new BloomWrapper (msg.peer, null, null, null, msg));
	    return;
	}

	int [] dest = RoutingTable.digits (msg.peer);
	int [] digit = {0};
	NodeId next = _routing_tbl.route_to_root (dest, true, false, digit);

	if (next == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
	      //MARK
	      System.out.println("bad message: "+msg);
	      System.out.println(_routing_tbl);

  	      Carp.out (tag + ": " + 
  			"got TapestryRouteMsg for non-existent " +
  			"destination.  Dropping it.");
	    }
	    return;
	}

	if (next.equals (_self_node_id)) {
	    // This is to us.

	    TapestryRouteMsg inb = (TapestryRouteMsg) ByteUtils.qs_clone (msg);
	    inb.peer = _self_guid;
	    inb.inbound = true;
	    inb.hopCount = 0;
	    //no worries about the TTL, since we took no hops
	    // The MAC need not be verified; it's from us, after all.

	    dispatch (inb);
	}
	else {

	    ICMPPingRouteMsg outb = 
		new  ICMPPingRouteMsg(next, _self_guid, msg.peer, false,
				      digit [0], msg, msg.TTL);
	    if (DEBUG)
		Debug.printtagln(tag, "Get an ICMPPingMsg = " + msg + 
				 " and dispatch it out as " + outb);
	    //Added by Ling Huang: handle reliable message.
	    outb.set_qos(msg.qos);
	    
	    dispatch (outb);
	}
    }

    protected void handle_tapestry_route_msg (TapestryRouteMsg msg) {
	if (DEBUG) 
	    Debug.printtagln (tag, "processing TapestryRouteMsg " + msg);

	if (_using_bloom) {
	    dispatch (new BloomWrapper (msg.peer, null, null, null, msg));
	    return;
	}

	int [] dest = RoutingTable.digits (msg.peer);
	int [] digit = {0};

	RouteEntry entry = _routing_tbl._route_to_root (dest, true, false, digit);
	NodeId next = (entry==null?null:entry.getFRLS());

	if (next == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
	      //MARK
	      System.out.println("bad message: "+msg);
	      System.out.println(_routing_tbl);

  	      Carp.out (tag + ": " + 
  			"got TapestryRouteMsg for non-existent " +
  			"destination.  Dropping it.");
	    }
	    return;
	}

	if (next.equals (_self_node_id)) {
	    // This is to us.

	    TapestryRouteMsg inb = (TapestryRouteMsg) ByteUtils.qs_clone (msg);
	    inb.peer = _self_guid;
	    inb.inbound = true;
	    inb.hopCount = 0;
	    //no worries about the TTL, since we took no hops
	    // The MAC need not be verified; it's from us, after all.

	    dispatch (inb);
	}
	else {
	    RouteMsg outb = 
		new RouteMsg (next, _self_guid, msg.peer, false,
			digit [0], msg, msg.TTL);

	    //Added by Ling Huang: handle reliable message.
	    outb.set_qos(msg.qos);

	    if (DO_MACS && (msg instanceof TapestryMacRouteMsg)) {

		if (_mac_stage == null) {
		    throw new AssertionViolatedException (
			    "DO_MACS=true but no MAC stage is running.");
		}

		if (MAC_DEBUG) Debug.printtagln (tag, 
			"Adding a MAC to " + outb + ".");

		try {
		    dispatch (new MacRequest (outb, msg.peer, 
				outb.data_to_mac ()));
		}
		catch (QSException e) {
		    Carp.die (tag + ": unreachable 2.");
		}
		return;

		/*
		QSByteArray mac = null;
		try {
		    mac = _mac_stage.doMAC (outb.data_to_mac (), msg.peer);
		}
		catch (MacHandler.GuidNotEstablished e) {
		    if (MAC_DEBUG) Debug.printtagln (tag, 
			    "No secret key established with " +
			    msg.peer + " yet.  Enqueuing " + outb + ".");

		    synchronized (_msgs_needing_macs) {
			LinkedList l = 
			    (LinkedList) _msgs_needing_macs.get (msg.peer);
			if (l == null) {
			    if (MAC_DEBUG) Debug.printtagln (tag, 
				    "Starting a key agreement with " 
				    + msg.peer + ".");

			    l = new LinkedList ();
			    _msgs_needing_macs.put (msg.peer, l);
			    _mac_stage.establish (msg.peer);
			}
			else if (MAC_DEBUG) Debug.printtagln (tag, 
				"Key agreement already started with "
				+ msg.peer + ".");

			l.addLast (outb);
		    }
		    return;
		}

		outb.mac = mac;

		if (MAC_DEBUG) Debug.printtagln (tag, "MAC added to " + 
			outb + ".");
		*/
	    }

	    if (MAC_DEBUG && (msg instanceof ostore.mac.DHMessage))
		Debug.printtagln (tag, "Dispatching " + outb + ".");

	    //Modified by Ling Huang for fault-tolerance routing.
	    //if link quality is not so good, do multicast.
	    if (MULTICAST && entry.need_multicast(THRESHOLD_MULTICAST)) {
	      dispatch_multicast(outb, entry);
	    } else
	      dispatch (outb);
	}
    }

    protected void handle_tapestry_prefix_route_msg (
	    TapestryPrefixRouteMsg msg) {
	if (DEBUG) Debug.printtagln (tag, "processing " + 
		"TapestryPrefixRouteMsg, full type " + 
		msg.getClass ().getName () + ". msg==" + msg);

	int [] dest = RoutingTable.digits (msg.peer);
	int [] digit = {0};
	// note: now prefix message just routes as a routemsg
	// looking for an inexact match. This way, we use the same
	// surrogate routing routine as the object location msg.

	//Modified by Ling Huang for fault-tolerance routing.
	//	NodeId next = _routing_tbl.route_to_root (dest, false, false, digit);
	RouteEntry entry = _routing_tbl._route_to_root (dest, false, false, digit);
	NodeId next = (entry==null?null:entry.getFRLS());

	if (next == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
	      //MARK
	      System.out.println("bad message: "+msg);
	      System.out.println(_routing_tbl);

  	      Carp.out (tag + ": " + 
  			"got TapestryPrefixRouteMsg for non-existent " +
  			"destination.  Dropping it.");
	    }
	    return;
	}

	if (next.equals (_self_node_id)) {

	    if (DEBUG) Debug.printtagln (tag, 
		    "It's to us.  Dispatching it locally.");
	
	    // This is to us.

	    TapestryPrefixRouteMsg inb = 
		(TapestryPrefixRouteMsg) ByteUtils.qs_clone (msg);
	    inb.peer = _self_guid;
	    inb.inbound = true;
	    inb.hopCount = 0;
	    dispatch (inb);
	    //no need to worry about TTsLs, hop=0
	}
	else {
	    if (DEBUG) Debug.printtagln (tag, 
		    "It's not to us.  Forwarding it to " + next + ".");
	
	    RouteMsg outb = 
		new RouteMsg (next, _self_guid, msg.peer, true,
			digit [0], msg, msg.TTL);

	    //Added by Ling Huang: handle reliable message.
	    outb.set_qos(msg.qos);

	    //Modified by Ling Huang for fault-tolerance routing.
	    //if link quality is not so good, do multicast.
	    if (MULTICAST && entry.need_multicast(THRESHOLD_MULTICAST)) {
	      dispatch_multicast(outb, entry);
	    } else
	      dispatch (outb);
	}
    }

    protected void handle_route_msg (RouteMsg msg) {
	if (DEBUG) 
	    Debug.printtagln (tag, "processing RouteMsg, " +
			      "msg==" + msg);

	//Added by Ling Huang. If a router gets a reliable msg, the first thing is
	//to send a ack to msg's previous Tapestry hop.
	if ( msg.is_reliable() )
	{
	    //If this node is not the one that this msg originates, 
	    //it should send ack to its previous node.
	    if ( (msg.src).equals(_self_guid) )
	    {
		if (DEBUG) 
		    Debug.printtagln (tag, "It is the original hop," +
		    "don't send ack" + ".");
	    }
	    else
	    {
		if (DEBUG) 
		    Debug.printtagln (tag, "It is a intermediate" +
		    "hop, send ack to " + msg.peer + ".");
		_channel_handler.send_ack(msg.peer,msg.msg_id);
	    }
	}

	int [] dest = RoutingTable.digits (msg.dest);
	int [] digit = {msg.hop + 1};

	//	NodeId next = _routing_tbl.route_to_root (dest, !msg.prefix_only, 
	//						  false/*msg.prefix_only*/, digit);
	RouteEntry entry = _routing_tbl._route_to_root (dest, !msg.prefix_only, 
						  false/*msg.prefix_only*/, digit);
	NodeId next = (entry==null?null:entry.getFRLS());

	int TTLleft = msg.TTL - 1;

	if (next == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
	      QuickSerializable user_data = null;
	      try {
		user_data = msg.user_data();
	      } catch(QSException qse) {
		Carp.out(qse.toString());
	      }
	      Carp.out (tag + ": bad message: user_data= "+ user_data);
	      Carp.out (tag + ": bad message: "+msg);
	      Carp.out (tag + ": " + _routing_tbl);

	      Carp.out (tag + ": " + 
		    "got RouteMsg for non-existent " +
		    "destination: " + msg.dest + ".  Dropping it.");
	      Carp.out (tag + ": " + 
		    "Current RouteEntry is: "+entry);
	      if( entry == null ) {
		  return;
	      }
	      Carp.out (tag + ": " + 
		    "The frls is "+entry.get_frls());
	      Carp.out (tag + ": " + 
		    "The routes in entry are:");
	      int i;
	      RouteEntry.NodeEntry [] routes = entry.getRoutes();
	      for (i = 0; i < 3; i++)
	      {
		  if (routes[i] != null)
		      Carp.out (tag + ": route["+i+"] = "+routes[i].address);
		  else
		      Carp.out (tag + ": route["+i+"] = null");
	      }
	    }
	    return;
	}

	if (next.equals (_self_node_id) || TTLleft == 0) {
	    // This is to us.
	    if (DEBUG) Debug.printtagln (tag, "We're the end of the line for " + msg);

	    QuickSerializable user_data = null;
	    QuickSerializable data_to_mac = null;
	    try {
		user_data = msg.user_data ();
		data_to_mac = msg.data_to_mac ();
	    }
	    catch (QSException e) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
		  Carp.out (tag + ": " +"caught " + e + 
			    " while decoding " + msg + ".  Dropping it...");
		  return;
	    }
	    
	    if (user_data instanceof TapestryRouteMsg) {

		TapestryRouteMsg inb = null;
		try {
		    inb = (TapestryRouteMsg) user_data;
		}
		catch (ClassCastException e) {
		    BUG (tag + ": should have caught this above.");
		    return;
		}
		inb.peer = msg.src;
		inb.inbound = true;
		inb.hopCount = msg.hop+1;

		if ((inb instanceof TapestryMacRouteMsg) && DO_MACS) {

		    if (_mac_stage == null) {
			if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
			    Carp.out (tag +  ": got " + msg + 
				    " but no MAC stage is running.  " +
				    "Dropping it.");
			return;
		    }

		    if (MAC_DEBUG) Debug.printtagln (tag, 
			    "Checking MAC from " + msg.src + ".");

		    try {
			if ((msg.mac == null) || 
			    (! _mac_stage.authenticate 
			     (data_to_mac, msg.mac, msg.src))) {

			    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out
				(tag + ": got " + msg + " from " + msg.src 
				 + "via " + msg.peer + " with bad MAC." +
				 "  Dropping it.");
			    // TODO: maybe retry it instead?
			    return;
			}
		    }
		    catch (MacHandler.GuidNotEstablished e) {
			if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out
			    (tag + ": got " + msg + " from " + msg.src + 
			     "via " + msg.peer + " with a MAC, but " +
			     "no secret key is currently established" +
			     " with that node.  Dropping it.");
			// TODO: maybe retry it instead?
			return;
		    }

		    if (MAC_DEBUG) Debug.printtagln (tag, 
			    "MAC checked out on " + inb + ".");
		}

		if (DO_MACS && (inb instanceof ostore.mac.DHMessage)) {
		    if (_mac_stage == null) {
			if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
			    Carp.out (tag +  ": got " + msg + 
				    " but no MAC stage is running.  " +
				    "Dropping it.");
			return;
		    }

		    _mac_stage.handle_dh_message ((ostore.mac.DHMessage) inb);
		}
		else
		    dispatch (inb);
	    }
	    else if (user_data instanceof TapestryPrefixRouteMsg) {
		TapestryPrefixRouteMsg inb = null;
		try {
		    inb = (TapestryPrefixRouteMsg) user_data;
		}
		catch (ClassCastException e) {
		    BUG (tag + ": should have caught this above.");
		    return;
		}
		inb.peer = msg.src;
		inb.inbound = true;
		inb.hopCount = msg.hop+1;

		dispatch (inb);
		
	    }
	    else {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
		  Carp.out (tag + ": bad message: "+msg);
		  Carp.out (tag + ": " + _routing_tbl);
		  Carp.out (tag + ": " +
			    "got RouteMsg with bad payload.  " +
			    "Dropping it.");
		}
		return;
	    }
	}
	else 
	{
	    if (DEBUG) Debug.printtagln (tag, "Forwarding to " + next + ".");
	    RouteMsg outb = null;
	    try {
		outb = (RouteMsg) msg.clone ();
	    }
	    catch (CloneNotSupportedException e) {
		BUG (tag + ": wtf?");
	    }
	    outb.hop = digit [0];
	    outb.peer = next;
	    outb.inbound = false;
	    outb.TTL = TTLleft;

	    //Modified by Ling Huang for fault-tolerance routing.
	    //if link quality is not so good, do multicast.
	    if (MULTICAST && entry.need_multicast(THRESHOLD_MULTICAST)) {
	      dispatch_multicast(outb, entry);
	    } else 
	      dispatch (outb);
	}
    }

    protected void handle_tapestry_publish_msg (TapestryPublishMsg msg) {
	if (DEBUG) Debug.printtagln (tag, "processing TapestryPublishMsg " + msg);
	localobjects.put(msg.guid, msg);
	publish (_self_guid, msg.guid, msg.tag, true /*publishing*/,
		 _self_node_id, msg.qos, msg.local_opts, LOCAL_OPT_HOPS);
    }

    protected void handle_tapestry_unpublish_msg(TapestryUnpublishMsg msg) {
      if (DEBUG) Debug.printtagln(tag, "processing TapestryUnpublishMsg "+
				  msg);
      localobjects.remove(msg.guid);
      publish(_self_guid, msg.guid, msg.tag, false /*unpublishing*/,
	      _self_node_id, msg.qos, msg.local_opts, LOCAL_OPT_HOPS );
    }

    protected void handle_republish_request(RepublishRequestMsg msg) {
      if (DEBUG) Debug.printtagln(tag, "processing RequestRepublishMsg "+
				  msg);
      TapestryPublishMsg tpm = (TapestryPublishMsg) 
	localobjects.get(msg.objGuid);
      if (tpm != null) 
	publish(_self_guid, tpm.guid, tpm.tag, true /*publishing*/,
		_self_node_id, tpm.qos);
      else if (DEBUG)
	Debug.printtagln(tag, "Missing TapestryPublishMsg for object to"+
			 " be republished.");
    }

    protected void handle_publish_msg (PublishMsg msg) {
	if (DEBUG) Debug.printtagln (tag, "processing PublishMsg " + msg);

	//Added by Ling Huang. If a router gets a reliable msg, the first thing is
	//to send a ack to msg's previous Tapestry hop.
	if ( msg.is_reliable() )
	{
	    //If this node is not the one that this msg originates, 
	    //it should send ack to its previous node.
	    if ( (msg.src).equals(_self_guid) )
	    {
		if (DEBUG) Debug.printtagln (tag, "It is the original hop," +
			   "don't send ack" + ".");
	    }
	    else
	    {
		if (DEBUG) Debug.printtagln (tag, "It is a intermediate" +
			   "hop, send ack to " + msg.peer + ".");
		_channel_handler.send_ack(msg.peer,msg.msg_id);
	    }
	}

	publish (msg.src, msg.guid, msg.tag, msg.publish, msg.peer, 
		 msg.get_qos(), msg.local_opts, msg.local_opts_ttl);
    }

    protected void handle_bloom_wrapper (BloomWrapper wrapper) {
	if (wrapper.success) {
	    if (wrapper.user_data instanceof TapestryLocateMsg) {
		TapestryLocateMsg inb = 
		    (TapestryLocateMsg) ByteUtils.qs_clone (wrapper.user_data);

		inb.peer = wrapper.src;
		inb.inbound = true;
		inb.guid = wrapper.dest;
		inb.query = wrapper.query;
		inb.query_state = wrapper.query_state;
		inb.id = wrapper.id;

		dispatch (inb);
	    }
	    else if (wrapper.user_data instanceof TapestryRouteMsg) {
		TapestryRouteMsg inb = 
		    (TapestryRouteMsg) ByteUtils.qs_clone (wrapper.user_data);

		inb.peer = wrapper.src;
		inb.inbound = true;

		dispatch (inb);
	    }
	    else {
		BUG (tag + ": don't know what to do with " + 
			wrapper.user_data + ".");
	    }
	}
	else {
	    if (wrapper.user_data instanceof TapestryLocateMsg) {

		RouteObjRootMsg r = new RouteObjRootMsg (null, wrapper.src,
			wrapper.dest, wrapper.query, wrapper.query_state, 
			wrapper.id, -1, wrapper.user_data, _self_node_id);

		//Added by Ling Huang: handle reliable message.
		r.set_qos(((TapestryLocateMsg)wrapper.user_data).qos);

		handle_route_obj_root_msg (r);
	    }
	    else if (wrapper.user_data instanceof TapestryRouteMsg) {

		RouteMsg r = new RouteMsg (null, wrapper.src, wrapper.dest,
			false, -1, wrapper.user_data, -1);

		//Added by Ling Huang: handle reliable message.
		r.set_qos(((TapestryRouteMsg)wrapper.user_data).qos);

		handle_route_msg (r);
	    }
	    else {
		BUG (tag + ": don't know what to do with " + 
			wrapper.user_data + ".");
	    }
	}
    }

    protected void handle_tapestry_locate_msg (TapestryLocateMsg msg) {
	if (DEBUG) Debug.printtagln (tag, "processing TapestryLocateMessage "+
				     msg);

	if (_using_bloom && 
	    ((msg.query == null) || 

	     /* TODO: hack for Patrick's web stuff */
	     (msg.query instanceof ostore.replica.ReplicaQuery))) {

	    dispatch (new BloomWrapper (msg.guid, msg.id, msg.query,
			msg.query_state, msg));
	    return;
	}

	try {
	// save this stuff because the values in the msg might be
	// removed by route_to_dest if we send this to the local node
	TapestryQuery query = msg.query;
	TapestryQueryState query_state = msg.query_state;
	if (! route_to_dest (msg.guid, _self_guid, msg.query,
		    msg.query_state, msg, 0, null, 0, msg.TTL, msg.qos)) {
	    //returns false, means the query is not satisfied by the
	    //current node's object pointers, must keep routing to root
  	    if (msg.TTL == 0) {
	      // SENDING TAPESTRYFAILURE BACK
	      if (msg.id != null) {
		if (DEBUG) 
		  Debug.printtagln(tag, "we've run out on the TTL count" +
				   "even before reaching the object root.\n"+
				   "Sending back tapestryLocateFailure.");
		dispatch(new TapestryLocateFailure
		  (_self_guid, msg.id, msg.guid, _self_node_id, 0));
	      } else if (DEBUG) {
		Debug.printtagln(tag, "we've run out on the TTL count but"+
				 "no msg.id found, so dropping it.");
	      }
	      return;
	    }

	    int [] dest = RoutingTable.digits (msg.guid);
	    int [] digit = {0};
	    NodeId next = _routing_tbl.route_to_root (dest, false, false, digit);

	    if (! next.equals (_self_node_id)) {
		RouteObjRootMsg outb = new RouteObjRootMsg (
			next, _self_guid, msg.guid, query,
			query_state, msg.id, digit [0], msg,
			_self_node_id);
		outb.TTL = msg.TTL;
		if(DEBUG) Debug.printtagln(tag, 
					   "we are NOT the root." +
					   " we are forwarding the" +
					   " TapestryLocatemsg==" + msg +
					   " towards the root." +
					   "  outbound==" + outb);

		//Added by Ling Huang: handle reliable message.
		outb.set_qos(msg.qos);
		
		// keep the local opts going
		outb.local_opts = msg.local_opts;
		outb.local_opts_ttl = LOCAL_OPT_HOPS;

		boolean do_send = perform_local_locate_opts( outb, dest, 
							 _self_node_id, 0 );
		    		
		if( do_send ) {
		    dispatch (outb);
		}
	    }
	    else {
		// If we are the root, and we have no matching
		// backpointers, the location request has failed.
		if (msg.id != null)
		{
		    if(DEBUG) 
		      Debug.printtagln(tag, "If we are the root, and we" +
				       " have no matching backpointers, the location" +
				       " request has failed." +
				       " backpointers=" + _backpointers);
		    dispatch (new TapestryLocateFailure 
		      (_self_guid, msg.id, msg.guid, _self_node_id, 0));
		}
		else
		  if(DEBUG) Debug.printtagln(tag, 
					     "we ARE the root." +
					     " TapestryLocateMsg==" + msg +
					     " we have followed some" +
					     " backpointers so we are" +
					     " dropping msg!");
	    }
	}
	else
	  if(DEBUG) Debug.printtagln(tag, 
				     "We have followed ENOUGH backpointers" +
				     " so we are stopping!" +
				     " TapestryLocateMsg==" + msg);
	}
	catch (QSException e) {
	    Carp.die (tag + ": unreachable 4.");
	}
    }

    protected void handle_route_obj_root_msg (RouteObjRootMsg msg) {
	if (DEBUG) Debug.printtagln (tag, "processing RouteObjRootMsg " + msg);

	//Added by Ling Huang. If a router gets a reliable msg, the first thing is
	//to send a ack to msg's previous Tapestry hop.
	if ( msg.is_reliable() )
	{
	    //If this node is not the one that this msg originates, 
	    //it should send ack to its previous node.
	    if ( (msg.src).equals(_self_guid) )
	    {
		if (DEBUG) Debug.printtagln (tag, "It is the original hop," +
			   "don't send ack" + ".");
	    }
	    else
	    {
		if (DEBUG) Debug.printtagln (tag, "It is a intermediate" +
			   "hop, send ack to " + msg.peer + ".");
		_channel_handler.send_ack(msg.peer,msg.get_msg_id());
	    }
	}

	//TTLs are always deducted upon reception by the destination node
	int TTL = msg.TTL -1;
	try {
	    boolean result = false;
	    
	    if (msg._user_data_buffer != null) {
		result = route_to_dest (msg.guid, msg.src, msg.query, 
			msg.query_state, null, msg._user_data_length, 
			msg._user_data_buffer, msg.hop+1, TTL, msg.get_qos());
	    }
	    else {
		// if we're running under the simulator, the message has
		// never been serialized, so we have to use user_data
		// instead of the _user_data_buffer
		result = route_to_dest (msg.guid, msg.src, msg.query, 
					msg.query_state, msg.user_data (), 
					0, null, msg.hop+1, TTL, msg.get_qos());
	    }

        if (! result) {
  	    if (TTL == 0) {
	      // need to route more, but ttl=0, sending TapestryFailure
	      if (msg.id != null) {
		if (DEBUG) 
		  Debug.printtagln(tag, "we've run out on the TTL count" +
				   "even before reaching the object root.\n"+
				   "Sending back tapestryLocateFailure.");
		dispatch(new TapestryLocateFailure
		  (_self_guid, msg.id, msg.guid, msg.requestor, msg.hop+1));
	      } else if (DEBUG) {
		Debug.printtagln(tag, "we've run out on the TTL count but"+
				 "no msg.id found, so dropping it.");
	      }
	      return;
	    }

	    if (DEBUG) Debug.printtagln (tag, "no route to dest" );
	    int [] dest = RoutingTable.digits (msg.guid);
	    int [] digit = {msg.hop + 1};
	    NodeId next = _routing_tbl.route_to_root (dest, false, false, digit);
	    if (DEBUG) Debug.printtagln (tag, "next: " + next);

	    if (! next.equals (_self_node_id)) {
		RouteObjRootMsg outb = null;
		try {
		    outb = (RouteObjRootMsg) msg.clone ();
		}
		catch (CloneNotSupportedException e) {
		    BUG (tag + ": unreachable 5.");
		}
		outb.peer = next;
		outb.inbound = false;
		outb.hop = digit [0];
		if (DEBUG) Debug.printtagln (tag, "outb: " + outb );

		// keep the local opts going
		msg.local_opts_ttl--;
		outb.local_opts = msg.local_opts;
		outb.local_opts_ttl = msg.local_opts_ttl;
		
		boolean do_send = perform_local_locate_opts( outb, dest, 
							     msg.peer,msg.hop);

		if( do_send ) {
		    dispatch (outb);
		}
	    } 
	    else {
		// If we are the root, and we have no matching
		// backpointers, the location request has failed.
		if (msg.id != null) {
		    dispatch (new TapestryLocateFailure (
				_self_guid, msg.id, msg.guid, msg.requestor,
				msg.hop+1));
		}
                else
                {
                    if (DebugFlags.LOG_BAD_NETWORK_EVENTS)
			Debug.printtagln(tag, "Got LocateMessage for " +
					 msg.guid + " from " + msg.requestor +
					 "msg=" + msg + "Dropping it.");
                }
	    }
	}
	}
	catch (QSException e) {
	    e.printStackTrace ();
	    System.err.println ("msg = " + msg);
	    Carp.die (tag + ": unreachable 6.");
	}
    }

    private Class LOCATE_CLASS;

    public void handle_route_obj_dest_msg (RouteObjDestMsg msg) {
	if (DEBUG) Debug.printtagln (tag, "processing RouteObjDestMsg " + msg);

	//Added by Ling Huang. If a router gets a reliable msg, the first thing is
	//to send a ack to msg's previous Tapestry hop.
	if ( msg.is_reliable() )
	{
	    //If this node is not the one that this msg originates, 
	    //it should send ack to its previous node.
	    if ( (msg.src).equals(_self_guid) )
	    {
		if (DEBUG) Debug.printtagln (tag, "It is the original hop," +
			   "don't send ack" + ".");
	    }
	    else
	    {
		if (DEBUG) Debug.printtagln (tag, "It is a intermediate" +
			   "hop, send ack to " + msg.peer + ".");
		_channel_handler.send_ack(msg.peer,msg.msg_id);
	    }
	}


	int [] dest = RoutingTable.digits (msg.dest);
	int [] digit = {msg.hop + 1};
	NodeId next = _routing_tbl.route_to_root (dest, true, false, digit);

	if (next == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
	      QuickSerializable user_data = null;
	      try {
		user_data = msg.user_data();
	      } catch(QSException qse) {
		Carp.out(qse.toString());
	      }
	      Carp.out (tag + ": bad message: user_data= "+ user_data);
		Carp.out(tag + ": bad message: "+msg);
		Carp.out(tag + ": " + _routing_tbl);
		Carp.out (tag + ": " + 
			  "got RouteObjDestMsg for non-existent " +
			  "destination.  Dropping it.");
	    }
	    return;
	}

	if (next.equals (_self_node_id)) {
	    QuickSerializable user_data = null;
	    try {
		user_data = msg.user_data ();
	    }
	    catch (QSException e) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": "
		+"caught " + e + 
			" while decoding " + msg + ".  Dropping it...");
		return;
	    }

	    if (user_data instanceof TapestryLocateMsg) {
		TapestryLocateMsg inb = (TapestryLocateMsg) user_data;
		inb.peer = msg.src;
		inb.inbound = true;
		inb.guid = msg.guid;
		inb.query = null;
		inb.query_state = null;
		inb.hopCount = msg.hopInit + msg.hop+1;
		dispatch (inb);
	    }
	    else {
		/*RouteObjTMessage inb = (RouteObjTMessage) user_data;
		inb.inbound = true;
		inb.usingMAC = false;
		inb.verifiedMAC = false;
		inb.peer = msg.src;
		if (!inb.objGuid.equals (msg.guid))*/
		  throw new AssertionViolatedException 
		    ( tag + ": unexplained error!");
		//dispatch (inb);
	    }
	}
	else {
	    RouteObjDestMsg outb = null;
	    try {
		outb = (RouteObjDestMsg) msg.clone ();
	    }
	    catch (CloneNotSupportedException e) {
		BUG (tag + ": wtf?");
	    }
	    outb.peer = next;
	    outb.inbound = false;
            if (digit[0] > 20) { digit[0] = 0; }
	    outb.hop = digit [0];
	    dispatch (outb);
	}
    }

    /**
     * Take into account the result of a NetworkLatencyReq query.  Unlike
     * handle_network_latency_req, below, this will remain part of Router
     * even after the network introspection code that services a
     * NetworkLatencyReq is moved out.
     */
    private void handle_network_latency_resp (NetworkLatencyResp resp) {
	if (DEBUG) Debug.printtagln (tag, "processing NetworkLatencyResp " + resp);

	LinkedList s = null;
	synchronized (_pending_back_pointer_times) {
	    s = (LinkedList) 
		_pending_back_pointer_times.remove (resp.node_guid);
	}

	if (s == null) 
	    return;

	while (! s.isEmpty ()) {
	    Pair p = (Pair) s.removeFirst ();
	    PublishInfo pi = (PublishInfo) p.first;
	    LinkedList l = (LinkedList) p.second;
	    if (DEBUG) Debug.printtagln (tag, "Changing distance of " +
		    pi + " to " + resp.latency_millis + " ms.  " +
		    "Current elts are " + l);

	    synchronized (l) {
		PublishInfo pi2 = null;
		ListIterator i = l.listIterator (); 
		while (i.hasNext ()) {
		    pi2 = (PublishInfo) i.next ();
		    if (pi.equals (pi2)){
			i.remove ();
			break;
		    }
		}

		if( pi2 == null ) {
		    // Probably unpublished already.
		    return;
		}
		pi2.distance = resp.latency_millis;

		i = l.listIterator ();
		while (i.hasNext ()) {
		    PublishInfo pi3 = (PublishInfo) i.next ();
		    if (pi2.distance < pi3.distance) {
			i.previous ();
			break;
		    }
		}
		i.add (pi2);
	    }
	}
    }

    /** 
     * Handle the event requesting the latency to some other network node
     * through Tapestry.  This should be moved to some introspective
     * network stage at some point in the future.
     */
    private void handle_network_latency_req (NetworkLatencyReq req) {
	if (DEBUG) Debug.printtagln (tag, "processing NetworkLatencyReq " + req);

	Long l = (Long) _network_latencies.get (req.node_guid);
	if (l != null) {
	    if (DEBUG) Debug.printtagln (tag, "Distance to " +
		    req.node_guid + " is already known to be " +
		    l + ".");
	    dispatch (new NetworkLatencyResp (req.node_guid, l.longValue ()));
	}
	else {
	    if (DEBUG) Debug.printtagln (tag, "Distance to " +
		    req.node_guid + " is not yet known.");
	    Long k = (Long) _pings_in_flight.get (req.node_guid);
	    if (k == null) {
		if (DEBUG) Debug.printtagln (tag, "Sending a ping.");
		_pings_in_flight.put (req.node_guid, 
			new Long ((Clock.current_date (_self_node_id)).getTime ()));
		handle_tapestry_route_msg (new PingMsg (req.node_guid));
	    }
	}
    }

    /**
     * Implements the functionality of NetworkLatencyReq.  This should be
     * moved to some introspective network stage at some point in the
     * future.
     */
    private void handle_ping_msg (PingMsg msg) {
	if (DEBUG) Debug.printtagln (tag, "processing PingMsg " + msg);

	// TODO: think about flood attacks, timeouts, sequence numbers...
	handle_tapestry_route_msg (new PongMsg (msg.peer));
    }

    /**
     * Implements the functionality of NetworkLatencyReq.  This should be
     * moved to some introspective network stage at some point in the
     * future.
     */
    private void handle_pong_msg (PongMsg msg) {
	if (DEBUG) Debug.printtagln (tag, "processing PongMsg " + msg);

	Long l = (Long) _pings_in_flight.get (msg.peer);
	if (l != null) {
	    long now = (Clock.current_date (_self_node_id)).getTime ();
	    long lat = now - l.longValue ();

	    if (DEBUG) Debug.printtagln (tag, "Updating distance to " +
		    msg.peer + " to be " + lat + " ms.");
	    
	    _network_latencies.put (msg.peer, new Long (lat));
	    _pings_in_flight.remove (msg.peer);
	    dispatch (new NetworkLatencyResp (msg.peer, lat));
	}
	else if (DEBUG) Debug.printtagln (tag, "got pong without " +
		"corresponding ping from node " + msg.peer + ".");
    }

  private void handle_object_map (Router.ObjectMap msg) {
    if (DEBUG) Debug.printtagln(tag, "processing ObjectMap "+msg);
    if(_backpointers != null)
      throw new AssertionViolatedException 
	(tag + ": got two versions of object map");
    
    _backpointers = msg.objmap;

    //check if routing table is also complete
    //if so, then start normal operation and execute queued events
    if (_routing_tbl != null) {
      try {
	while (! _waiting_reqs.isEmpty ()) {
	  QueueElementIF e = 
	    (QueueElementIF) _waiting_reqs.removeFirst ();
	  handleEvent (e);
	}
      }
      catch (EventHandlerException e) {
	BUG (tag + ": what do I do now?"); // TODO
      }
    }
  }

    /**
     * Return the set of object pointers for a guid, and whether or not
     * we're the root.
     */
    private void handle_obj_pointer_set_req( TapestryObjPointerSetReq msg ) {

	// get the linked list
	LinkedList pis = (LinkedList) _backpointers.get(msg.objguid);
	PublishInfo[] pi_ar;
	boolean root = false;
	if( pis == null ) {
	    pi_ar = new PublishInfo[0];
	} else {
	    synchronized( pis ) {
		pi_ar = new PublishInfo[pis.size()];
		for( int i = 0; i < pi_ar.length; i++ ) {
		    pi_ar[i] = (PublishInfo) pis.get(i);
		}
	    }
	    // see if we're root (no need to synchronize, no modification
	    // or iteration. right??
	    HashSet roots = 
		(HashSet) _objlevels.get(new Integer(LOCALROOTOBJS));
	    root = roots.contains(msg.objguid);
	}
	try {
	    msg.sink.enqueue( new TapestryObjPointerSetResp( msg.objguid, 
							     pi_ar, 
							     root ) );
	} catch( SinkException e ) {
	    Carp.die( "Couldn't enqueue ObjPointerSetResp: " + e );
	}
    }
    
    private void handle_routing_table_available (RoutingTable.Available msg) {

	if (DEBUG) Debug.printtagln (tag, "processing RoutingTable.Available " + msg);

	if (_routing_tbl != null)
	  throw new AssertionViolatedException 
	    (tag + ": got two versions of the routing table.");

	_routing_tbl = msg.table;
	_revpointers = msg.revpointers;
	_objlevels = msg.objlevels;
	_self_digits = RoutingTable.digits (_self_guid);

	if (DEBUG)
	  dumpRevpointers(msg.revpointers, 3);

	/*
	if (dynamic_route != null && 
	    dynamic_route.equals("static")) {
	    // Why would you need to send out another one, when you 
	    // just got one???
	  dispatch (new
	    RoutingTable.Available (_routing_tbl, _revpointers, _objlevels));
	} else {
	  //TODO: is this really necessary? reexamine!!
	}	  
	*/
	System.out.println("Tapestry: ready.\n");
	System.out.flush();


	//check, and only execute queued events if _backpointers is there
	if (_backpointers != null) {
	  try {
	    while (! _waiting_reqs.isEmpty ()) {
	      QueueElementIF e = 
		(QueueElementIF) _waiting_reqs.removeFirst ();
	      handleEvent (e);
	    }
	  }
	  catch (EventHandlerException e) {
	    BUG (tag + ": what do I do now?"); // TODO
	  }
	}
    }

    private void handle_key_established (MacHandler.KeyEstablished event) {

	if (DO_MACS) {

	    if (_mac_stage == null) {
		throw new AssertionViolatedException (
			"DO_MACS=true but no MAC stage is running.");
	    }

	    if (MAC_DEBUG) Debug.printtagln (tag, 
		    "processing " + event + ".");

	    if (MAC_DEBUG) Debug.printtagln (tag, "Key established for node " +
		    "with guid " + event.guid + ".");
	    
	    /*
	    LinkedList l = null;
	    synchronized (_msgs_needing_macs) {
		l = (LinkedList) _msgs_needing_macs.remove (event.guid);
	    }
	    while ((l != null) && (! l.isEmpty ())) {
		RouteMsg outb = (RouteMsg) l.removeFirst ();
		QSByteArray mac = null;

		if (MAC_DEBUG) Debug.printtagln (tag, 
			"Calculating MAC for " + outb + ".");

		try {
		    mac = _mac_stage.doMAC (outb.data_to_mac (), event.guid);
		}
		catch (MacHandler.GuidNotEstablished e) {
		    
		    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (
			    "I think we should have been established here.");

		    synchronized (_msgs_needing_macs) {
			l.addFirst (outb);
			LinkedList l2 = 
			    (LinkedList) _msgs_needing_macs.get (event.guid);
			if (l2 != null) {
			    while (! l2.isEmpty ()) 
				l.addLast (l2.removeFirst ());
			}
			_msgs_needing_macs.put (event.guid, l);
			_mac_stage.establish (event.guid);
		    }
		    return;
		}
		catch (QSException e) {
		    Carp.die (tag + ": unreachable 7.");
		}

		outb.mac = mac;

		if (MAC_DEBUG) Debug.printtagln (tag, 
			"MAC calculated for " + outb + ".");

		int [] dest = RoutingTable.digits (outb.dest);
		int [] digit = {0};
		NodeId next = _routing_tbl.route_to_root (dest, true, false, digit);

		if (next == _self_node_id)
		    throw new AssertionViolatedException(tag + ": bug.");

		outb.peer = next;
		outb.hop = digit [0];

		dispatch (outb); 
	    }
	    */
	}
    }

    public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {
	//Added by Ling Huang for Fault tolerance routing.
      if ( MULTICAST == true )
      {
	  if ( isDuplicate(item) )
	      return;
      }
	
      try
      {
	if (DEBUG) Debug.printtagln (tag, "got " + item.getClass () + 
		    " event: " + item + ".");
	
	if (item instanceof StagesInitializedSignal) {
	    handle_stages_init_signal ();
	} else if (item instanceof TapestryIntegrateMsg) {
	    _routing_tbl = null;
	    if( dynamic_route == null || !dynamic_route.equals( "dynamic" ) ) {
		dynamic_route = "dynamic";
		// need to look out for the ObjectMap now that we're dynamic
		try {
		    Filter filter = new Filter ();
		    if (! filter.requireType(
		       Class.forName("ostore.tapestry.impl.Router$ObjectMap")))
			BUG (tag + ": could not require type " + 
			     "ostore.tapestry.impl.Router$ObjectMap");
		    Classifier.getClassifier(_self_node_id).
			subscribe (filter, _sink);
		} catch( ClassNotFoundException cnfe ) {
		    BUG (tag + ": could not require type " + 
			 "ostore.tapestry.impl.Router$ObjectMap. Exception: "
			 + cnfe);
		}
	    }
	} else if (item instanceof RoutingTable.Available) {
	    handle_routing_table_available ((RoutingTable.Available) item);
	}
	else if (item instanceof Router.ObjectMap) {
	    handle_object_map((Router.ObjectMap) item);
	}
	else if (item instanceof RouterFaultMsg) {
	    // update the faults
	    // note that this must be before the code that drops messages
	    _fault_controller.updateFaults( (RouterFaultMsg) item );
	}
	else if (item instanceof BloomWrapper) {
	    handle_bloom_wrapper ((BloomWrapper) item);
	}
	else if (item instanceof MacResponse) {
	    MacResponse resp = (MacResponse) item;
	    if (MAC_DEBUG) Debug.printtagln (tag, "got " + resp);

	    RouteMsg outb = (RouteMsg) resp.id;
	    outb.mac = resp.mac;

	    if (MAC_DEBUG) Debug.printtagln (tag, "MAC added to " + 
		    outb + ".");

	    dispatch (outb);
	}
	else if (_routing_tbl == null || 
		 _backpointers == null) {
	    if (DEBUG) Debug.printtagln (tag, "Not up yet, queuing it."+
					 "tbl="+_routing_tbl+"; backs="+
					 _backpointers);
	    _waiting_reqs.addLast (item);
	    return;
	}
	else {

	    // should we drop this message?
	    if( item instanceof NetworkMessage &&
		_fault_controller.dropMessage(((NetworkMessage)item).peer) ) {
		return;
	    }

	    if (item instanceof DeltaMsg) {
		if( DEBUG ) {
		    Debug.printtagln(tag, "Received DeltaMsg "+(DeltaMsg) item+
				     " calling pwupdate");
		}
		_routing_tbl.pw_update ((DeltaMsg) item);
	    }

	    else if (item instanceof ICMPPingMsg) {
		ICMPPingMsg msg = (ICMPPingMsg)item;
		handle_icmp_ping_msg(msg); //equivalent of handle_tapestry_route_msg
	    }
	    else if (item instanceof TapestryRouteMsg) {
		
		if (((TapestryRouteMsg) item).inbound) {
		    if (item instanceof PingMsg) {
			handle_ping_msg ((PingMsg) item);
		    }
		    else if (item instanceof PongMsg) {
			handle_pong_msg ((PongMsg) item);
		    }
		    else {
			BUG (tag + ": should be a ping or pong");
		    }
		}
		else
		    handle_tapestry_route_msg ((TapestryRouteMsg) item);
	    }
	    else if (item instanceof TapestryPrefixRouteMsg) {
		handle_tapestry_prefix_route_msg ((TapestryPrefixRouteMsg) item);
	    } else if (item instanceof ICMPPingRouteMsg){
	      ICMPPingMsg msg=null;
	      try {
		  msg = (ICMPPingMsg)((ICMPPingRouteMsg)item).user_data();
	      } catch (QSException e) {
		Carp.die("Bad user_data with exception: " + e);
	      }
	      handle_icmp_ping_route_msg(msg);
	      handle_route_msg ((RouteMsg) item);
	      //need to do this to fill in the peer field
	    }
	    else if (item instanceof RouteMsg) {
		handle_route_msg ((RouteMsg) item);
	    }
	    else if (item instanceof TapestryPublishMsg) {
		handle_tapestry_publish_msg ((TapestryPublishMsg) item);
	    }
	    else if (item instanceof TapestryUnpublishMsg) {
		handle_tapestry_unpublish_msg ((TapestryUnpublishMsg) item);
	    } else if (item instanceof RepublishRequestMsg) {
	      handle_republish_request((RepublishRequestMsg) item);

	    } else if (item instanceof PublishMsg) {
		handle_publish_msg ((PublishMsg) item);
	    }
	    /*else if (item instanceof RouteObjTMessage) {
		handle_route_obj_t_message ((RouteObjTMessage) item);
	    }*/
	    else if (item instanceof TapestryObjPointerSetReq) {
		handle_obj_pointer_set_req((TapestryObjPointerSetReq)item);
	    }
	    else if (item instanceof TapestryLocateMsg) {
		handle_tapestry_locate_msg ((TapestryLocateMsg) item);
	    }
	    else if (item instanceof RouteObjRootMsg) {
		handle_route_obj_root_msg ((RouteObjRootMsg) item);
	    }
	    else if (item instanceof RouteObjDestMsg) {
		handle_route_obj_dest_msg ((RouteObjDestMsg) item);
	    }
	    else if (item instanceof NetworkLatencyReq) {
		handle_network_latency_req ((NetworkLatencyReq) item);
	    }
	    else if (item instanceof NetworkLatencyResp) {
		handle_network_latency_resp ((NetworkLatencyResp) item);
	    }
	    else if (item instanceof MacHandler.KeyEstablished) {
		handle_key_established ((MacHandler.KeyEstablished) item);
	    }
	    else if (item instanceof ChannelACKMsg) 
	    {
		//Added by Ling Huang. Router should handel ack msg.
		_channel_handler.handle_msg((ChannelACKMsg)item);
	    }
	    else if (item instanceof RetransmissionCheckMsg) 
	    {
		//Added by Ling Huang. Router should do retransmission
		//check.
		_channel_handler.retransmit();
	    } 
	    else if (item instanceof TapestryDetachConfirm) {
		//the node has just detached. nullify all data structures
		_routing_tbl = null;
		_revpointers = null;
		_objlevels = null;
		_backpointers = null;

	    } else if (DEBUG)
	      Carp.die("ERROR in Router: unrecognized incoming message:"
		       +item);
	}
      }
      catch(Exception e)
      {
        Carp.dieGracefully(e, _self_node_id); 
      }
    }

    // if user_data is null, we will use the remaining four arguments
    // instead.  this is done so that we don't deserialize payload until we
    // reach the destination
  /*************************************************
   * Method to check local pointers for an object, and if
   * found, sends RouteObjDest messages to the actual locations
   * of the objects.
   * @param guid The GUID of the object searched for
   * @param hopInit The # of hops taken to get here from the query source
   * @return Boolean on whether the query is satisfied. If true, it means
   * we're done.
   */
  protected boolean route_to_dest (SecureHash guid, SecureHash src, 
	    TapestryQuery query, TapestryQueryState state, 
	    QuickSerializable user_data, int user_data_length,
	    InputBuffer user_data_buffer, int hopInit, int TTL, int qos) 
    throws QSException {

    if (DEBUG) Debug.printtagln (tag, "called" );
    
    if(DEBUG)
    { 
      if(ByteUtils.memcmp(_match_bytes, 0, guid.bytes(), 0, 6))
	Debug.printtagln(tag, "At time=" + 
			 Clock.current_date(_self_node_id) +
			 " attempting to route guid=" + guid);
    }

    LinkedList l = (LinkedList) _backpointers.get (guid);
    boolean stop = false;
    
    boolean keeprouting = false;
    if (l != null) {
      if (DEBUG) Debug.printtagln (tag, "scanning backpointers " + l +
				   " l.size==" + l.size());
      LinkedList dispatches = new LinkedList ();
      synchronized( l ) {
	  Iterator i = l.iterator ();
	  while (i.hasNext ()) {
	      PublishInfo ii = (PublishInfo) i.next ();
	      if (DEBUG) Debug.printtagln (tag, "checking: " + ii );
	  
	      SecureHash [] matches = null;
	      if (query != null) {
		  if (DEBUG) Debug.printtagln (tag, "running query: " + 
					       query );
		  TapestryQueryResult qr = 
		      query.query (_self_guid, ii.src, ii.tag, state);
		  if (DEBUG) Debug.printtagln (tag, "result: " + qr );
		  if (qr instanceof TapestryQueryResultMatch) {
		      TapestryQueryResultMatch m = 
			  (TapestryQueryResultMatch) qr;
		      matches = m.matches ();
		      stop = m.stop ();
		  } else 
		      stop = false;
	      } else {
		  if (DEBUG) Debug.printtagln (tag, "no query" );
		  matches = new SecureHash [1];
		  matches [0] = ii.src;
		  stop = true;
	      }
	  
	      if(DEBUG) {
		  if(ByteUtils.memcmp(_match_bytes, 0, guid.bytes(), 0, 6))
		      {
			  String str = new String();
			  str += "{";
			  for(int kk = 0; matches != null && 
				  kk < matches.length; kk++)
			      str += matches[kk] 
				  + (kk!=matches.length-1?", ":"");
			  str += "}";
			  Debug.printtagln(tag, "At time=" + 
					   Clock.current_date(_self_node_id) +
					   " attempting to route guid=" + guid
					   + " matches=" + str + " stop=" + 
					   stop);
		      }
	      }
	      if (DEBUG) Debug.printtagln (tag, "matches: " + matches );
	      if( matches != null ) {
		  boolean foundmatch = false;
		  for (int j = 0; j < matches.length; ++j) {
		  
		      if (DEBUG) Debug.printtagln (tag, "matches[" + j +
						   "]: " + matches[j] );
		      int [] dest = RoutingTable.digits (matches [j]);
		      int [] digit = {0};
		  
		      NodeId next = _routing_tbl.route_to_root(dest, true, 
							       false, digit);
		      if (DEBUG) Debug.printtagln (tag, "next: " + next );
		  
		      if (next == _self_node_id) {
			  foundmatch = true;
			  if (DEBUG) Debug.printtagln (tag, "next is self" );
		      
			  QuickSerializable event = user_data;
			  if (event == null) {
			      try {
				  // Since I may hand this user_data_buffer 
				  // to more than one RouteObjDestMsg (if I 
				  // follow two backpointers, for example), 
				  // I need to clone it before passing it
				  // on.  Otherwise, once one of them reads 
				  // the data out of it, the other's buffer 
				  // will be empty, too.
				  event = 
				      ((InputBuffer)user_data_buffer.clone ()).
				      nextObject ();
			      }
			      catch (CloneNotSupportedException e) {
				  throw new AssertionViolatedException 
				      ("not cloneable");
			      }
			  }
		      
			  TapestryLocateMsg r = (TapestryLocateMsg) event;
			  r.peer = src;
			  r.inbound = true;
			  r.guid = guid;
			  r.query = null;
			  r.query_state = null;
			  r.hopCount = hopInit;
			  r.TTL = TTL;
			  if (DEBUG) Debug.printtagln (tag, 
						       "event==" + event +
						       " r==" + r);
			  dispatches.addLast (event);
		      } else if (next != null) {
			  RouteObjDestMsg outb = null;
		      
			  if (digit[0] > 20)
			      { digit[0] = 0; }
		      
			  //here we start to send out messages to where objects
			  //are stored, so if TTLs haven't taken into effect, 
			  //they are ignored from now on.
			  if (user_data_buffer == null) {
			      // if we're running under the simulator,
			      // the message has never been serialized,
			      // so we have to use user_data instead of
			      // the _user_data_buffer
			  
			      outb = new RouteObjDestMsg (next, src, ii.src,
							  guid, digit [0], 
							  user_data);
			  }
			  else {
			      try {
				  // Since I may hand this user_data_buffer 
				  // to more than one RouteObjDestMsg (if I 
				  // follow two backpointers, for example), 
				  // I need to clone it before passing it
				  // on.  Otherwise, once one of them reads 
				  // the data out of it, the other's buffer 
				  // will be empty, too.
				  outb = new RouteObjDestMsg (next, src, 
							      ii.src, guid, 
							      digit [0], 
							      user_data_length,
				      (InputBuffer) user_data_buffer.clone ());
			      }
			      catch (CloneNotSupportedException e) {
				  throw new AssertionViolatedException 
				      ("must be cloneable");
			      }
			  }
		      
			  //Added by Ling Huang: handle reliable message.
			  outb.set_qos(qos);
		      
			  outb.hopInit = hopInit;
			  dispatches.addLast (outb);
		      } else {
			  if( DEBUG ) {
			      Debug.printtagln( tag, "next is null, " + 
						"routing table: " +
						_routing_tbl );
			  }
			  //inconsistent routing table: we ignore this and 
			  //make sure we keep routing
			  keeprouting = true;
			  //Carp.die("Inconsistent routing table: " + 
			  //"obj location not found in table.");
		      }
		  }
	      }
	      if (stop) 
		  break;
	  }
      }

      while (! dispatches.isEmpty ()) {
	QueueElementIF d = (QueueElementIF) dispatches.removeFirst ();
	if (DEBUG) Debug.printtagln (tag, "dispatching: "+ d );
	dispatch (d);
      }
    } else {
      if (DEBUG) Debug.printtagln (tag, "no backpointers exist" );
    }
    if (DEBUG) Debug.printtagln (tag, "done (" + (stop?"stop":"go") + ")");
    return (stop || keeprouting);
  }

  /**
   * Compare two guids and return the index (starting with 0) of the
   * first differing digit.
   */
  public static int guidCompare(SecureHash one, SecureHash two) {
    int [] onedigits = RoutingTable.digits(one);
    int [] twodigits = RoutingTable.digits(two);
    int i;
    for (i=0; i<RoutingTable.DIGITS_PER_GUID; i++) {
      if (onedigits[i] != twodigits[i]) break;
    }
    if (i == RoutingTable.DIGITS_PER_GUID) i = LOCALROOTOBJS; //=-1
    return i;
  }

    protected void publish (SecureHash src, SecureHash guid,
	    TapestryTag ttag, boolean publish, NodeId lasthop, int qos) {
	// no local opts
	publish( src, guid, ttag, publish, lasthop, qos, 0, 0 );
    }

    protected void publish (SecureHash src, SecureHash guid,
	    TapestryTag ttag, boolean publish, NodeId lasthop, int qos,
	    int local_opts, int local_opts_ttl) {

	// Store a backpointer here, initially at infinite distance.
	PublishInfo pi = new PublishInfo (src, ttag, lasthop);
	if (src.equals (_self_guid))
	  pi.distance = 0;

	LinkedList l = null;
	synchronized( _backpointers ) {
	    l = (LinkedList) _backpointers.get (guid);
	    if (l == null) {
		l = new LinkedList ();
		if (publish) {
		    _backpointers.put (guid, l);
		}
		else if (DEBUG)
		    Debug.printtagln(tag, "Unpublishmsg received for " +
				     "non-existent object pointer. " + 
				     "Will forward on.");
	    }
	}
	if(DEBUG)
	  Debug.printtagln(tag, (publish ? "adding publishInfo==" :
				 "removing publishInfo==") + pi + 
			   " for guid==" + guid + 
			   " to backpointers set==" + l + " at time " + 
			   Clock.current_date(_self_node_id).getTime() + 
			   " BEFORE."); 
	synchronized (l) {
	  if (publish) {
	      if( !l.contains(pi) ) {
		  if (pi.distance == 0)
		      l.addFirst (pi);
		  else
		      l.addLast (pi);
	      }
	  } else {
	    if (l.contains(pi)) {
	      l.remove(pi);
	      if (DEBUG)
		Debug.printtagln(tag, "Removing publishinfo...");
	    }
	  }
	}
	if (publish && pi.distance == Long.MAX_VALUE) {
	    // Start looking for the real distance to the publisher.
	    synchronized (_pending_back_pointer_times) {
		LinkedList s = (LinkedList) 
		    _pending_back_pointer_times.get (src);
		if (s == null) {
		    s = new LinkedList ();
		    _pending_back_pointer_times.put (src, s);
		    dispatch (new NetworkLatencyReq (src));
		}
		s.addLast (new Pair (pi, l));
	    }
	} else if (!publish && pi!=null && !lasthop.equals(pi.lasthop)) {
	  //if we're unpublishing, check to see if we need to backtrack
	  PublishMsg back = new PublishMsg(pi.lasthop, src, guid,
					   ttag, 0, false);

	  // Do we need to do optimizations here?
	  // perform any unpublish backtracking optimizations
	  if( local_opts_ttl > 0 ) {
	      // keep the local opts going
	      back.local_opts = local_opts;
	      back.local_opts_ttl = local_opts_ttl;
	  }

	  //Added by Ling Huang: handle reliable message.
	  back.set_qos(qos);

	  dispatch(back);
	}

	// Route the message on towards the root.
	int [] dest = RoutingTable.digits (guid);
	int [] digit = {0};
	NodeId next = _routing_tbl.route_to_root (dest, false, false, digit);

	Integer i;
	if (next != null && ! next.equals (_self_node_id)) {
	    PublishMsg outb = new PublishMsg (next, src, guid, ttag, 
		    digit [0], publish);

	    //Added by Ling Huang: handle reliable message.
	    outb.set_qos(qos);

	    local_opts_ttl--;
	    // keep the local opts going
	    outb.local_opts = local_opts;
	    outb.local_opts_ttl = local_opts_ttl;

	    perform_local_publish_opts( outb, dest, lasthop,
					guidCompare(guid, _self_guid ) );

	    // if the node we're publishing to is leaving, we also need to
	    // publish to the backup for this node.  that can either be the
	    // backup, or just loop around the same level and see if we find
	    // someone else.  given good knowledge about the network, this
	    // should be right.
	    if( DEBUG ) {
		Debug.printtagln( tag, "Looking up guid for " + next );
	    }
	    int [] next_digits = 
		RoutingTable.digits( _routing_tbl.node_id_to_guid( next ) );
	    RouteEntry re = _routing_tbl.getEntry( digit[0], 
						   next_digits[ digit[0] ] );
	    if( re == null || re.getFRLS() == null ) {
		Carp.die( "RouteEntry is null when looking for " + next +
			  " in the Routing Table: " + 
			  _routing_tbl.toShortString() );
	    }
	    if( !re.getFRLS().equals( next ) ) {
		Carp.die( next + 
			  " not in RouteEntry where we expect it to be." );
	    }
	    // if the next node is leaving, we'll try to find a replacement
	    if( publish && local_opts != -1 && 
		re.isLeavingAtIndex( re.get_frls() ) ) {

		NodeId backup = null;
		int val = next_digits[ digit[0] ];
		if( re.size() > 1 ) {
		    backup = re.getAtIndex(1);
		} else {
		    int z = (val == _routing_tbl.digit_values()-1) ? 0 :val+1;
		    while (_routing_tbl.getEntry(digit[0], z) == null ||
			 _routing_tbl.getEntry(digit[0], z).getFirst() == null)
			   z = (z == _routing_tbl.digit_values() - 1) ? 0 :z+1;
		    if (z != val)
			backup = _routing_tbl.getEntry(digit[0], z).getFirst();
		}
		if( backup != null ) {
		    PublishMsg clonemsg = null;
		    try {
			clonemsg = (PublishMsg) outb.clone();
		    } catch( CloneNotSupportedException cnse ) {
			Carp.die( "No clone! " + cnse );
		    }
		    clonemsg.peer = backup;
		    if( DEBUG ) {
			Debug.printtagln( tag, "Sending clone of publish msg ("
					  + outb + ") to " + backup + 
					  " because " + next + " is marked as "
					  + "leaving" );
		    }
		    dispatch( clonemsg );
		}
	    }

	    // don't route towards the root if this message was the result
	    // of a local_optimization
	    if( local_opts != -1 ) {
		dispatch (outb);
	    }

	    // Be careful with setting this value, since it later
	    // determines what replacement is suggested for an object
	    // when this node detaches
	    //i = new Integer(guidCompare(_self_guid, 
	    // 			_routing_tbl.node_id_to_guid(next)));
	    i = new Integer(digit[0]);
	} else {
	  i = new Integer(LOCALROOTOBJS);
	  if(DEBUG) Debug.printtagln(tag, "NOT forwarding" +
				     " because we are _root_ for" +
				     " guid=" + guid +
				     " tag=" + ttag + 
				     " src=" + src +
				     " publish=" + publish);
	  // we're the root, but maybe we want to publish to our surrogate
	  // for reasons of fault tolerance
	  if( _publish_to_root_surrogate ) {
	      NodeId surr = _routing_tbl.nextSurrogate( null, _self_guid );
	      // TODO: is digit[0] really the right thing to use here?
	      // should it be LOCALROOTOBJS? or something else?
	      PublishMsg outb = new PublishMsg (surr, src, guid, ttag, 
						digit [0], publish);
	      // make sure the next guy knows he shouldn't forward this along.
	      outb.local_opts = -1;
	      outb.local_opts_ttl = 0;
	      if( DEBUG ) Debug.printtagln( tag, "Sending duplicate publish " 
					    + "message to root\'s surrogate " 
					    + surr + " --  " + outb );
	      dispatch(outb);
	  }

	}

	modify_objlevels( guid, i, publish );

	// if we're not the root, we may need to put this object in other 
	// levels as well (due to the strangeness of surrogate routing
	if( i.intValue() != -1 ) {
	    int j = guidCompare(_self_guid, guid );
	    if( j != i.intValue() ) {
		for( ; j < i.intValue(); j++ ) {
		    modify_objlevels( guid, new Integer(j), publish );
		}
	    }
	}

    }

    /**
     * Put in or remove a guid form a certain level of the objlevels 
     * data structure
     */
    private void modify_objlevels( SecureHash guid, Integer level, 
				   boolean publish ) {
	dispatch( new ModifyObjLevelsMsg( guid, level, publish ) );
    }

    private void perform_local_publish_opts( PublishMsg pubmsg, int[] dest, 
					     NodeId lasthop, int myhop ) {

	// if the local opts is -1, must return now because that
	// matches all the flags
	if( pubmsg.local_opts == -1 ) {
	    return;
	}

	///////////////////////////////////////
	// LOCAL_OPT #1: publish along backups 
	///////////////////////////////////////
	if( pubmsg.local_opts_ttl > -1 && 
	    (pubmsg.local_opts & LOCAL_OPT_PUBLISH_BACKUP) != 0 ) {
	    
	    // get the backup of the next hop, if there is one
	    RouteEntry re = _routing_tbl.getEntry( pubmsg.hop,
						   dest[pubmsg.hop] );
	    if( re == null || re.getFirst() == null || 
		!re.getFirst().equals( pubmsg.peer ) ) {

		// This indicates we used a surrogate
		// TODO:  Should we publish along surrogate's backup?
		if(DEBUG) { 
		    Debug.printtagln( tag, "Although we're forwarding a " + 
				      "PublishMsg to " + pubmsg.peer + 
				      " for guid " + pubmsg.guid + " at hop " 
				      + pubmsg.hop + ", the RouteEntry in " + 
				      "that spot doesn't match: " + re + 
				      ", probably indicates a surrogate.  " +
				      "Ignoring Local opt #1." );
		}
		return;
	    }
	    
	    // if there's a backup
	    for( int i = 1; i < re.size() && i <= LOCAL_OPT_BACKUPS; i++ ) {
		// get secondary link (primary backup)
		NodeId backup = re.getAtIndex(i);
		if( DEBUG ) {
		    Debug.printtagln( tag + "LOCAL_OPT_PUBLISH_BACKUP",
				      "sending publish to backup link "
				      + backup );
		}
		
		// clone the message and send it yourself
		PublishMsg backup_msg = null;
		try {
		    backup_msg = (PublishMsg) pubmsg.clone();
		} catch( CloneNotSupportedException cnse ) {
		    Carp.die( "Can't clone PublishMsg: " + cnse );
		}
		backup_msg.peer = backup;

		// perform other optimizations on this backup_msg, so recurse
		// backup_msg.local_opts = 
		//    backup_msg.local_opts^LOCAL_OPT_PUBLISH_BACKUP; //xor
		// perform_local_publish_opts( backup_msg, dest, lasthop );
		// don't allow backups to perform opt or publish further
		backup_msg.local_opts_ttl = 0;
		backup_msg.local_opts = -1;
		// dispatch
		dispatch( backup_msg );
	    }
	    
	} // end LOCAL_OPT #1

	///////////////////////////////////////
	// LOCAL_OPT #2: publish to nearest N neighbors on this level
	///////////////////////////////////////
	if( pubmsg.local_opts_ttl > -1 && 
	    (pubmsg.local_opts & LOCAL_OPT_PUBLISH_ALL_LINKS) != 0 ) {

	    // find LOCAL_OPT_NEAREST neighbors (including backups)
	    PriorityQueue neighbors = new PriorityQueue( LOCAL_OPT_NEAREST );
	    for( int j = 0; j < _routing_tbl.digit_values(); j++ ) {
		RouteEntry re = _routing_tbl.getEntry( pubmsg.hop, j );
		if( re != null ) {
		    for( int k = 0; k < re.size(); k++ ) {

			NodeId neighbor = re.getAtIndex(k);
			if( !neighbor.equals(_self_node_id) &&
			    !neighbor.equals(pubmsg.peer) ) {

			    long distance = 
				(_routing_tbl.node_to_latency(neighbor))
				                                  .longValue();
			    neighbors.insert( neighbor, distance, false );
					      
			}
		    }
		}

	    }

	    // now publish to the first LOCAL_OPT_NEAREST of those
	    for(int i = 0; i<LOCAL_OPT_NEAREST && i<neighbors.size(); i++) {

		NodeId neighbor = (NodeId) neighbors.extract( false );
		
		// clone the message and send it yourself
		PublishMsg neighbor_msg = null;
		try {
		    neighbor_msg = (PublishMsg) pubmsg.clone();
		} catch( CloneNotSupportedException cnse ) {
		    Carp.die( "Can't clone PublishMsg: " + cnse );
		}
		neighbor_msg.peer = neighbor;

		// don't allow backups to perform opt or publish further
		neighbor_msg.local_opts_ttl = 0;
		neighbor_msg.local_opts = -1;
		// dispatch
		dispatch( neighbor_msg );

	    }

	} // end LOCAL_OPT #2

	///////////////////////////////////////
	// LOCAL_OPT #3: misroute publishes before taking large hops 
	///////////////////////////////////////
	if( (pubmsg.local_opts & LOCAL_OPT_PUBLISH_LOCAL_MISROUTE) != 0 ) {


	    NodeId surr = find_misroute_node( pubmsg.peer, pubmsg.hop,
					      dest[pubmsg.hop], lasthop, dest);
	    
	    // we found a surrogate, so route a publish to it
	    if( surr != null ) {
		
		// clone the message and send it yourself
		PublishMsg misroute_msg = null;
		try {
		    misroute_msg = (PublishMsg) pubmsg.clone();
		} catch( CloneNotSupportedException cnse ) {
		    Carp.die( "Can't clone PublishMsg: " + cnse );
		}
		misroute_msg.peer = surr;
		// don't allow backups to perform opt or publish further
		misroute_msg.local_opts_ttl = 0;
		misroute_msg.local_opts = -1;
		misroute_msg.hop = myhop;
		// dispatch
		dispatch( misroute_msg );
		
	    }
	    
	} // end LOCAL_OPT #3

    }

    /**
     * Returns whether or not the Router should send along the original
     * locate message or not.
     */
    private boolean perform_local_locate_opts( RouteObjRootMsg rootmsg, 
					       int[] dest, NodeId lasthop, 
					       int myhop ) {

	boolean do_orig_route = true;
	///////////////////////////////////////
	// LOCAL_OPT #3: misroute locate before taking large hops 
	///////////////////////////////////////
	if( (rootmsg.local_opts & LOCAL_OPT_PUBLISH_LOCAL_MISROUTE) != 0 ) {


	    NodeId surr = find_misroute_node( rootmsg.peer, rootmsg.hop,
					      dest[rootmsg.hop], lasthop, 
					      dest);
	    
	    // we found a surrogate, so route a locate to it
	    if( surr != null ) {
		
		// clone the message and send it yourself
		RouteObjRootMsg misroute_msg = null;
		try {
		    misroute_msg = (RouteObjRootMsg) rootmsg.clone();
		} catch( CloneNotSupportedException cnse ) {
		    Carp.die( "Can't clone RouteObjRootMsg: " + cnse );
		}
		misroute_msg.peer = surr;
		// don't set ttl to 0 -- allow further optimization!
		misroute_msg.hop = myhop;
		// dispatch
		dispatch( misroute_msg );
		do_orig_route = false;
	    }
	    
	}

	return do_orig_route;

    }

    private NodeId find_misroute_node(NodeId next, int digit, int value, 
				      NodeId lasthop, int[] dest ) {

	// if this is going to be a large hop, pretend
	// this guy doesn't exist, and publish to someone closer	    
	
	// find the latency to the next guy we're publishing to.
	Double next_lat = _routing_tbl.node_to_latency( next );
	if( next_lat == null ) {
	    Carp.die( "1) No latency stored for " + next );
	}
	
	// check the latency of the last hop
	Double last_lat = _routing_tbl.node_to_latency( lasthop );
	if( last_lat == null ) {
	    if( DEBUG ) {
		Debug.printtagln( tag + "LOCAL_OPT_LOCAL_MISROUTE",
				  "2) No latency stored for last hop, " +
				  lasthop + ", using row 0 latency" );
	    }
	    // guess that your first-level neighbors should have a local
	    // latency
	    boolean set = false;
	    for( int i = 0; i < _routing_tbl.digit_values(); i++ ) {
		RouteEntry re = _routing_tbl.getEntry( 0, i );
		if( re != null && re.getFirst() != null && 
		    !re.getFirst().equals(_self_node_id)) {
		    Double possible_last_lat = 
			_routing_tbl.node_to_latency( re.getFirst() );
		    if( possible_last_lat == null ) {
			Carp.die( "3) No latency stored for " + re.getFirst() );
		    }
		    // find the smallest one
		    if( last_lat == null ||
			possible_last_lat.compareTo( last_lat ) < 0 ) {
			last_lat = possible_last_lat;
			set = true;
		    }
		}
	    }
	    // if we haven't set it for some reason, just use the same node
	    if( !set ) {
		last_lat = next_lat;
	    }
	}	    
	
	// if the next latency is more than 5 times the last latency, 
	// try misrouting.
	if( !areSameDomain( _self_node_id, next, next_lat, last_lat ) ) {
	    
	    // see if the would-be surrogate is closer
	    // to find the surrogate, look to the right circularly

	    // find the surrogate with the same relationship 
	    // (on the same level) as the current node, wrt the guid
	    int surr_level;
	    int[] my_digits = RoutingTable.digits(_self_guid);
	    for (surr_level=0; surr_level<RoutingTable.DIGITS_PER_GUID; 
		 surr_level++) {
		if (dest[surr_level] != my_digits[surr_level]) break;
	    }
	    // this can't match our guid, cause otherwise we couldn't
	    // have gotten here (we'd have been the root)

	    NodeId surr = null;
	    Double surr_lat = null;
	    // loop until we find a surrogate that's not us
	    // (Don't worry, we break this loop 
	    //  if we really are the intra-node surrogate)
	    while( surr == null || surr.equals( _self_node_id ) ) {

		boolean around = false;

		// start one to the right of the objects value for this digit
		int digit_value = dest[surr_level]; 
		if( digit_value+1 == _routing_tbl.digit_values() ) {
		    digit_value = -1;
		}
		
		for( int j = digit_value+1; 
		     ((!around && j < _routing_tbl.digit_values()) ||
		      (around && j < digit_value)); j++ ) {
		    
		    RouteEntry re = _routing_tbl.getEntry( surr_level, j );
		    if( re != null && re.getFirst() != null ) {
			
			// check to see if the primary neighbor is closer
			// than the threshold
			surr_lat = 
			    _routing_tbl.node_to_latency( re.getFirst() );
			if( surr_lat == null ) {
			    Carp.die( "4) No latency stored for " + 
				      re.getFirst() );
			}
			if( areSameDomain( _self_node_id, re.getFirst(), 
					   surr_lat, last_lat ) ) {
			    // this node's the one
			    surr = re.getFirst();
			    break;
			}
			
		    }
		    
		    // think circularly
		    if( j+1 == _routing_tbl.digit_values() ) {
			j = -1;
			around = true;
		    }
		
		}

		// get ready to move up a level, if this node was us
		surr_level++;
		
		if( surr_level == _routing_tbl.digits_per_guid() ) {
		    // we've reached the end . . . 
		    // we must be the intra-node surrogate
		    surr = null;
		    break;
		}

	    }
	    
	    if( surr != null ) {

		if( DEBUG ) {
		    Debug.printtagln( tag + 
				      "LOCAL_OPT_LOCAL_MISROUTE",
				      "Misrouting to " + surr +
				      " with lat " + surr_lat + 
				      ", rather than to " + next +
				      " with lat " + next_lat + 
				      " (last lat was " + last_lat + ")");
		}
		
		return surr;
	    }

	}

	return null;

    }

    private boolean areSameDomain( NodeId one, NodeId two, Double next_lat,
				   Double last_lat ) {

	// the simulator knows for sure (but only use that knowledge if
	// directed to do so by setting thresh to 0)
	if( _simulator_running && LOCAL_OPT_THRESHOLD_FACTOR == 0 ) {

	    return soss.network.Network.instance().areSameDomain( one, two );

	} else {

	    // figure it out via the threshold factor
	    return next_lat.doubleValue() <= 
		LOCAL_OPT_THRESHOLD_FACTOR*last_lat.doubleValue();

	}

    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for (int i=0; i < items.length; i++)
	    handleEvent (items[i]);
    }


    /**
     * A <code>Comparator</code> used to compare the enqueue time
     * for message in multicast queue.
     * <p>
     */
    public class MsgTime {
	public SecureHash msg_id;
	public long time;

	public MsgTime(SecureHash id, long t)
	{
	    msg_id = id;
	    time = t;
	}
    }

    private static class EnqueueTimeComparator implements Comparator
    {
	//////////////////// 	  METHODS	////////////////////
	/**
	 * Defines an ordering among <code>VersionInfo</code>s and
	 * <code>Date</code>s based on creation dates.
	 */
	public int compare( Object o1, Object o2 )
	{
	    long a, b;
	    
	    if ( (o1 instanceof MsgTime) && o2 instanceof MsgTime) 
	    {
		a = (int)(( (MsgTime)o1 ).time);
		b = (int)(( (MsgTime)o2 ).time);
		if (a > b)
		    return 1;
		else if (a == b)
		    return 0;
		else
		    return -1;
	    }
	    else
		return -1;
	}

	/**
	 * Returns <code>true</code> for any
	 * <code>CreationDateComparator</code>.
	 */
	public boolean equals( Object obj )
	{
	    return( obj instanceof EnqueueTimeComparator );
	}
    }


  public static class ObjectMap implements QueueElementIF {
    public ObjectMap (Map backpointers) {
      this.objmap = backpointers;
    }

    public Map objmap;
  }

    private boolean _drop_all;
    private HashSet _multicast_msgs;
    private TreeSet _sorted_multicast_msgs;
    private HashMap localobjects;
    private String dynamic_route;
    private NodeId _self_node_id;
    private SecureHash _self_guid;
    private int [] _self_digits;
    private SinkIF _sink;
    private LinkedList _waiting_reqs;
    private Map _backpointers; //object location pointers, not rev routes
    private HashMap _revpointers; //pointers to those pointing at us
    private HashMap _objlevels; //map of levels to objects
    private QSPublicKey _pkey;
    private boolean _simulator_running;
    // publish to the alternate root?  config option.
    private boolean _publish_to_root_surrogate;

    private Map _network_latencies;
    private Map _pings_in_flight;
    private Map _pending_back_pointer_times;

    private RoutingTable _routing_tbl;
    //private Map _msgs_needing_macs;
    private boolean _using_bloom;

    // fault controlling mechanisms
    private FaultController _fault_controller;

    // Added by Ling Huang for Channel operation: 06/28/2002.
    private ChannelHandler    _channel_handler;
    private HMACDHHandler _mac_stage;

    public NodeId get_self_node_id()        {    return _self_node_id;    }
    public SecureHash get_self_guid()       {    return _self_guid;       }
    public SinkIF get_sink()                {    return _sink;            }
    public ChannelHandler get_channel_handler() {    return _channel_handler; }
    public RoutingTable get_routing_table() {    return _routing_tbl; }
}










