/**
 *  ShuttleIO.java
 *
 *  @author     Calvin L. Hubble
 *  @version    $Id: ShuttleIO.java,v 1.9 2003/02/19 10:32:31 hweather Exp $
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl.Shuttle;
import ostore.tapestry.impl.*;
import java.util.*;
import java.io.*;
import java.security.*;
import ostore.util.*;
import ostore.dispatch.*;
import ostore.tapestry.*;
import ostore.tapestry.api.*;
import ostore.dataobj.*;
import ostore.update.*;
import ostore.security.*;
import ostore.security.thresh.*;
import ostore.startrek.*;
import seda.sandStorm.api.*;
import seda.sandStorm.core.*;
import seda.sandStorm.internal.SinkProxy;
import cryptix.jce.provider.random.DevRandom;


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class ShuttleIO extends JFrame
    implements EventHandlerIF, ProfilableIF, ActionListener {

    public String class_tag = "ShuttleIO";
 
    JTextField textField;
    JTextArea textArea;
    String newline = "\n";
    JTextArea contactTextArea;
    

    public ShuttleIO() {
	super("Shuttle");
	textField = new JTextField(20);
        textField.addActionListener(this);

	contactTextArea = new JTextArea(20, 10);
	contactTextArea.setEditable(false);
	JScrollPane contactScrollPane = new JScrollPane( contactTextArea,
							 JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
							 JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);

	

        textArea = new JTextArea(5, 20);
        textArea.setEditable(false);
        JScrollPane scrollPane = new JScrollPane(textArea,
                                       JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                                       JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);

        //Add Components to the Applet. 
        GridBagLayout gridBag = new GridBagLayout();
        Container contentPane = getContentPane();
        contentPane.setLayout(gridBag);
        GridBagConstraints c = new GridBagConstraints();
        



	c.gridwidth = GridBagConstraints.REMAINDER;

	c.gridheight = 1;
        c.fill = GridBagConstraints.HORIZONTAL;
        gridBag.setConstraints(textField, c);
        contentPane.add(textField);

        c.fill = GridBagConstraints.BOTH;
        c.weightx = 1.0;
        c.weighty = 1.0;
        gridBag.setConstraints(scrollPane, c);
        contentPane.add(scrollPane);

	c.fill = GridBagConstraints.VERTICAL;
	c.gridheight = 2;
	c.weighty = 1.0;
	c.weightx = 1.0;
	gridBag.setConstraints(contactTextArea, c);
	contentPane.add(contactTextArea);

	pack();
	setVisible(true);
    }


    public void actionPerformed(ActionEvent evt) {
        String text = textField.getText();
        
	//textArea.append(text + newline);
        
	textField.selectAll();

	HandleInput(text);

    }

    private void dispatch (QueueElementIF item) {
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    Carp.out ("Could not dispatch " + item);
	    System.exit (1);
	}
    }

    public void handleEvents(QueueElementIF items[]) 
	throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    public void printtagln(String tag, String msg) {
	System.err.print(tag+":"+msg+"\n");
    }

    

    public void init (ConfigDataIF config) throws Exception {
	
	DEBUG = true;
	
	Types.registerAllTypes ();

	_classifier = Classifier.getClassifier (
                new NodeId (config.getString ("node_id")));
	
	//initialize the numerical arguments
	//_oldobjs = Integer.parseInt(config.getString ("oldobjs"));
	//_numid = config.getString ("nameindex");
	//_id = Integer.parseInt(_numid);

	_self_node_id = new NodeId (config.getString ("node_id"));

	class_tag = class_tag + ".["+_self_node_id.port()+"]";
	if (DEBUG) ostore.util.Debug.turnOn(class_tag);
	
	final String tag = class_tag +".init";
	String on = config.getString("power");
	if (on != null && on.equalsIgnoreCase("off")) return;

	//generate GUIDs for all clients
	/*SecureHash guidx;
	  _guid_located = new HashMap();
	  for (int i=0; i<_oldobjs; i++) {
	  guidx = new SHA1Hash ((_numid+i).getBytes());
	  _guid_located.put(guidx, Boolean.FALSE);
	  } */

	if (DEBUG) printtagln(tag, "Done initializing guid hashmaps");

	String stageName = config.getStage().getName();
	try 
	{
	  if(config.getBoolean("Profile"))
	  {
	    config.getManager().getProfiler().add("ostore.tapestry.impl.ShuttleIO", this);
	    SinkIF sink =
	      config.getManager().getStage(stageName).getSink();
	    this_sink = new SinkProxy(sink,
				      config.getManager(),
				      config.getStage().getWrapper());
	  }
	  else
	    this_sink =
	      config.getManager().getStage(stageName).getSink();
	} 
	catch (NoSuchStageException e) 
	{
	  Carp.die (tag + ": could not get own sink.");
	}

	Filter filter;
	String [] event_types = {
	    "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.tapestry.impl.TapestryReadyMsg",
	    "ostore.tapestry.impl.Shuttle.ShuttleIOToMsg",	
	};	  

	for (int i=0; i< event_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(event_types[i]))) {
	    Carp.out ("Could not require type " + event_types[i]);
	    System.exit (1);
	  }
	  _classifier.subscribe (filter, this_sink);
	}

	//Only listen to messgae To the shuttleIO
	String [] msg_types = {
	   
	    //"ostore.tapestry.impl.TestLocateMsg"
	    //"ostore.tapestry.impl.ShuttleIOToMsg"
	};

	for (int i=0; i< msg_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(msg_types[i])))
	    Carp.die ("Could not require type " + msg_types[i]);
	  if (! filter.requireValue("inbound", new Boolean(true)))
	    Carp.die ("Could not require inbound = true for "+
		      msg_types[i]);
	  _classifier.subscribe (filter, this_sink);
	}

	//System.out.println("[ShuttleIO]: Tapestry stage initialized!!");
    }



    public void handleEvent(QueueElementIF item) 
	throws EventHandlerException {
	
	DEBUG = true;
	final String tag = class_tag+".handleEvent";
	//System.out.println("[SHUTTLEIO] HANDLE EVENT!!!");
	if (item instanceof StagesInitializedSignal) {
	    
	    // do nothing.

	} else if (item instanceof TapestryReadyMsg) {
	    //	    System.out.println("[SHUTTLEIO]: Tapestry Initialized");
	    _client_guid = ((TapestryReadyMsg)item).node_guid;
	    if(inputThread == null)
		inputThread = new Thread(new ShuttleIOInputThread(this));
	    /*NOTE: When i fire off this thread, the event handler does NOT receive events ..
	      Ask Ben about this one!!
	      For now ... i have placed a copy of HandleShuttleIOToMsg in the Shuttle java*/


	    //	    inputThread.run();

	    SecureHash objguid;
	    TapestryPublishMsg msg;
	   
	    
	} else if (item instanceof ShuttleIOToMsg) {
	    //	    System.out.println("[ShuttleIO]: Received as TO Message");
	    HandleShuttleIOToMsg( (ShuttleIOToMsg) item );
	    
	} else {
	    System.out.println("UNKNOWN MESSAGE");
	    Carp.die ("ostore.tapestry.impl.Shuttle.ShuttleIO: Received unknown "
		      + "QueueElementIF: " + item);
	}
    } 
    
    public void HandleShuttleIOToMsg(ShuttleIOToMsg msg) {
	
	String newText;
	//Do the actual Sending and Handling Data here
	System.out.print("Received msg of type "+msg.msgType);
	System.out.println(" String: " + msg.msg);

	
	
	switch (msg.msgType) {
	case(ShuttleIOToMsg.SHUTTLEIO_WHOAMI) :
	    //System.out.println("Current User: "+msg.msg);
	    textArea.append("Current User: "+msg.msg+"\n");
	    break;

	case(ShuttleIOToMsg.SHUTTLEIO_ONLINECONTACTLIST) :
	    //System.out.println("Your Contact List: ");	  
	    newText = new String("Online Contacts: \n");	   
	    for(int i=0; i<msg.contactList.length; i++) {
		//System.out.println(msg.contactList[i]);
		newText += msg.contactList[i];
		newText += "\n";
	    }
	    contactTextArea.setText(newText);
	    break;

	case(ShuttleIOToMsg.SHUTTLEIO_OFFLINECONTACTLIST) :
	    System.out.println("Your Offline contact list: ");
	    newText = new String("Offline Contacts: \n");
	    for(int i=0; i<msg.contactList.length; i++) {
		//System.out.println(msg.contactList[i]);
		newText += msg.contactList[i];
		newText += "\n";
	    }
	    contactTextArea.setText(newText);
	    break;

	case(ShuttleIOToMsg.SHUTTLEIO_ALLCONTACTLIST) :
	    System.out.println("List is "+msg.contactList.length+" long");
	    System.out.println("Your total contact list: ");
	    newText = new String("All Contacts\n");
	    for(int i=0; i<msg.contactList.length; i++) {
		//System.out.println(msg.contactList[i]);
		newText += msg.contactList[i];
		newText += "\n";
	    }	    
	    contactTextArea.setText(newText);
	    break;
	case(ShuttleIOToMsg.SHUTTLEIO_RECEIVEDTEXTMESSAGE):
	    //System.out.println("<"+msg.sender+"> : "+msg.msg);
	    newText = new String("<"+msg.sender+"> : "+msg.msg+"\n");
	    textArea.append(newText);
	    break;
	}
	
	
    }
    
    /* Handles input from standard in*/
    public void HandleInput(String s)
    {
	/*Dending on the string ...send a message out ... right now
	  SEnd the string */
	
	StringTokenizer st = new StringTokenizer(s, " ");
	if(!st.hasMoreTokens()) return;
	String msgType = st.nextToken();
	String msg = "";
	String recipient = "";

	ShuttleIOFromMsg ioMessage = null;

	msgType = msgType.trim().toLowerCase();
	
	if(msgType.equals("getonlinecontacts"))
	{
	    ioMessage = new ShuttleIOFromMsg(ShuttleIOFromMsg.SHUTTLEIO_REQUESTONLINECONTACTS,null, null);
	
	}else if (msgType.equals("getallcontacts")) {
	    ioMessage = new ShuttleIOFromMsg(ShuttleIOFromMsg.SHUTTLEIO_REQUESTALLCONTACTS, null, null);
	
	}else if (msgType.equals("getofflinecontacts")) {
	    ioMessage = new ShuttleIOFromMsg(ShuttleIOFromMsg.SHUTTLEIO_REQUESTOFFLINECONTACTS, null, null);    
	
	}else if(msgType.equals("sendtextmessage")) {
	    recipient = st.nextToken();
	    while(st.hasMoreTokens()) {
		msg += st.nextToken();
		msg += " ";
	    }
	    ioMessage = new ShuttleIOFromMsg(ShuttleIOFromMsg.SHUTTLEIO_SENDTEXTMESSAGE, recipient, msg);
	}else if (msgType.equals("addcontact")) {
	    ioMessage = new ShuttleIOFromMsg(ShuttleIOFromMsg.SHUTTLEIO_ADDCONTACT, st.nextToken().trim(), null);


	}else if (msgType.equals("removecontact")) {
	    ioMessage = new ShuttleIOFromMsg(ShuttleIOFromMsg.SHUTTLEIO_REMOVECONTACT, st.nextToken().trim(), null);


	}else if (msgType.equals("goodbye")) {
	    ioMessage = new ShuttleIOFromMsg(ShuttleIOFromMsg.SHUTTLEIO_GOODBYE, null, null);
	}else if (msgType.equals("whoami")) {
	    ioMessage = new ShuttleIOFromMsg(ShuttleIOFromMsg.SHUTTLEIO_WHOAMI, null, null);
	}else if (msgType.equals("ping")) {
	    ioMessage = new ShuttleIOFromMsg(ShuttleIOFromMsg.SHUTTLEIO_PING, st.nextToken().trim(), null);
	}else if(msgType.equals("?")) {
	    System.out.println("Available commands are: ");
	    System.out.println("getonlinecontacts, getallcontacts, getofflinecontacts, sendtextmessage, addcontact, removecontact, goodbye, whoami");

	}else{
	    System.out.println("Unrecognized command: "+msgType);
	}
	    
	if(msg != null)
	    dispatch(ioMessage);
    }
	
  //our input thread running in the background

    //********************************************************************************/

  private class ShuttleIOInputThread implements Runnable {
      ShuttleIO owner;
      BufferedReader in;
      String inputLine;
      
      public ShuttleIOInputThread(ShuttleIO o) {
	  owner = o;
	  in = new java.io.BufferedReader(new java.io.InputStreamReader(System.in)); 
      }
      public void run() {

	  while(true) {
	      inputLine = "//";
	      System.out.print("\n\n[ShuttleIO]: Enter message: ");
	      while(inputLine.indexOf("//") == 0) {
		  try {
		      inputLine = in.readLine();
		  }catch(Exception e) {
		      e.printStackTrace();
		  }
		  owner.HandleInput(inputLine);
	      }
	  }
      }
  } 
    //********************************************************************************/
    


    public void destroy() throws Exception {
	// TODO
    }
    
    //  methods for ProfilableIF
    /* Specified by seda.sandStorm.api.ProfilableIF */
    public int profileSize()
    {
	return this_sink.size();
    }
    //  end methods for ProfilableIF
    
    public static String printGuid (SecureHash guid) {
        String out="";
	int [] gdigits = RoutingTable.digits(guid);
	for (int i = 0; i < gdigits.length; ++i) {
	    out += Integer.toHexString (gdigits [i]);
	}   
	return out;
    }
    
    private boolean DEBUG;
    private String _numid;
    private int _id;
    private int _oldobjs;
    private SecureHash _client_guid;
    
    private HashMap _guid_located;  //guid->bool: done locatemsg on them
    private SinkIF this_sink;
    
    private NodeId _self_node_id;
    private Classifier _classifier;
 
    private Thread inputThread = null;
}







