/**
 * StaticPtrMsg.
 * A class to notify other statically configured tapestry nodes that we're
 * pointing to them for some entry in our routing table.  This message is
 * used to build up a nodes set of backpointers (_revpointers).  For a node
 * X which we point to in multiple levels of our routing table, we should
 * send off a message for each level it occurs on.
 *
 * @author     Ben Y. Zhao
 * @version    $Id: StaticPtrMsg.java,v 1.3 2002/08/03 18:22:34 ravenben Exp $
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.QSException;
import ostore.util.SecureHash;

public class StaticPtrMsg 
extends NetworkMessage implements QuickSerializable {

    public SecureHash srcguid;
    public int level;

    public StaticPtrMsg (NodeId peer, SecureHash srcguid, int level) {
	super (peer, false /* outbound */);
	this.srcguid = srcguid;
	this.level = level;
    }

    public StaticPtrMsg (InputBuffer buffer) 
    throws QSException {
      super (buffer); 
      srcguid = (SecureHash) buffer.nextObject();
      level = buffer.nextInt();
    }

    public void add_to_digest (OutputBuffer buffer) {
      throw new NoSuchMethodError ();
    }
  
    public void serialize (OutputBuffer buffer) {
      super.serialize (buffer);
      buffer.add(srcguid);
      buffer.add(level);
    }

    public String toString () {
	return "(StaticPtrMsg from " + srcguid+ ", level="+level+")";
    }
}

