/**
 *  Test.java
 *
 *  @author     Ben Y. Zhao
 *  @version    $Id: Test.java,v 1.24 2003/02/19 10:32:31 hweather Exp $
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.*;
import java.io.*;
import java.security.*;
import ostore.util.*;
import ostore.dispatch.*;
import ostore.tapestry.*;
import ostore.tapestry.api.*;
import seda.sandStorm.api.*;
import seda.sandStorm.core.*;
import seda.sandStorm.internal.SinkProxy;
import cryptix.jce.provider.random.DevRandom;
import ostore.tapestry.api.TapestryQuery;
import ostore.tapestry.api.TapestryQueryResult;
import ostore.tapestry.api.TapestryTag;
import ostore.tapestry.api.TapestryQueryState;
import ostore.tapestry.api.TapestryQueryResultMatch;
import ostore.tapestry.api.TapestryQueryResultNoMatch;

public class Test implements EventHandlerIF, ProfilableIF {

    public String class_tag = "Test";

    private void dispatch (QueueElementIF item) {
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    Carp.out ("Could not dispatch " + item);
	    System.exit (1);
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

  public void printtagln(String tag, String msg) {
    System.err.print(tag+":"+msg+"\n");
  }

    public void init (ConfigDataIF config) throws Exception {

	DEBUG = true;

	Types.registerAllTypes ();

	_classifier = Classifier.getClassifier (
                new NodeId (config.getString ("node_id")));
	
	//initialize the numerical arguments
	_clients = Integer.parseInt(config.getString ("numclients"));
	_oldobjs = Integer.parseInt(config.getString ("oldobjs"));

	_guid_located = new HashMap();
	_guid_oldobjs = new HashMap();
	readycount = 0;
	foundcount = 0;
	_running = 0;

	_self_node_id = new NodeId (config.getString ("node_id"));

	class_tag = class_tag + ".["+_self_node_id.port()+"]";
	if (DEBUG) ostore.util.Debug.turnOn(class_tag);

	final String tag = class_tag +".init";
	String on = config.getString("power");
	if (on != null && on.equalsIgnoreCase("off")) return;

	//generate GUIDs for all clients
	SecureHash guidx;
	for (int i=0; i<_clients; i++) {
	  for (int j=0; j<_oldobjs; j++) {
	    guidx = new SHA1Hash ((""+i+j).getBytes());
	    _guid_oldobjs.put(guidx, Boolean.FALSE);
	  }
	}
	if (DEBUG) printtagln(tag, "Done initializing guid hashmaps");

	String stageName = config.getStage().getName();
	try 
	{
	  if(config.getBoolean("Profile"))
	  {
	    config.getManager().getProfiler().add("ostore.tapestry.impl.Test", this);
	    SinkIF sink =
	      config.getManager().getStage(stageName).getSink();
	    this_sink = new SinkProxy(sink,
				      config.getManager(),
				      config.getStage().getWrapper());
	  }
	  else
	    this_sink =
	      config.getManager().getStage(stageName).getSink();
	} 
	catch (NoSuchStageException e) 
	{
	  Carp.die (tag + ": could not get own sink.");
	}

	Filter filter;
	String [] event_types = {
	  "seda.sandStorm.api.StagesInitializedSignal",
	  "ostore.tapestry.impl.TapestryReadyMsg",
	  "ostore.tapestry.api.TapestryDetachConfirm"
	};	  

	for (int i=0; i< event_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(event_types[i]))) {
	    Carp.out ("Could not require type " + event_types[i]);
	    System.exit (1);
	  }
	  _classifier.subscribe (filter, this_sink);
	}

	String [] msg_types = {
	  "ostore.tapestry.impl.TestLocateMsg",
	  "ostore.tapestry.impl.TestFoundMsg",
	  "ostore.tapestry.api.TapestryLocateFailure",
	  "ostore.tapestry.impl.TestDeleteMsg"
	};

	for (int i=0; i< msg_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(msg_types[i])))
	    Carp.die ("Could not require type " + msg_types[i]);
	  if (! filter.requireValue("inbound", new Boolean(true)))
	    Carp.die ("Could not require inbound = true for "+
		      msg_types[i]);
	  _classifier.subscribe (filter, this_sink);
	}

	count = 0;
	System.out.println("Test Tapestry stage initialized!!");
    }

  public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {

    DEBUG = true;
    final String tag = class_tag+".handleEvent";

    if (item instanceof StagesInitializedSignal) {
      // do nothing
      
    } else if (item instanceof TapestryReadyMsg) {

      SecureHash guidx, objguid;
      TapestryPublishMsg msg;
      _client_guid = ((TapestryReadyMsg)item).node_guid;
      if (_running == 0) {
	for (int i=0; i<_clients; i++) {
	  guidx = new SHA1Hash ((""+i).getBytes());
	  msg = new TapestryPublishMsg(guidx, new NullTag());
	  if (DEBUG) printtagln(tag, "Publishing " + printGuid(guidx)
				+ ".");
	  dispatch(msg);
	  _guid_located.put(guidx, Boolean.FALSE);
	}

	_running = 1;
	_classifier.dispatch_later(item, 2000);
      } else { //_running == 1
	//now publishing should be done, now do locate
	TestLocateMsg outmsg;
	for (Iterator i=_guid_located.keySet().iterator(); i.hasNext();) {
	  objguid = (SecureHash) i.next();
	  outmsg = new TestLocateMsg(_self_node_id, objguid, 
				     new NullTagQuery(), 
				     null);
	  dispatch(outmsg);
	  if (DEBUG) printtagln(tag, "(1) Sending locate msg to Guid:"+
					printGuid(objguid));
	}
	for (Iterator i=_guid_oldobjs.keySet().iterator(); i.hasNext();) {
	  objguid = (SecureHash) i.next();
	  outmsg = new TestLocateMsg(_self_node_id, objguid, 
				     new NullTagQuery(), null);
	  dispatch(outmsg);
	  if (DEBUG) printtagln(tag, "(2) Sending locate msg to Guid:"+
					printGuid(objguid));
	}
      }
    } else if (item instanceof TestFoundMsg) {
      TestFoundMsg msg = (TestFoundMsg) item;
      System.out.println("Got TestFoundMsg for obj "+msg.objGuid+" at node"+
			 msg.peer);
      if (_running < 2) {
	if (DEBUG) printtagln(tag, 
			      "Found object "+msg.objGuid+" at node "+
			      msg.peer);
	Object x, y;
	x = _guid_located.get(msg.objGuid);
	y = _guid_oldobjs.get(msg.objGuid);       
	if ((x != null && x.equals(Boolean.FALSE)) ||
	    (y != null && y.equals(Boolean.FALSE))) {
	  if (_guid_located.containsKey(msg.objGuid))
	    _guid_located.put(msg.objGuid, Boolean.TRUE);
	  else
	    _guid_oldobjs.put(msg.objGuid, Boolean.TRUE);
	  foundcount++;
	  System.out.println("Foundct: "+foundcount+
			     " locatesize: "+_guid_located.size()+
			     " oldsize: "+_guid_oldobjs.size());
	  if (foundcount == _guid_located.size()+_guid_oldobjs.size()) {
	    //	    System.exit(0);
	    if (DEBUG) 
	      printtagln(tag, "ALL SUCCESS with publish/locate!");
	    _running = 2;
	    //now we test the unpublish
	    SecureHash objguid;
	    TapestryUnpublishMsg outmsg;
	    for (Iterator i=_guid_located.keySet().iterator(); i.hasNext();) {
	      objguid = (SecureHash) i.next();
	      outmsg = new TapestryUnpublishMsg(objguid, new NullTag());
	      dispatch(outmsg);
	      if (DEBUG) printtagln(tag, "Sending unpublish msg to Guid:"+
					    printGuid(objguid));
	    }
	    _classifier.dispatch_later(item, 2000);
	  }
	}
      } else if (_running == 2) {
	SecureHash objguid;
	TestLocateMsg outmsg;
	System.out.println("Now entering _running==2\n");
	_running = 3;
	for (Iterator i=_guid_located.keySet().iterator(); i.hasNext();) {
	  //now send locate messages for the non-oldobj objects to see
	  //if they have been removed
	  objguid = (SecureHash) i.next();
	  outmsg = new TestLocateMsg(_self_node_id, objguid, 
				     new NullTagQuery(), null);
	  dispatch(outmsg);
	  if (DEBUG) printtagln(tag, "Sending locate msg to Guid:"+
					printGuid(objguid));
	}
      } else if (_running == 3) {
	//if response from the locates, then we failed in the delete
	if (DEBUG) 
	  printtagln(tag, "Got response after objects unpublished."+
			   "\nFailing.");
				      
	System.exit(-1);
      }

    } else if (item instanceof TestLocateMsg) {
      
      TestLocateMsg msg = (TestLocateMsg) item;
      if (DEBUG) printtagln(tag, "Got TestLocateMsg from "+
			    msg.peer+" for guid: "+
			    printGuid(msg.guid)+" w/ hop="+
			    msg.hopCount);
      //now send back a routemsg to the originator of the locatemsg
      TestFoundMsg outmsg = new TestFoundMsg(msg.querysrc, msg.guid);
      dispatch(outmsg);

    } else if (item instanceof TapestryLocateFailure) {
      TapestryLocateFailure msg = (TapestryLocateFailure) item;
      if (DEBUG) printtagln(tag, "Got TapestryLocateFailure from "+
			    msg.peer+" for guid: " +
			    printGuid(msg.obj_guid)+" w/ hop="+
			    msg.hopCount);
      if (_running == 3) failedcount++;
      if (failedcount == _clients) {
	System.err.println("SUCCESS, all failures received.");
	//locate requests all failed, now begin dynamic deletion test
	/*	SecureHash guidx;
	TapestryPublishMsg m;
	for (int i=0; i<12; i++) {
	  guidx = new SHA1Hash ((""+i).getBytes());
	  m = new TapestryPublishMsg(guidx, null);
	  if (DEBUG) printtagln(tag, "Publishing " + printGuid(guidx)
				+ ".");
	  dispatch(m);
	  }*/
	TapestryNodeDetachMsg detach = new TapestryNodeDetachMsg();
	dispatch (detach);
      }
    } else if (item instanceof TapestryDetachConfirm) {
      System.err.println("Node detached!!  Now waiting for others to query");
    } else if (item instanceof TestDeleteMsg) {
      System.err.println("Received TestDeleteMsg");
      TestDeleteMsg msg = (TestDeleteMsg) item;
      
      if (msg.type == TestDeleteMsg.REQUEST_TYPE) {
	count ++;
	if (DEBUG)
	  printtagln(tag, "Sending TestDelete.RESPONSE to: "+msg.peer+
		     ", count now:"+count);
	TestDeleteMsg resp = new TestDeleteMsg(msg.peer, 
					       TestDeleteMsg.RESPONSE_TYPE);
	dispatch(resp);
      } else if (msg.type == TestDeleteMsg.SUCCESS_TYPE) {
	count --;
	if (DEBUG)
	  printtagln(tag, "Got TestDelete.SUCCESS from:"+msg.peer+
		     ", count now:"+count);
	if (count == 0) {
	  //now we're done
	  System.err.println("SUCCESS, DELETION VERIFIED.");
	  System.exit(0);
	}
      } else if (msg.type == TestDeleteMsg.FAILURE_TYPE) {
	System.err.println("ERROR: received failed msg from: "+msg.peer);
	System.exit(-1);
      } else 
	System.err.println("ERROR: received unexpected TestDelete: "+msg);
    } else {
      Carp.die ("ostore.tapestry.impl.Test: Received unknown "
		+ "QueueElementIF: " + item);
    }
  }
  
  public void destroy() throws Exception {
    // TODO
  }

    //  methods for ProfilableIF
    /* Specified by seda.sandStorm.api.ProfilableIF */
    public int profileSize()
    {
	return this_sink.size();
    }
    //  end methods for ProfilableIF

    public static String printGuid (SecureHash guid) {
        String out="";
	int [] gdigits = RoutingTable.digits(guid);
	for (int i = 0; i < gdigits.length; ++i) {
	    out += Integer.toHexString (gdigits [i]);
	}   
	return out;
    }

  private int _running;
  /* _running goes in 4 stages
   * 0. just started
   * 1. done publishing, sending out locateMsgs
   * 2. done unpublishing, send out locateMsgs
   * 3. waiting for negative results
   */
  private int count;
  private ssTimer timer;
  private boolean DEBUG;
  private int _clients;
  private int _oldobjs;
  private SecureHash _client_guid;
  
  private HashMap _guid_oldobjs;
  private HashMap _guid_located;  //guid->bool: done locatemsg on them
  private int readycount;
  private int foundcount;
  private int failedcount;
  private SinkIF this_sink;

  private NodeId _self_node_id;
  private Classifier _classifier;

  /*    private RPCert _rp_cert;
    private SecureRandom _secure_random;
    private PRCert _pr_cert;
    private QSPublicKey _rp_pkey;
    private Signature _sengine;
    private QSPublicKey _principal_pkey;
    private QSPrivateKey _principal_skey;
  */
}

