package ostore.tapestry.impl;
import ostore.util.QSException;
import ostore.util.SecureHash;
import ostore.util.OutputBuffer;
import ostore.tapestry.api.TapestryRouteMsg;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;

/**
 * A generic test message routed through the Tapestry network. Used in 
 * various tests, including the FullTest suite.
 *
 * @author     Ben Y. Zhao
 * @version    $Id: TestReadyMsg.java,v 1.13 2003/04/28 21:57:24 hling Exp $
 */
public class TestReadyMsg 
  extends TapestryRouteMsg implements QuickSerializable {

  public boolean request;
    public double  failure;
  public SecureHash src;
  public SecureHash id;
  public int seq;
  public String  comment = " ";

    public TestReadyMsg (SecureHash dest_guid, SecureHash src,
			 SecureHash id, boolean request) {
      super (dest_guid);
      this.src = src;
      this.id = id;
      this.request = request;
      this.seq = 0;
      this.failure = 0.0;
    }

    public TestReadyMsg (SecureHash dest_guid, SecureHash src,
			 SecureHash id, int seq, boolean request) {
      super (dest_guid);
      this.src = src;
      this.id = id;
      this.request = request;
      this.seq = seq;
      this.failure = 0.0;
    }

    public TestReadyMsg (SecureHash dest_guid, SecureHash src,
			 SecureHash id, double failure, boolean request) {
      super (dest_guid);
      this.src = src;
      this.id = id;
      this.request = request;
      this.seq = 0;
      this.failure = failure;
    }

    public TestReadyMsg (InputBuffer buffer) throws QSException {
      super(buffer);
      src = (SecureHash) buffer.nextObject();
      id = (SecureHash) buffer.nextObject();
      seq = buffer.nextInt();
      request = buffer.nextBoolean();
      failure = buffer.nextDouble();

      boolean have_comment = buffer.nextBoolean();
      if (have_comment)
	  comment = buffer.nextString();
      else comment = " ";
    }

    public void serialize (OutputBuffer buffer) {
      super.serialize (buffer);
      buffer.add(src);
      buffer.add(id);
      buffer.add(seq);
      buffer.add(request);
      buffer.add(failure);

      if (comment == null)
	  buffer.add(false);
      else {
	  buffer.add(true);
	  buffer.add(comment);
      }
    }

    public String toString () {
	return "TestReadyMsg: src = " + src + " dest = " +
	    super.peer + " id = " + id +  " seq = " + seq;
    }
}

