/**
 *  Microbench.java
 *
 *  @author     Ben Y. Zhao
 *  @version    $Id: Microbench.java,v 1.20 2003/02/19 10:32:31 hweather Exp $
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. either the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl.test;
import java.util.*;
import java.io.*;
import java.security.*;
import ostore.util.*;
import ostore.dispatch.*;
import ostore.tapestry.*;
import ostore.tapestry.api.*;
import ostore.tapestry.impl.*;
import seda.sandStorm.api.*;
import seda.sandStorm.core.*;
import seda.sandStorm.internal.SinkProxy;

public class Microbench implements EventHandlerIF, ProfilableIF {

    public String class_tag = "Test";

    private void dispatch (QueueElementIF item) {
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    Carp.out ("Could not dispatch " + item);
	    e.printStackTrace(System.err);
	    System.exit (1);
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

  public void printtagln(String tag, String msg) {
    System.err.print(tag+":"+msg+"\n");
  }

    public void init (ConfigDataIF config) throws Exception {

	DEBUG = true;

	Types.registerAllTypes ();

	_classifier = Classifier.getClassifier (
                new NodeId (config.getString ("node_id")));
	
	//initialize the numerical arguments

	_running = 0;
	warmedup = 0;
	sst = new ssTimer();

	_self_node_id = new NodeId (config.getString ("node_id"));

	class_tag = class_tag + ".["+_self_node_id.port()+"]";
	if (DEBUG) ostore.util.Debug.turnOn(class_tag);

	final String tag = class_tag +".init";
	sender = config.getString("sender").equalsIgnoreCase("true");

	sender_id = new NodeId(config.getString("sender_nodeID"));
	receiver_id = new NodeId(config.getString("receiver_nodeID"));
	receiver_guid = new SHA1Hash(receiver_id);
	_messages = new MicroRouteMsg [16];

	if (sender) {
	  //generate messages for all sizes
	  int size = 64;
	  for (int i = 0; i < 16; i ++) {
	    _messages[i] = new MicroRouteMsg(receiver_guid, 
					     new byte [size], size);
	    size = size * 2;
	  }
	}
	if (DEBUG) printtagln(tag, "Done initializing byte-sized routemsgs.");

	String stageName = config.getStage().getName();
	try {
	  if(config.getBoolean("Profile")) {
	    config.getManager().getProfiler().add(
			      "ostore.tapestry.impl.test.Microbench", this);
	    SinkIF sink =
	      config.getManager().getStage(stageName).getSink();
	    this_sink = new SinkProxy(sink,
				      config.getManager(),
				      config.getStage().getWrapper());
	  } else
	    this_sink = config.getManager().getStage(stageName).getSink();
	} catch (NoSuchStageException e) {
	  Carp.die (tag + ": could not get own sink.");
	}

	Filter filter;
	String [] event_types = {
	  "seda.sandStorm.api.StagesInitializedSignal",
	  "ostore.tapestry.impl.TapestryReadyMsg"
	};

	for (int i=0; i< event_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(event_types[i]))) {
	    Carp.out ("Could not require type " + event_types[i]);
	    System.exit (1);
	  }
	  _classifier.subscribe (filter, this_sink);
	}

	String [] msg_types = {
	  "ostore.tapestry.impl.test.MicroRouteMsg",
	  "ostore.tapestry.impl.test.MicroBarrierMsg"
//	  "ostore.tapestry.impl.test.MicroLocateMsg",
//	  "ostore.tapestry.impl.test.MicroPublishMsg",
	};

	for (int i=0; i< msg_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(msg_types[i])))
	    Carp.die ("Could not require type " + msg_types[i]);
	  if (! filter.requireValue("inbound", new Boolean(true)))
	    Carp.die ("Could not require inbound = true for "+
		      msg_types[i]);
	  _classifier.subscribe (filter, this_sink);
	}

	System.out.println("Tapestry Microbenchmark initialized!!");
    }

  public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {

    DEBUG = true;
    final String tag = class_tag+".handleEvent";

    if (item instanceof StagesInitializedSignal) {
      // do nothing.

    } else if (item instanceof TapestryReadyMsg) {
      _client_guid = ((TapestryReadyMsg)item).node_guid;
      currSize = 64;
      currIndex = 0;
      total = 10001;
      if (sender) {
	//start sending route messages
	MicroRouteMsg mrm;
	System.out.println("Sending out MicroRouteMsg of size "+currSize);
	mrm = _messages[0];

	long start = System.currentTimeMillis();
	for (int i=0; i<total; i++) {
	  //	    mrm.setTime();
	  dispatch(mrm);
	}
	start = System.currentTimeMillis() - start;
	System.out.println("Processed "+total+" MicroRouteMsgs in "+
			   start+" ms.");
      } else {
	System.out.println("initializing storage arrays.");
	count = 0;
	timer = 0;
	System.out.println("storage arrays initialized.");
	//if receiver, init variables, init output
	System.out.println("Msgs          MsgSize(Bytes)           Time(ms)");
      }

    } else if (item instanceof MicroRouteMsg) {
      MicroRouteMsg msg = (MicroRouteMsg) item;
      //we must be the receiver
      int s = msg.size;
      if (timer == 0) 
	timer = System.currentTimeMillis();
      count ++;
      if (count == total) {
	//we got the last message, now calculate thruput
	timer = System.currentTimeMillis() - timer;
	System.out.println(" "+(total-1)+"           "+s+"            "+timer);
	System.out.flush();
	System.gc();
	currSize = currSize *2;
	timer =0;
	count =0;
	int newtotal = 838861000 / currSize;
	if (newtotal < total)
	  total = newtotal+1;
	if (total < 300) {
	  if (warmedup < runs) {
	    total = 10001;
	    currSize = 64;
	    warmedup++;
	    MicroBarrierMsg mbm = new MicroBarrierMsg(msg.peer, total);
	    System.out.println("Sent out MicroBarrier to ask for size "+
			       currSize);
	    dispatch(mbm);
	  } else {
	    MicroBarrierMsg mbm = new MicroBarrierMsg(msg.peer, total);
	    System.out.println("Sent out MicroBarrier to ask sender to quit");
	    dispatch(mbm);
	    System.exit(0);
	  }
	} else {
	    MicroBarrierMsg mbm = new MicroBarrierMsg(msg.peer, total);
	    System.out.println("Sent out MicroBarrier to ask for size "+
			       currSize);
	    dispatch(mbm);
	}
      }

    } else if (item instanceof MicroBarrierMsg) {
      //sender receives confirmation that last set was received
	System.out.println("Got microbarriermsg");
      if (_running == 0) {
	System.out.println("Garbage collecting for 2 seconds");
	System.out.flush();
	System.gc();
	_running++;
	sst.registerEvent(2000, item, this_sink);
	return;
      } else {
	System.out.println("_running was not 0, changing to 0.");
	_running = 0;
      }

      MicroBarrierMsg mbm = (MicroBarrierMsg) item;
      currSize = currSize * 2;
      currIndex ++;
      if (currIndex == 16) {
	if (warmedup==runs)
	  System.exit(0);

	warmedup++;
	currSize = 64;
	currIndex = 0;
	total = 10001;
      } else {
	total = mbm.size;
      }
      MicroRouteMsg mrm = _messages[currIndex];
      System.out.println("Got MicroBarrier for size "+mbm.size);
      System.out.println("Sending out "+total+" MicroRouteMsg of size "
			 +currSize);
      long start = System.currentTimeMillis();
      for (int i=0; i<total; i++) {
	// mrm.setTime();
	dispatch(mrm);
      }
      start = System.currentTimeMillis() - start;
      System.out.println("Processed "+total+" MicroRouteMsgs of size "+
			 currSize+" in "+ start+" ms.");

    } else {
      Carp.die ("ostore.tapestry.impl.Test: Received unknown "
		+ "QueueElementIF: " + item);
    }
  }

  public void output() {
    System.out.println("Output for Tapestry throughput microbenchmark:");
    System.out.println("Message size(B)                   Time/10000msg");
    int size;
    for (int i=0; i<10; i++) {
      size = 64<<i;
      System.out.println(""+size + "                              "+
			 timer);
    }
  }
  
  public void destroy() throws Exception {
    // TODO
  }

  //  methods for ProfilableIF
  /* Specified by seda.sandStorm.api.ProfilableIF */
  public int profileSize() {
    return this_sink.size();
  }
  //  end methods for ProfilableIF

  public static String printGuid (SecureHash guid) {
    String out="";
    int [] gdigits = RoutingTable.digits(guid);
    for (int i = 0; i < gdigits.length; ++i) {
      out += Integer.toHexString (gdigits [i]);
    }   
    return out;
  }

  private ssTimer sst;
  private int warmedup;
  private int _running;
  private final int INITIAL = 0;
  private final int ROUTETHROUGHPUT = 1;
  private final int LOCATETHROUGHPUT = 2;
  private final int PUBLISHTHROUGHPUT = 3;
  private boolean DEBUG;
  private boolean sender;
  private NodeId sender_id;
  private NodeId receiver_id;
  private SecureHash receiver_guid;
  private long timer;
  private int count;
  private int currSize;
  private int currIndex;
  private int total;
  private static final int runs = 7;
  
  private MicroRouteMsg [] _messages;
  private int readycount;
  private int foundcount;
  private int failedcount;
  private SinkIF this_sink;

  private NodeId _self_node_id;
  private SecureHash _client_guid;
  private Classifier _classifier;
}

