package ostore.tapestry.interweave;
import ostore.tapestry.api.*;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;

/**
 * InterweaveFileRequestMsg.java
 *
 * The message that carries a request for the file FILENAME on PEER
 *
 * @author Benjamin K. Poon
 */
public class InterweaveFileRequestMsg
    extends TapestryRouteMsg implements QuickSerializable {

    public SecureHash from;
    public String filename; // full filename including extension

    public InterweaveFileRequestMsg (SecureHash to, SecureHash from, String filename) {
	super(to);
	this.from = from;
	this.filename = filename;
    }

    public InterweaveFileRequestMsg (InputBuffer buffer) throws QSException {
	super(buffer);
	from = (SecureHash)buffer.nextObject();
	filename = buffer.nextString();
    }
    
    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);
	buffer.add(from);
	buffer.add(filename);
    }
    
    public Object clone () throws CloneNotSupportedException {
	InterweaveFileRequestMsg result = (InterweaveFileRequestMsg) super.clone();
	result.from = from;
	result.filename = filename;
	return result;
    }

    public String toString() {
	return "InterweaveFileRequestMsg: filename="+filename;
    }

}
