/**
 * PatchworkChangePeriodMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: PatchworkChangePeriodMsg.java,v 1.3 2002/08/05 10:33:25 strib Exp $
 */

package ostore.tapestry.patchwork;
import seda.sandStorm.api.QueueElementIF;
import ostore.util.NodeId;

/**
 * This represents a command to the <code>PatchworkStage</code>
 * indicating that it needs to change the period of one of its
 * messages to a new value (in milliseconds).
 */
public class PatchworkChangePeriodMsg implements QueueElementIF {

    public static final int LATENCY = 0;
    public static final int BANDWIDTH = 1;
    public static final int LOSS = 2;
    public static final int RECALC_MODEL = 3;
    public static final int RESPOND = 4;

    public static final int OP_ADD = 0;
    public static final int OP_MULT = 1;
    public static final int OP_SUB = 2;    
    public static final int OP_DIV = 3;    
    public static final int OP_EXP = 4;
   
    public int type; 
    public int newPeriod;
    public int newOp;
    public double newFactor;

    public NodeId node;

    public PatchworkChangePeriodMsg( int type, int newPeriod ) {

        if( (type != LATENCY) && (type != BANDWIDTH) &&
            (type != LOSS) && (type != RECALC_MODEL) && (type != RESPOND) ) {
          throw new RuntimeException( "Invalid type (" + type +
                                      ") for PatchworkChangePeriodMsg" );
        } else {
          this.type = type;
        }

	if( newPeriod < 0 ) {
          throw new RuntimeException( "New period (" + newPeriod +
                                      ") invalid for PatchworkChangePeriodMsg" 
					+ ", must be >0" );
	} else {
	  this.newPeriod = newPeriod;
	}

	newOp = -1;
	newFactor = -1;
	node = null;

    }

    public PatchworkChangePeriodMsg( int type, int newPeriod, int newOp, 
				     double newFactor ) {

	this( type, newPeriod );

        if( (newOp != OP_ADD) && (newOp != OP_MULT) &&
            (newOp != OP_SUB) && (newOp != OP_DIV) && (newOp != OP_EXP) ) {
          throw new RuntimeException( "Invalid op (" + newOp +
                                      ") for PatchworkChangePeriodMsg" );
        } else {
          this.newOp = newOp;
        }

	if( newFactor < 0 ) {
          throw new RuntimeException( "New factor (" + newFactor +
                                      ") invalid for PatchworkChangePeriodMsg" 
					+ ", must be >0" );
	} else {
	  this.newFactor = newFactor;
	}

    }

    public PatchworkChangePeriodMsg( int type, int newPeriod, NodeId node ) {

	this( type, newPeriod );

        if( (type == RECALC_MODEL) || (type == RESPOND) ) {
          throw new RuntimeException( "Cannot consruct a node-based change" +
				      " period message for recalculating" +
				      " or responding" );
	}

	if( node == null ) {
	    throw new RuntimeException( "The node cannot be null" );
	}
	    
	this.node = node;

    }

    public String toString () {
	return "(PatchworkChangePeriodMsg, type = " + type + 
	", new period = " + newPeriod + ")";
    }
}

