/**
 * PatchworkNodeStatusMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: PatchworkNodeStatusMsg.java,v 1.6 2003/02/19 10:32:31 hweather Exp $
 */

package ostore.tapestry.patchwork;
import seda.sandStorm.api.QueueElementIF;
import ostore.util.SecureHash;
import ostore.util.NodeId;

/**
 * This event lets any other listening stage know that
 * the status of a particluar node has changed
 * (i.e. it is has been tagged as hibernating or dead).
 */
public class PatchworkNodeStatusMsg implements QueueElementIF {

    public static final int NORMAL = 0;
    public static final int HIBERNATING = 1;
    public static final int DEAD = 2;
   
    public int type; 
    public SecureHash node;
    public NodeId node_id;
    public int hops; // can also be used for level.

    /**
     * Constructor.
     *
     * @param node   the id of the node whose status has changed
     * @param node   the guid of the node whose status has changed
     * @param type   the new status, either NORMAL, HIBERNATING, or DEAD
     * @param hops    Number of 
     *               {@link ostore.tapestry.impl.TapestryReadyMsg tapestry} 
     *               between <CODE>node</CODE> and self.  
     *               NOTE: this also can be the <I>level</I> in the 
     *               {@link ostore.tapestry.impl.RoutingTable} between 
     *               other <CODE>node</CODE> and self. */
    public PatchworkNodeStatusMsg( NodeId node_id, SecureHash node, 
				   int type, int hops ) {

        if( (type != NORMAL) && (type != HIBERNATING) &&
            (type != DEAD) ) {
          throw new RuntimeException( "Invalid type (" + type +
                                      ") for PatchworkNodeStatusMsg" );
        } else {
          this.type = type;
        }

	this.node = node;
	this.node_id = node_id;
	this.hops = hops;

    }

    /**
     * Constructor.
     *
     * @param node   the guid of the node whose status has changed
     * @param type   the new status, either NORMAL, HIBERNATING, or DEAD
     * @param hops    Number of 
     *               {@link ostore.tapestry.impl.TapestryReadyMsg tapestry} 
     *               between <CODE>node</CODE> and self.  
     *               NOTE: this also can be the <I>level</I> in the 
     *               {@link ostore.tapestry.impl.RoutingTable} between 
     *               other <CODE>node</CODE> and self. */
    public PatchworkNodeStatusMsg( SecureHash node, int type, int hops ) {
	this( null, node, type, hops );
    }

    public String toString () {
	return "(PatchworkNodeStatusMsg, type = " + type + 
	    ", node = " + node + 
	    ", hops/level = " + hops + ")";
    }
}

