/**
 * FromBytesInputBuffer.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;

/**
 * For help in porting.  Use like this:
 * <pre>
 *     public Foo (byte [] data, int [] offset) throws QSException {
 *         this (new FromBytesInputBuffer (data, offset));
 *     }
 * </pre>
 *
 * @author	Sean C. Rhea
 * @version	$Id: FromBytesInputBuffer.java,v 1.2 2002/06/29 05:55:35 hweather Exp $
 */
public class FromBytesInputBuffer extends InputBufferImpl {
    
    private byte [] data;
    private int [] offset;
    private int end;

    public FromBytesInputBuffer (byte [] data, int [] offset) {
	this.data = data;
	this.offset = offset;
	end = data.length;
    }

    public byte nextByte () {
	byte result = data [offset [0]];
	++offset [0];
	return result;
    }

    public void nextBytes (byte [] out, int off, int length) {
        if ((out != null) && (data != null))
        {
	    if (out.length < off + length)
                throw new AssertionViolatedException("dst.length = " + out.length + ", but " +
                       "dst_offset = " + off + " and length = " +
                       length + ".");
	    if (data.length < offset[0] + length)
                throw new AssertionViolatedException("data.length = " + data.length + ", but " +
                       "src_offset = " + offset[0] + " and length = " +
                       length + ".");
        }
	System.arraycopy (data, offset [0], out, off, length);
	offset [0] += length;
    }
    
    public void nextBytes (OutputBuffer output) {
	output.add (data, offset [0], end - offset [0]);
    }

    public void convertToBytes (byte out_buffer [][], int out_offset [][]) {
	out_buffer [0] = data;
	out_offset [0][0] = offset [0];
    }

    public void skipBytes (int count) {
	offset [0] += count;
    }

    public InputBuffer subBuffer (int length) {
	FromBytesInputBuffer result = new FromBytesInputBuffer (data, offset);
	result.end = result.offset [0] + length;
	return result;
    }

    public Object clone () throws CloneNotSupportedException {
	FromBytesInputBuffer result = (FromBytesInputBuffer) super.clone ();
	result.data = data;
	result.offset = new int [1];
	result.offset [0] = offset [0];
	result.end = end;
	return result;
    }
}

