/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
 * Pair for QS classes.
 *
 * @see ostore.util.Pair
 *
 * @author     Jeff Pang
 * @version    $Id: QSPair.java,v 1.1 2003/03/14 22:28:03 jeffpang Exp $
 */
public final class QSPair implements QuickSerializable{
    public QuickSerializable first, second;

    /**
     * Default constructor
     */
    public QSPair() {}

    /**
     * Constructor
     * @param f == first object of the pair
     * @param s == second object of the pair
     */
    public QSPair(QuickSerializable f, QuickSerializable s) {
        first = f;
        second = s;
    }

    /* Implements ostore.util.QuickSerializable */
    public QSPair(InputBuffer buffer) throws QSException
    {
	first = buffer.nextObject();
	second = buffer.nextObject();
    }

    /* Implements ostore.util.QuickSerializable */
    public void serialize(OutputBuffer buffer)
    {
	buffer.add(first);
	buffer.add(second);
    }

    /**
    ***  Returns the first element of the pair.
    ***
    ***  @return The first element of the pair.
    **/
    public final QuickSerializable first()
    {
        return first;
    }

    /**
    ***  Returns the second element of the pair.
    ***
    ***  @return The second element of the pair.
    **/
    public final QuickSerializable second()
    {
        return second;
    }

    public int hashCode () {
	if(first == null && second == null)
	  return 0;
	else if(first == null)
	  return second.hashCode();
	else if(second == null)
	  return first.hashCode();
	else
	  return first.hashCode () ^ second.hashCode ();
    }

    public boolean equals (Object other) {
	if (! (other instanceof QSPair)) return false;
	QSPair rhs = (QSPair) other;
	if(first == null && second == null)
	  return (rhs.first == null && rhs.second == null);
	else if(first == null)
	  return (rhs.first == null && second.equals(rhs.second));
	else if(second == null)
	  return (rhs.second == null && first.equals(rhs.first));
	else
	{
	  if (!first.equals (rhs.first)) return false;
	  return second.equals (rhs.second);
	}
    }

    public String toString()
    {
	String str = new String();
	str += "(QSPair";
	str += " first==" + first;
	str += " second==" + second;
	str += ")";
	return str;
    }
}
