/**
 * InputAnalyzer.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package Sonar.PacketAnalyzers;

import Sonar.*;
import Sonar.util.*;
import Sonar.event.*;
import Sonar.Dialog.*;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.accessibility.*;
import java.util.zip.*;

import ostore.util.NodeId;
import ostore.util.QSString;
import ostore.util.QSVector;
import ostore.tapestry.impl.RouteMsg;
import ostore.tapestry.impl.RouteObjRootMsg;
import ostore.tapestry.impl.RouteObjDestMsg;

import ostore.apps.visdemo.SonarReqMessage;

/**
 * 
 * @author	Jason Lee <jlee81@cs.berkeley.edu>
 */

public class InputAnalyzer extends PacketAnalyzer {
    protected MenuManager menu_manager;

    protected QSVector networkMsgs;
    protected QSVector routeMsgs;
    protected QSVector routeObjMsgs;
    
    public InputAnalyzer() {
	menu_manager = new MenuManager();
        MainClass.addOptionsPanelContributor(this);

	networkMsgs = new QSVector();
	routeMsgs = new QSVector();
	routeObjMsgs = new QSVector();

	/* TODO : can just be initialized here */
    }

    public void sendMessage() {
	//TODO : hack
	/*
	for(Enumeration e = ((GridLocationAnalyzer) MainClass.locationAnalyzer).group_leaders.elements(); e.hasMoreElements();) {
	    //SonarReqMessage msg = new SonarReqMessage((NodeId)e.nextElement(),
	    //	    false, new ostore.util.SHA1Hash(), networkMsgs, routeMsgs, routeObjMsgs);
	    SonarReqMessage msg = new SonarReqMessage((NodeId)e.nextElement(),
	    	    false, new ostore.util.SHA1Hash(), SonarReqMessage.ROUTEMSG, routeMsgs); 
	    MainClass.packetSender.sendMessage(msg);
	}
	*/
	if(networkMsgs.size() > 0) {
	    MainClass.routingAnalyzer.registerMessages(networkMsgs, SonarReqMessage.NETWORK);
	}
	if(routeMsgs.size() > 0) {
	    MainClass.routingAnalyzer.registerMessages(routeMsgs, SonarReqMessage.ROUTEMSG);
	}
	if(routeObjMsgs.size() > 0) {
	    MainClass.routingAnalyzer.registerMessages(routeObjMsgs, SonarReqMessage.ROUTEOBJ);
	}
    }

    public void showOptionsDialog() {
	StandardDialog newDialog = new StandardDialog(new OptionsPanel());
	newDialog.show(); 
    } 

    public ActivePanel getOptionsPanel() { 
	return new OptionsPanel(); 
    }

    public class OptionsPanel extends Sonar.Dialog.ActivePanel 
    {		 
	public OptionsPanel() 
	{ 
	    tabTitle = "Input";//this will be the title of the tab 
	    setLayout(null); 
	    setSize(307,168); 

	    //input string
	    /*
	    j_label1.setText("currently:"+class_input);
	    j_label1.setBounds(20,34, 300, 24);
	    add(j_label1);
	    */
	    
	    j_label2.setText("class:");
	    j_label2.setBounds(20,68, 80, 24);
	    add(j_label2);

	    j_textfield1.setToolTipText("type the class name of the msg you would like to display");
	    add(j_textfield1);
	    j_textfield1.setBounds(100, 68, 200, 24);

	    //choose a color
	    j_label3.setText("via:");
	    j_label3.setBounds(20, 92, 80, 24);
	    add(j_label3);
	    
	    //mark how its sent
	    j_combobox1.addItem(new String("as is"));
	    j_combobox1.addItem(new String("in a RouteMsg"));
	    j_combobox1.addItem(new String("in a RouteObjRootMsg"));
	    j_combobox1.addItem(new String("in a RouteObjDestMsg"));
	    add(j_combobox1);
	    j_combobox1.setBounds(90, 116, 210, 24);
	} 

	javax.swing.JTextField j_textfield1 = new javax.swing.JTextField(); 
	javax.swing.JComboBox j_combobox1 = new javax.swing.JComboBox();

	javax.swing.JLabel j_label1 = new javax.swing.JLabel();
	javax.swing.JLabel j_label2 = new javax.swing.JLabel();
	javax.swing.JLabel j_label3 = new javax.swing.JLabel();


    
	public void applyChanges() { 
	    String class_input = j_textfield1.getAccessibleContext().getAccessibleText().getAtIndex(AccessibleText.SENTENCE, 0);
	    //MainClass.out(class_input);
	    
	    if(j_combobox1.getSelectedItem().equals("as is") && 
		    !networkMsgs.contains(new QSString(class_input))) {
		networkMsgs.add(new QSString(class_input));
	    } else if(j_combobox1.getSelectedItem().equals("in a RouteMsg") &&
		    !routeMsgs.contains(new QSString(class_input))) {
		routeMsgs.add(new QSString(class_input));
	    } else if((j_combobox1.getSelectedItem().equals("in a RouteObjRootMsg") || j_combobox1.getSelectedItem().equals("in a RouteObjDestMsg")) && !routeObjMsgs.contains(new QSString(class_input))) {
		routeObjMsgs.add(new QSString(class_input));
	    }
	} 
    
	//this function will be called when the panel is first shown 
	public void initializeDisplayValues()
	{
	} 

    }	           
 
    protected class MenuManager implements Serializable, ActionListener
    { 
	JMenu mainMenu = new JMenu(); 
	JSeparator separator1 = new JSeparator(); 
	JMenuItem propertiesItem = new JMenuItem(); 
	JSeparator separator2 = new JSeparator(); 

	JMenuItem sendReqMsgItem = new JMenuItem(); 
	JSeparator separator3 = new JSeparator(); 

	public MenuManager() 
	{ 
	    mainMenu.setText("Input Analyzer"); 
	    mainMenu.setActionCommand("Input Analyzer"); 

	   
	    sendReqMsgItem.setText("Send Request");
	    sendReqMsgItem.setActionCommand("Send");
	    mainMenu.add(sendReqMsgItem); 

	    mainMenu.add(separator1); 
	    propertiesItem.setText("Options"); 
	    propertiesItem.setActionCommand("Options"); 
	    mainMenu.add(propertiesItem); 

	    MainClass.mainFrame.PacketAnalyzersMenu.add(mainMenu);

	    propertiesItem.addActionListener(this); 
	    sendReqMsgItem.addActionListener(this); 
	} 

	public void actionPerformed(ActionEvent e) 
	{ 
	    Object object = e.getSource(); 
	    if (object == propertiesItem) 
		showOptionsDialog(); 
	    else if(object == sendReqMsgItem)
		sendMessage();
	}

    }
}
