/**
 * LocalOptimizationTest.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl.test;
import java.util.Random;
import java.util.HashSet;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.network.NetworkMessage;
import ostore.tapestry.api.TapestryLocateFailure;
import ostore.tapestry.api.TapestryPrefixRouteMsg;
import ostore.tapestry.api.TapestryPublishMsg;
import ostore.tapestry.impl.TapestryReadyMsg;
import ostore.tapestry.impl.TestLocateMsg;
import ostore.tapestry.impl.RoutingTable;
import ostore.tapestry.impl.Router;
import ostore.util.Carp;
import ostore.util.Clock;
import ostore.util.Debug;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.QSException;
import ostore.util.RandomExponential;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkFullException;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.StagesInitializedSignal;

/**
 * TODO
 *
 * @author     Jeremy Stribling
 * @version    $Id: RouteTest.java,v 1.5 2003/04/21 23:21:27 strib Exp $
 */
public class RouteTest 
implements SingleThreadedEventHandlerIF, EventHandlerIF {

    protected String TAG = "RouteTest";
    protected static final boolean DEBUG = true;

    protected SinkIF this_sink;
    protected boolean start;
    protected long start_time;
    protected Classifier classifier;
    protected NodeId node_id;

    // config options
    private int sleep_time;
    private long random_seed;
    private boolean sim_running;

    private SecureHash self_guid;

    private RandomExponential rand;

    private long lastTimeSent;
    private SecureHash lastGuid;
    private int lastIndex;
    private Object lastToken;
    private byte [] bytes;

    private HashSet published_guids;

    // static (WARNING: exploiting simulator!!!!!)
    private static int seed_counter = 0;
    private static int seed_inc = 1;
    private static long time_first_started = 0;
    private static Object static_lock = new Object();

    public static class AlarmEvent implements QueueElementIF {}

    public void init (ConfigDataIF config) throws Exception {

	// Set up some debug stuff.

	TAG = config.getStage().getName();
        if (DEBUG) Debug.turnOn (TAG);

	// Set this_sink.

	this_sink = config.getStage().getSink();
	node_id = new NodeId (config.getString ("node_id"));
	classifier = Classifier.getClassifier (node_id);
	sim_running = config.getBoolean( "simulator_running" );

	// config opts
	sleep_time = config.getInt ("sleep_time");
	double lambda = (1.0/sleep_time);

	String random_str = config.getString( "random_seed" );
	if( random_str == null ) {
	    random_seed = -1;
	} else {
	    random_seed = Long.parseLong( random_str );
	}

	// give it a minute to start
	start_time = Clock.current_date(node_id).getTime() + 60000;

	// if we're starting up dynamically, then assume
	// everyone got the same random seed.  Try to do a
	// decentralized bucketing into groups of 10
	if( config.getString( "dynamic_route" ).equals( "dynamic" ) ) {
	    if( sim_running ) {
		synchronized( static_lock ) {
		    if( seed_counter == 0 ) {
			time_first_started = start_time;
		    } else {
			start_time = time_first_started;
		    }
		    seed_counter++;
		    random_seed += seed_inc;
		    if( seed_counter == 10 ) {
			seed_counter = 0;
			seed_inc++;
		    }
		}
	    } else {
		random_seed += ( node_id.hashCode() % 10 ) + 1;
	    }
	    if( DEBUG ) {
		Debug.printtagln( TAG, "Setting seed to " + random_seed );
	    }
	}

	rand = new RandomExponential( lambda, random_seed );
	
	// Create all of our filters.

	String [] inb_msg_types = {
	    "ostore.tapestry.impl.test.RouteTest$RouteMsg",
	    "ostore.tapestry.impl.test.RouteTest$Response",
	    "ostore.tapestry.api.TapestryLocateFailure"
	};

	ostore.util.TypeTable.register_type( "ostore.tapestry.impl.test.RouteTest$RouteMsg" );
	ostore.util.TypeTable.register_type( "ostore.tapestry.impl.test.RouteTest$Response" );

	Filter filter;
	for (int i = 0; i < inb_msg_types.length; ++i) {
	    filter = new Filter ();
	    if (! filter.requireType (Class.forName (inb_msg_types [i])))
		Carp.die ("could not require type " + inb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (true)))
		Carp.die ("could not require inbound = true for " 
			  + inb_msg_types [i]);
	    classifier.subscribe (filter, this_sink);
	}

	String [] event_types = {
	    "ostore.tapestry.impl.RoutingTable$Available",
	    "ostore.tapestry.impl.test.RouteTest$AlarmEvent",
	    "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.tapestry.impl.TapestryReadyMsg"
	};

	for (int i = 0; i < event_types.length; ++i) {
	    filter = new Filter ();
	    if (! filter.requireType (Class.forName (event_types [i])))
		Carp.die("could not require type " + event_types [i]);
	    classifier.subscribe (filter, this_sink);
	}

	bytes = new byte [21];

    }

    public void handleEvent (QueueElementIF item) 
    throws EventHandlerException {

	if( item instanceof RoutingTable.Available ) {
	    // do nothing
	    return;
	}

	if (DEBUG) Debug.printtagln (TAG, "event " + item);

	if (item instanceof StagesInitializedSignal) {
	    // do nothing

	}
	else if( item instanceof TapestryReadyMsg ) {
	    self_guid = ((TapestryReadyMsg)item).node_guid;
	    // don't start for a minute
	    long time_to_start = start_time - 
		Clock.current_date(node_id).getTime();
	    if( time_to_start < 0 ) {
		time_to_start = play_catchup( -time_to_start );
	    }
	    if( time_to_start == 0 ) {
		dispatch( new AlarmEvent() );
	    } else {
		classifier.dispatch_later (new AlarmEvent (), time_to_start );
	    }
	}
	else if (item instanceof RouteMsg) {
	    RouteMsg msg = (RouteMsg) item;
	    // send a response off
	    dispatch( new Response( msg.requestor, msg.hopCount, 
				    Clock.current_date(node_id).getTime(),
				    msg.guid, self_guid ) );
	}
	else if (item instanceof Response) {
	    got_resp ( (Response) item );
	    set_new_alarm();
	}
	else if (item instanceof TapestryLocateFailure) {
	    TapestryLocateFailure msg = (TapestryLocateFailure) item;
	    if( lastGuid == null || !msg.obj_guid.equals( lastGuid ) ) {
		if( DEBUG ) {
		    Debug.printtagln( TAG, 
				      "Got incorrect or late failure; " + 
				      "ignoring " + msg );
		}
		return;
	    }
	    handle_failure( msg );
	    set_new_alarm();
	}
	else if (item instanceof AlarmEvent) {
	    send_new_msg();
	} 
	else {
	    Carp.die (TAG + ": got unknown event of type " + 
		    item.getClass ().getName () + ": " + item + ".");
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    public void destroy() throws Exception {

    }

    // how long until my next message should be sent?
    private long get_next_time() {
	return (long) rand.nextExponential();
    }

    private void set_new_alarm() {

	// find the next time to start, and start then
	long time_to_next = get_next_time();
	long time_left = (time_to_next+lastTimeSent) - 
	    Clock.current_date(node_id).getTime();

	if (DEBUG) Debug.printtagln (TAG, "time_to_next: " + time_to_next + 
				     ":" + time_left + ":" + lastTimeSent + 
				     ":" + 
				     Clock.current_date(node_id).getTime());

	if( time_left < 0 ) {
	    time_left = play_catchup( -time_left );
	}
	if( time_left == 0 ) {
	    dispatch( new AlarmEvent() );
	} else {
	    classifier.dispatch_later (new AlarmEvent (), time_left);
	}

    }

    protected void got_resp( Response msg ) {
		
	if (msg != null) {
	    long currtime = Clock.current_date( node_id ).getTime();
	    long ping_time = currtime - msg.time;
	    long locate_time = msg.time - lastTimeSent;
	    double stretch;
	    if( ping_time != 0 ) {
		stretch = ((double)locate_time) / ((double)ping_time);
		if( msg.guid.equals( lastGuid ) ) {
		    // we found something, print out the time and the hopcount
		    System.out.println( "FOUND: " + node_id + 
					" routed to obj " + 
					msg.guid + " on node " + msg.peer + 
					" with guid " + msg.root_guid + 
					" at time " + currtime + 
					" in time " + locate_time +
					" with hopcount " + msg.hopcount + 
					" with stretch " + stretch + 
					" and ping time " + ping_time + 
					", index " + lastIndex );
		} else {
		    if( DEBUG ) {
			System.out.println( node_id + ": got duplicate or " + 
					    "unrequested response for " + 
					    msg.guid + ", ignoring." );
		    }
		    return;
		}
	    }
	    // cancel the last timeout
	    classifier.cancel( lastToken );
	}
	
    }

    private SecureHash get_next_guid() {

	rand.nextBytes(bytes);
	return new SHA1Hash( bytes );	

    }

    private void send_new_msg() {

	// find the next one
	SecureHash objguid = get_next_guid();	

	lastTimeSent = Clock.current_date( node_id ).getTime();

	if( DEBUG ) {
	    System.out.println( node_id + ": looking for obj " +
				objguid + " at time " + lastTimeSent );
	}	    

	lastGuid = objguid;
	RouteMsg rm = new RouteMsg( objguid, node_id );
	dispatch( rm );
	// if it takes more than a minute, timeout
	lastToken = 
	    classifier.dispatch_later( new TapestryLocateFailure(self_guid,
								 self_guid,
								 objguid,
								 node_id,
								 0 ),
				       10000 );
    
    }

    private boolean handle_failure( TapestryLocateFailure msg ) {
	boolean timeout = false;
	if( msg.root.equals(self_guid) && msg.id.equals(self_guid) ) {
	    timeout = true;
	}
	if( msg.obj_guid.equals( lastGuid ) ) {
	    long currtime = Clock.current_date( node_id ).getTime();
	    // we found something, print out the time and the hopcount
	    System.out.println( "FAILED: " + node_id + " failed to route " + 
				"to obj " 
				+ msg.obj_guid + " on node " + msg.peer + 
				" rootguid " + msg.root + " at time " + 
				currtime + " object index " +
				lastIndex + " timeout " + 
				timeout );
	} else {
	    if( DEBUG ) {
		System.out.println( node_id + ": got duplicate or " + 
				    "unrequested failure for " + 
				    msg.obj_guid + ", ignoring." );
	    }
	}
	// cancel the last timeout
	classifier.cancel( lastToken );
	return timeout;
    }

    private long play_catchup( long makeup_time ) {

	if( DEBUG ) {
	    Debug.printtagln( TAG, node_id + " is playing catchup for " + 
			      makeup_time + " ms." );
	}

	long passed_time = 0;
	while( passed_time < makeup_time ) {

	    // use the RNG
	    get_next_guid();
	    passed_time += get_next_time();

	}

	return (passed_time - makeup_time);

    }

    /**
     * An internal method used for dispatching events
     */
    private void dispatch( QueueElementIF event ) {

	if( DEBUG ) {
	    System.out.println( node_id + ": Dispatching " + event );
	}
                                                
      try {
        classifier.dispatch( event );
      } catch ( SinkFullException sfe ) {
        //TODO: Better way to deal with SinkFull
        try {
          Thread.currentThread().sleep( 5000 );
        } catch( InterruptedException ie ) {
          //ignore
        }
      }                                         
      
    }

    public static class RouteMsg extends TapestryPrefixRouteMsg {

	// the node who requests this prefix route message
	NodeId requestor;

	// the desired guid
	SecureHash guid;

	public RouteMsg( SecureHash id, NodeId requestor ) {
	    super(id, false);
	    this.requestor = requestor;
	    this.guid = id;
	}
	
	public RouteMsg (InputBuffer buffer) throws QSException {
	    //super( buffer );
	    requestor = (NodeId) buffer.nextObject();
	    guid = (SecureHash) buffer.nextObject();
	}

	public void serialize (OutputBuffer buffer) {
	    //super.serialize( buffer );
	    buffer.add( requestor );
	    buffer.add( guid );
	}

	public Object clone () throws CloneNotSupportedException {
	    RouteMsg rt = (RouteMsg) super.clone();
	    rt.requestor = requestor;
	    rt.guid = guid;
	    return rt;
	}
       
	public String toString () {
	    return "(RouteMsg, requestor=" + requestor + 
		", guid=" + guid + "), (" + super.toString () + ")";
	}

    }

    public static class Response extends NetworkMessage 
	implements QuickSerializable {

	int hopcount;
	long time;
	SecureHash guid;
	SecureHash root_guid;

	public Response (NodeId peer, int hopcount, long time, 
			 SecureHash guid, SecureHash root_guid) {
	    
	    super (peer, false /* outbound */);
	    this.hopcount = hopcount;
	    this.time = time;
	    this.guid = guid;
	    this.root_guid = root_guid;
	    if( root_guid == null ) {
		this.root_guid = new SHA1Hash().nullHash();
	    }
	}

	public Response (InputBuffer buffer) throws QSException {
	    super( buffer );
	    hopcount = buffer.nextInt();
	    time = buffer.nextLong();
	    guid = (SecureHash) buffer.nextObject();
	    root_guid = (SecureHash) buffer.nextObject();
	}

	public void serialize (OutputBuffer buffer) {
	    super.serialize( buffer );
	    buffer.add( hopcount );
	    buffer.add( time );
	    buffer.add( guid );
	    buffer.add( root_guid );
	}

	public Object clone () throws CloneNotSupportedException {
	    Response rt = (Response) super.clone();
	    rt.hopcount  = hopcount;
	    rt.time = time;
	    rt.guid = guid;
	    rt.root_guid = root_guid;
	    return rt;
	}
	
	public String toString () {
	    return "(Response hopcount=" + hopcount + ", time=" + time + 
		", guid=" + guid + ", root=" + root_guid + 
		"), (" + super.toString () + ")";
	}
    }

}
